/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.bioatlas.modules;

import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Vector;

import org.srs3d.viewer.bioatlas.colorschemes.MeshColorScheme;
import org.srs3d.viewer.bioatlas.objects.Residue;
import org.srs3d.viewer.bioatlas.objects.Subchain;
import org.srs3d.viewer.j3d.ColorScheme;
import org.srs3d.viewer.j3d.ColorSchemeBucket;
import org.srs3d.viewer.j3d.ColorSchemeManager;
import org.srs3d.viewer.j3d.ContextData;
import org.srs3d.viewer.j3d.commands.ColorCommand;
import org.srs3d.viewer.j3d.commands.ExecuteClassCommand;
import org.srs3d.viewer.j3d.commands.ParentCommand;
import org.srs3d.viewer.j3d.commands.PreSelectCommand;
import org.srs3d.viewer.j3d.commands.PropagateCommand;
import org.srs3d.viewer.objects.AbstractObject;
import org.srs3d.viewer.objects.Command;

/**
 * @author Karsten Klein
 *
 * @created August 06, 2001
 */
public class MeshCheck extends RepresentationCheck {
    private ColorSchemeBucket colorSchemeBucket = null;
    private transient ColorCommand onColorCommand = null;
    private transient ColorCommand offColorCommand = null;

    /**
     * Method description.
     *
     * @param contextData Parameter description.
     */
    public void create(ContextData contextData) {
        if (colorSchemeBucket == null) {
            ColorScheme colorScheme =
                (ColorScheme) contextData.getProperty(MeshColorScheme.class);
            if (colorScheme == null) {
                colorScheme = new MeshColorScheme(contextData);
                contextData.setProperty(MeshColorScheme.class, colorScheme);
            }
            colorSchemeBucket = new ColorSchemeBucket(colorScheme);
            colorSchemeBucket.setRemoving(false);
            colorSchemeBucket.setLocked(true);
            onColorCommand =
                new ColorCommand(contextData, getColorSchemeBucket());
            offColorCommand =
                new ColorCommand(contextData, getColorSchemeBucket());
        }
        onColorCommand.setForceRecoloring(true);
        onColorCommand.setExtendBucket(true);
        onColorCommand.setCollecting(false);
        offColorCommand.setForceRecoloring(true);
        offColorCommand.setRemoving(true);
        offColorCommand.setCollecting(false);
    }

    /**
     * Method description.
     *
     * @param contextData Parameter description.
     * @param objects Parameter description.
     *
     * @return Return description.
     */
    public int check(ContextData contextData, Collection objects) {
        boolean isActive = false;
        if (!objects.isEmpty()) {
            HashSet set = new HashSet(objects);
            contextData.getObjectManager().getDownAssociations(objects, set);
            create(contextData);
            Iterator iterator = set.iterator();
            ColorSchemeBucket bucket;
            Object object;
            ColorSchemeManager colorSchemeManager =
                contextData.getColorSchemeManager();
            while (!isActive && iterator.hasNext()) {
                object = iterator.next();
                bucket =
                    colorSchemeManager.getColorSchemeBucket((AbstractObject) object);
                isActive = check(bucket);
            }
            return isActive ? CHECK_YES : CHECK_NO;
        }
        return CHECK_FAIL;
    }

    /**
     * Gets the <code>allOnCommands</code> attribute of the <code>WireframeCheck</code>
     * class.
     *
     * @param contextData Description of parameter.
     * @param onCommands Description of parameter.
     */
    public void getAllOnCommands(ContextData contextData, Collection onCommands) {
        Command command = new PreSelectCommand(contextData);
        command = new ExecuteClassCommand(contextData, command, Subchain.class);
        command = new ParentCommand(contextData, command);
        command = new ExecuteClassCommand(contextData, command, Residue.class);
        onCommands.add(command);
        create(contextData);
        command = new PropagateCommand(contextData, onColorCommand);
        onCommands.add(command);
    }

    /**
     * Gets the <code>allOffCommands</code> attribute of the <code>WireframeCheck</code>
     * class.
     *
     * @param contextData Description of parameter.
     * @param offCommands Description of parameter.
     */
    public void getAllOffCommands(ContextData contextData,
        Collection offCommands) {
        create(contextData);
        Command command = new PropagateCommand(contextData, offColorCommand);
        offCommands.add(command);
    }

    /**
     * Method description.
     *
     * @param contextData Parameter description.
     *
     * @return Return description.
     */
    public static RepresentationModule createRepresentationModule(
        ContextData contextData) {
        Vector onCommands = new Vector();
        Vector offCommands = new Vector();

        // we cannot use the shared instance mechanism, since MeshCheck is not stateless
        MeshCheck check = new MeshCheck();
        check.getAllOnCommands(contextData, onCommands);
        check.getAllOffCommands(contextData, offCommands);
        RepresentationModule module =
            new RepresentationModule("Mesh", contextData, onCommands,
                offCommands, check);
        module.setSpawing(false);
        return module;
    }

    private boolean check(ColorSchemeBucket bucket) {
        if (bucket != null) {
            if (bucket.isEmpty()) {
                return false;
            } else {
                return bucket.containsAll(getColorSchemeBucket());
            }
        }
        return false;
    }

    /**
     * Method description.
     *
     * @return Return description.
     */
    public ColorSchemeBucket getColorSchemeBucket() {
        return colorSchemeBucket;
    }
}
