/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.j3d;

import java.util.Collection;
import java.util.HashSet;

import org.srs3d.viewer.bioatlas.colorschemes.SelectionColorScheme;
import org.srs3d.viewer.j3d.commands.ColorCommand;
import org.srs3d.viewer.objects.AbstractObject;

/**
 * The <code>Selection</code> class collects objects and provides methods for adding and
 * removing items from the current selection. Operations on the current selection can
 * easily be applied by reading the collection and processing the single elements. Note
 * that the method prefix reset... always indicates a coloring without the selection
 * coloring, while the prefix apply... always applies the colors including the selection
 * coloring. The selection works only on <code>AbstractObject</code> inherited
 * instances.
 *
 * @author Karsten Fries, LION bioscience, 02/2001
 *
 * @created May 2, 2001
 * @reviewed Oct 16, 2001
 */
public class Selection extends HashSet implements Contextual {
    private ColorScheme colorScheme = null;
    private String name = new String("<default>");
    private ContextData contextData = null;
    private boolean isColoringLocked = false;
    private boolean isColoringRemoveLocked = false;
    private boolean isExpanding = false;
    private boolean isPropagatingColoring = true;
    private transient boolean isAutoColoring = true;

    /**
     * <code>Selection</code> constructor Creates a default instance. The
     * SelectionColorScheme is used to brighten the normal object color using a gray
     * coloring offset.
     *
     * @param contextData Description of parameter.
     */
    public Selection(ContextData contextData) {
        this.contextData = contextData;
        colorScheme = new SelectionColorScheme(contextData);

        //    colorScheme = new SelectionColorScheme( contextData, new Color3f( 0.5f, 0.5f, 0.5f ), false );
        // :FIXME: implement this in a different way; perhaps a context property
        if (contextData.getContext() instanceof org.srs3d.viewer.bioatlas.contexts.MoleculeContext) {
            isExpanding = true;
        }
    }

    /**
     * Sets the <code>ColorScheme</code> attribute of the <code>Selection</code> object.
     *
     * @param colorScheme The new <code>colorScheme</code> value.
     */
    public void setColorScheme(ColorScheme colorScheme) {
        this.colorScheme = colorScheme;
    }

    /**
     * Sets the <code>Name</code> attribute of the <code>Selection</code> object.
     *
     * @param name The new <code>Name</code> value.
     */
    public void setName(String name) {
        this.name = new String(name);
    }

    /**
     * Gets the <code>Name</code> attribute of the <code>Selection</code> object.
     *
     * @return The <code>Name</code> value.
     */
    public String getName() {
        return new String(name);
    }

    /**
     * Gets the <code>contextData</code> attribute of the <code>Selection</code> object.
     *
     * @return The <code>contextData</code> value.
     */
    public final ContextData getContextData() {
        return contextData;
    }

    /**
     * Gets the <code>colorScheme</code> attribute of the <code>Selection</code> object.
     *
     * @return The <code>colorScheme</code> value.
     */
    public ColorScheme getColorScheme() {
        return colorScheme;
    }

    /**
     * To permit normal usage of the add method this implementation executes an select
     * command on the specified object.
     *
     * @param object Object to add to selection.
     *
     * @return Always <code>true</code>.
     */
    public boolean add(Object object) {
        if (!contains(object)) {
            if (!isColoringLocked && isAutoColoring) {
                applyColoring((AbstractObject) object);
            }
            return super.add(object);
        }
        return true;
    }

    /**
     * Triggers a selection command on the selection.
     *
     * @param collection Objects to to be added to the selection.
     *
     * @return Always <code>true</code>.
     */
    public boolean addAll(Collection objects) {
        if (!isColoringLocked && isAutoColoring) {
            applyColoring(objects);
        }
        isAutoColoring = false;
        boolean result = super.addAll(objects);
        isAutoColoring = true;
        return result;
    }

    /**
     * Method description.
     *
     * @param objects Parameter description.
     * @param isColoring Parameter description.
     *
     * @return Return description.
     */
    public boolean addAll(Collection objects, boolean isColoring) {
        if (!isColoringLocked && isAutoColoring && isColoring) {
            applyColoring(objects);
        }
        isAutoColoring = false;
        boolean result = super.addAll(objects);
        isAutoColoring = true;
        return result;
    }

    /**
     * Applies the selection coloring to the given object.
     *
     * @param object Object to color.
     */
    public void applyColoring(AbstractObject object) {
        if (!isColoringLocked) {
            ColorCommand colorCommand =
                new ColorCommand(getContextData(), colorScheme);
            colorCommand.setForceRecoloring(true);
            colorCommand.setRespawning(false);
            colorCommand.setExpanding(isExpanding);
            colorCommand.setExtendBucket(!colorScheme.isVolatile());
            if (isPropagatingColoring) {
                getContextData().getStrategyManager().propagate(object,
                    colorCommand);
            } else {
                getContextData().getStrategyManager().execute(object,
                    colorCommand);
            }
        }
    }

    /**
     * Description of the method.
     *
     * @param objects Description of parameter.
     */
    public void applyColoring(Collection objects) {
        if (!isColoringLocked) {
            ColorCommand colorCommand =
                new ColorCommand(getContextData(), colorScheme);
            colorCommand.setForceRecoloring(true);
            colorCommand.setRespawning(false);
            colorCommand.setExpanding(isExpanding);
            colorCommand.setExtendBucket(!colorScheme.isVolatile());
            if (isPropagatingColoring) {
                getContextData().getStrategyManager().propagate(objects,
                    colorCommand, false);
            } else {
                getContextData().getStrategyManager().execute(objects,
                    colorCommand);
            }
        }
    }

    /**
     * Remove object from selection. The coloring of the object is reset.
     *
     * @param object Object to remove.
     *
     * @return Flag indicating success.
     */
    public final boolean remove(Object object) {

        // OPT 1906
        if (!isColoringRemoveLocked && isAutoColoring) {
            resetColoring((AbstractObject) object, true);
        }
        isAutoColoring = false;
        boolean result = super.remove(object);
        isAutoColoring = true;
        return result;
    }

    /**
     * Removes all object from the selection. All object coloring is reset.
     *
     * @param collection Objects to remove.
     *
     * @return Flag inicating success.
     */
    public boolean removeAll(Collection collection) {
        if (!isColoringRemoveLocked && isAutoColoring) {
            resetColoring(collection, true);
        }
        isAutoColoring = false;
        boolean result = super.removeAll(collection);
        isAutoColoring = true;
        return result;
    }

    /**
     * Description of the method.
     */
    public void clear() {
        if (!isEmpty()) {
            if (!isColoringRemoveLocked && isAutoColoring) {
                resetColoring(this, true);
            }
        }
        isAutoColoring = false;
        super.clear();
        isAutoColoring = true;
    }

    /**
     * Method description.
     *
     * @param isRecoloring Parameter description.
     */
    public void clear(boolean isRecoloring) {
        if (isRecoloring) {
            resetColoring(this, true);
        }
        isAutoColoring = false;
        super.clear();
        isAutoColoring = true;
    }

    /**
     * Resets the coloring of an objects to a colring without selection colors. Note that
     * the return value is only non-null if the applied flag is set. Otherwise only the
     * ColorSchemeBucket of the object is modified.
     *
     * @param object Description of parameter.
     * @param isApplied Description of parameter.
     */
    public void resetColoring(AbstractObject object, boolean isApplied) {
        ColorCommand colorCommand =
            new ColorCommand(getContextData(), colorScheme);
        colorCommand.setRespawning(false);
        colorCommand.setExpanding(false);
        colorCommand.setRemoving(true);
        if (isPropagatingColoring) {
            getContextData().getStrategyManager().propagate(object, colorCommand);
        } else {
            getContextData().getStrategyManager().execute(object, colorCommand);
        }
    }

    /**
     * Description of the method.
     *
     * @param objects Description of parameter.
     * @param isApplied Description of parameter.
     */
    public void resetColoring(Collection objects, boolean isApplied) {
        ColorCommand colorCommand =
            new ColorCommand(getContextData(), colorScheme);
        colorCommand.setRespawning(false);
        colorCommand.setExpanding(false);
        colorCommand.setRemoving(true);
        colorCommand.execute(objects);
        if (isPropagatingColoring) {
            getContextData().getStrategyManager().propagate(this, colorCommand,
                false);
        } else {
            getContextData().getStrategyManager().execute(this, colorCommand);
        }
    }

    /**
     * Resets the coloring of all objects in the selection.
     *
     * @param isApplied Applies the coloring to become visible if true.
     */
    public void resetColoring(boolean isApplied) {
        resetColoring(this, isApplied);
    }

    /**
     * Description of the method.
     */
    public void removeAll() {
        clear();
    }

    /**
     * Colors the objects in the selection in the selection color
     */
    public void applyColoring() {
        applyColoring(this);
    }

    /**
     * Resets the coloring of all objects in the selection.
     */
    public void resetColoring() {
        resetColoring(true);
    }

    /**
     * Recolors the selection. The selection will have the selection coloring.
     */
    public void reapplyColorScheme() {
        applyColoring();
    }

    /**
     * Method description.
     *
     * @param isRecolor Parameter description.
     */
    public void reapplyColorScheme(boolean isRecolor) {
        if (!isEmpty()) {
            applyColoring();
        }
    }

    /**
     * Complete disposition of the selection object.
     */
    public void cleanup() {
        contextData = null;
        colorScheme = null;
    }

    /**
     * Method description.
     *
     * @param isColoringAddLocked Parameter description.
     */
    public void setColoringLocked(boolean isColoringLocked) {
        this.isColoringLocked = isColoringLocked;
    }

    /**
     * Method description.
     *
     * @param isColoringRemoveLocked Parameter description.
     */
    public void setColoringRemoveLocked(boolean isColoringRemoveLocked) {
        this.isColoringRemoveLocked = isColoringRemoveLocked;
    }

    /**
     * Method description.
     *
     * @param isPropagatingColoring Parameter description.
     */
    public void setPropagatingColoring(boolean isPropagatingColoring) {
        this.isPropagatingColoring = isPropagatingColoring;
    }
}
