/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.j3d;

import java.util.Collection;
import java.util.ConcurrentModificationException;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;

import javax.media.j3d.Appearance;
import javax.media.j3d.CapabilityNotSetException;
import javax.media.j3d.GeometryArray;
import javax.media.j3d.LineArray;
import javax.media.j3d.QuadArray;
import javax.media.j3d.Shape3D;
import javax.media.j3d.TriangleArray;
import javax.media.j3d.TriangleFanArray;
import javax.media.j3d.TriangleStripArray;
import javax.vecmath.Point3f;
import javax.vecmath.TexCoord2f;
import javax.vecmath.Vector3f;

import org.srs3d.viewer.objects.AbstractManager;
import org.srs3d.viewer.objects.AbstractObject;
import org.srs3d.viewer.util.ExceptionHandler;
import org.srs3d.viewer.util.Log;

import com.sun.j3d.utils.picking.PickTool;

/**
 * The <code>ShapeManager</code> handles all <code>Shape</code> objects that are
 * associated with an <code>AbstractObject</code> . You can associate multiple shapes
 * with one object and single shapes to multiple objects.
 *
 * @author Karsten Klein
 *
 * @created June 14, 2001
 */
public class ShapeManager extends AbstractManager {
    private static final Log log = new Log(ShapeManager.class);
    private static final boolean isRuntimeTestEnabled = false;

    /** This <code>Map</code> maps the object to a <code>Collection</code> of shapes. */
    private Map objectMap = new Hashtable();
    private HashSet lockedShapes = new HashSet();

    /**
     * <code>ShapeManager</code> constructor.
     *
     * @param contextData associated <code>ContextData</code> instance.
     */
    public ShapeManager(ContextData contextData) {
        super(contextData);
    }

    /**
     * Sets the appearance of all shapes associated with <code>object</code> .
     *
     * @param object object specifying the shapes.
     * @param appearance <code>Appearance</code> object to apply to the shapes.
     */
    public void setUniformAppearance(AbstractObject object,
        Appearance appearance) {
        Collection shapes = getShapes(object);
        if (shapes != null) {
            Iterator iterator = shapes.iterator();
            Shape3D shape;
            while (iterator.hasNext()) {
                shape = (Shape3D) iterator.next();
                try {
                    shape.setAppearance(appearance);
                } catch (Exception e) {
                    ExceptionHandler.handleException(e,
                        ExceptionHandler.SILENT_IN_RELEASE, this);
                    if (isVerbose()) {
                        log.debug("appearance of " + object +
                            " not updated with " + appearance);
                        log.debug(shape);
                        Enumeration enumer = shape.getAllGeometries();
                        while (enumer.hasMoreElements()) {
                            log.debug(enumer.nextElement());
                        }
                        log.debug(shape.getUserData());
                    }
                }
            }
        }
    }

    /**
     * Sets the <code>shapeLock</code> attribute of the <code>ShapeManager</code> object.
     *
     * @param shape The new <code>shapeLock</code> value.
     * @param isShapeLock The new <code>shapeLock</code> value.
     */
    public void setShapeLock(Shape3D shape, boolean isShapeLock) {
        if (isShapeLock) {
            lockedShapes.add(shape);
        } else {
            lockedShapes.remove(shape);
        }
    }

    /**
     * Returns a <code>Collection</code> of shapes, that are associated with the
     * specified object.
     *
     * @param object the object defines the shapes of interest.
     *
     * @return <code>Collection</code> - the gathered shapes that are associated with
     *         <code>object</code> .
     */
    public Collection getShapes(AbstractObject object) {
        return (Collection) objectMap.get(object);
    }

    /**
     * Retrieves all objects that are associated with a shape.
     *
     * @return The <code>objects</code> value.
     */
    public Collection getObjects() {
        return objectMap.keySet();
    }

    /**
     * Registers an object to shape association.
     *
     * @param object object to register the shape to.
     * @param shape shape to register.
     */
    public void register(AbstractObject object, Shape3D shape) {
        Collection shapes = getShapes(object);
        if (shapes == null) {
            shapes = new HashSet();
            objectMap.put(object, shapes);
        }
        if (isVerbose()) {
            if (shape.getUserData() == null) {
                log.debug("shape with no user data registered to object " +
                    object + "(" + object.getClass() + ")");
            }
        }
        shapes.add(shape);
        Appearance appearance =
            getContextData().getAppearanceManager().getAppearance(object);
        shape.setAppearance(appearance);
        if (isRuntimeTestEnabled && !checkShape(shape)) {
            log.error("check failed on " + object);
        }
    }

    /**
     * Clears all data of the manager.
     */
    public void clear() {

        // release lock on all locked shapes
        lockedShapes.clear();

        // clear all shapes (recently unlocked included)
        Iterator iterator = objectMap.values().iterator();
        Collection shapes;
        while (iterator.hasNext()) {
            shapes = (Collection) iterator.next();
            cleanup(shapes);
        }
        objectMap.clear();
    }

    /**
     * Removes all object related shape references.
     *
     * @param object specifies the shapes to remove.
     */
    public void remove(AbstractObject object) {
        Collection shapes = getShapes(object);
        if (shapes != null) {
            if (cleanup(shapes)) {
                objectMap.remove(object);
            } else {

                // found a look on at least one shape of the list
            }
        }
    }

    /**
     * A unit test for JUnit.
     *
     * @param object Description of parameter.
     */
    public void testShapes(AbstractObject object) {
        Collection shapes = getShapes(object);
        if (shapes != null) {
            Iterator iterator = shapes.iterator();
            Shape3D shape;
            while (iterator.hasNext()) {
                shape = (Shape3D) iterator.next();
                if (shape.getUserData() == null) {
                    log.debug("object(" + object + object.hashCode() +
                        ") shape (" + shape +
                        ") without appropriate user data.");
                }
            }
        }
    }

    /**
     * A unit test for JUnit.
     */
    public void testShapes() {
        Iterator iterator = objectMap.keySet().iterator();
        while (iterator.hasNext()) {
            testShapes((AbstractObject) iterator.next());
        }
    }

    /**
     * Runtime check for shapes.
     *
     * @param shape Shape for testing the geoemtry.
     *
     * @return <code>true</code> if check passed without error.
     */
    public boolean checkShape(Shape3D shape) {
        Enumeration enumer = shape.getAllGeometries();
        GeometryArray geometry;
        boolean isOk = true;
        while (enumer.hasMoreElements()) {
            try {
                geometry = (GeometryArray) enumer.nextElement();
                isOk &= checkGeometry(geometry);
            } catch (ClassCastException e) {

                // silent, we're not yet checking al kinds of geometry
                // incomplete list in checkGeometry
                ExceptionHandler.handleException(e,
                    ExceptionHandler.SILENT_IN_DEBUG, this);
            }
        }
        return isOk;
    }

    /**
     * Checks a GeometryArray
     *
     * @param geometry GeometryArray for checking.
     *
     * @return <code>true</code> if check passed without error.
     */
    public boolean checkGeometry(GeometryArray geometry) {
        boolean isOk = true;
        if (geometry != null) {
            isOk = checkTexturing(geometry);
            if (geometry instanceof TriangleFanArray) {
                isOk &= checkGeometry((TriangleFanArray) geometry);
            } else if (geometry instanceof TriangleStripArray) {
                isOk &= checkGeometry((TriangleStripArray) geometry);
            } else if (geometry instanceof TriangleArray) {
                isOk &= checkGeometry((TriangleArray) geometry);
            } else if (geometry instanceof LineArray) {
                isOk &= checkGeometry((LineArray) geometry);
            } else if (geometry instanceof QuadArray) {
                isOk &= checkGeometry((QuadArray) geometry);
            } else {

                // no check implemented for this geometry type
            }
        }
        return isOk;
    }

    /**
     * Description of the method.
     *
     * @param geometry Description of parameter.
     *
     * @return Description of the returned value.
     */
    public boolean checkGeometry(TriangleStripArray geometry) {
        boolean isOk = true;
        float angle;
        Point3f point = new Point3f();
        Point3f secondPoint = new Point3f();
        Point3f thirdPoint = new Point3f();
        int[] strips = new int[geometry.getNumStrips()];
        geometry.getStripVertexCounts(strips);
        int counter = 0;
        int limit = 0;
        for (int j = 0; j < strips.length; j++) {
            limit = counter + strips[j];
            for (int i = counter; i < limit - 2; i += 1) {
                geometry.getCoordinate(i, point);
                geometry.getCoordinate(i + 1, secondPoint);
                geometry.getCoordinate(i + 2, thirdPoint);
                secondPoint.sub(point);
                thirdPoint.sub(point);
                angle =
                    new Vector3f(secondPoint).angle(new Vector3f(thirdPoint));
                if (angle < 0.00001f || angle > Math.PI - 0.00001f) {
                    log.error("degenerated strip triangle " + angle);
                    log.error("index: " + i + "(strip start:" + counter +
                        "; strip end:" + (limit - 1) + ")");
                    geometry.getCoordinate(i + 1, secondPoint);
                    geometry.getCoordinate(i + 2, thirdPoint);
                    log.error(point);
                    log.error(secondPoint);
                    log.error(thirdPoint);
                    isOk = false;
                }
            }
            counter = limit;
        }
        return isOk;
    }

    /**
     * Description of the method.
     *
     * @param geometry Description of parameter.
     *
     * @return Description of the returned value.
     */
    public boolean checkGeometry(QuadArray geometry) {
        return true;
    }

    /**
     * Description of the method.
     *
     * @param geometry Description of parameter.
     *
     * @return Description of the returned value.
     */
    public boolean checkGeometry(TriangleArray geometry) {
        boolean isOk = true;
        float angle;
        Point3f point = new Point3f();
        Point3f secondPoint = new Point3f();
        Point3f thirdPoint = new Point3f();
        for (int i = 0; i < geometry.getVertexCount() - 2; i += 3) {
            geometry.getCoordinate(i, point);
            geometry.getCoordinate(i + 1, secondPoint);
            geometry.getCoordinate(i + 2, thirdPoint);
            secondPoint.sub(point);
            thirdPoint.sub(point);
            angle = new Vector3f(secondPoint).angle(new Vector3f(thirdPoint));
            if (angle < 0.00001f || angle > Math.PI - 0.00001f) {
                log.error("degenerated triangle");
                log.error("index: " + i);
                geometry.getCoordinate(i + 1, secondPoint);
                geometry.getCoordinate(i + 2, thirdPoint);
                log.error(point);
                log.error(secondPoint);
                log.error(thirdPoint);
                isOk = false;
            }
        }
        return isOk;
    }

    /**
     * Description of the method.
     *
     * @param geometry Description of parameter.
     *
     * @return Description of the returned value.
     */
    public boolean checkGeometry(LineArray geometry) {
        Point3f point = new Point3f();
        for (int i = 0; i < geometry.getVertexCount(); i++) {
            geometry.getCoordinate(i, point);
            if (point.x == 0 && point.y == 0 && point.z == 0) {
                log.error("suspicious point at " + i + "(" +
                    geometry.getVertexCount() + ")");
                log.error(geometry);
                return false;
            }
        }
        return true;
    }

    /**
     * Description of the method.
     *
     * @param geometry Description of parameter.
     *
     * @return Description of the returned value.
     */
    public boolean checkGeometry(TriangleFanArray geometry) {
        boolean isOk = true;

        // iterator over coordinates an check for equal coordinates
        Point3f point = new Point3f();
        Point3f secondPoint = new Point3f();
        Point3f thirdPoint = new Point3f();
        for (int i = 0; i < geometry.getVertexCount(); i++) {
            geometry.getCoordinate(i, point);
            for (int j = i + 1; j < geometry.getVertexCount(); j++) {
                geometry.getCoordinate(j, secondPoint);
                if (secondPoint.epsilonEquals(point, 0.001f)) {
                    if (i != 0 && j != geometry.getVertexCount() - 1) {
                        log.error("error in " + geometry);
                        log.error("point " + i + " and " + j + " identical");
                        log.error("i: " + point);
                        log.error("j: " + secondPoint);
                        isOk = false;
                    }
                }
            }
        }
        geometry.getCoordinate(0, point);
        float angle;
        for (int i = 1; i < geometry.getVertexCount() - 1; i++) {
            geometry.getCoordinate(i, secondPoint);
            geometry.getCoordinate(i + 1, thirdPoint);
            secondPoint.sub(point);
            thirdPoint.sub(point);
            angle = new Vector3f(secondPoint).angle(new Vector3f(thirdPoint));
            if (angle < 0.00001f || angle > Math.PI - 0.00001f) {
                log.error("degenerated fan triangle");
                log.error("index: " + i);
                isOk = false;
            }
        }
        return isOk;
    }

    /**
     * Description of the method.
     *
     * @param geometry Description of parameter.
     *
     * @return Description of the returned value.
     */
    public boolean checkTexturing(GeometryArray geometry) {
        boolean isOk = true;
        if ((geometry.getVertexFormat() & GeometryArray.TEXTURE_COORDINATE_2) != 0) {
            Point3f point = new Point3f();
            Point3f secondPoint = new Point3f();
            TexCoord2f texCoordA = new TexCoord2f();
            TexCoord2f texCoordB = new TexCoord2f();
            for (int i = 0; i < geometry.getVertexCount(); i++) {
                geometry.getCoordinate(i, point);
                geometry.getTextureCoordinate(0, i, texCoordA);
                if (texCoordA.x < 0.0f || texCoordA.x > 1.0f) {
                    log.error("texture coordinate out of scope " + texCoordA);
                    isOk = false;
                }
                for (int j = i + 1; j < geometry.getVertexCount(); j++) {
                    geometry.getCoordinate(j, secondPoint);
                    if (secondPoint.epsilonEquals(point, 0.001f)) {

                        // compare texture coordinate
                        geometry.getTextureCoordinate(0, j, texCoordB);
                        if (!texCoordA.epsilonEquals(texCoordB, 0.001f)) {
                            log.error("texture coordinate check failed " + i +
                                " " + j);
                            log.error(point.toString() + " " + texCoordA);
                            log.error(secondPoint.toString() + " " + texCoordB);
                            isOk = false;
                        }
                    }
                }
            }
        }
        return isOk;
    }

    /**
     * Cleanup all the shapes in the collection for finalization.
     *
     * @param shapes Collection of <code>Shape3D</code> objects.
     *
     * @return True if all the shapes where removed.
     */
    private boolean cleanup(Collection shapes) {
        Iterator iterator = shapes.iterator();
        Shape3D shape;
        int index;
        while (iterator.hasNext()) {
            shape = (Shape3D) iterator.next();
            if (!lockedShapes.contains(shape)) {
                try {
                    if (shape.getCapability(Shape3D.ALLOW_APPEARANCE_WRITE)) {
                        shape.setAppearance(null);
                    }
                } catch (CapabilityNotSetException e) {

                    // catch silently, often we have shapes that don't allow appearance
                    // setting
                    ExceptionHandler.handleException(e,
                        ExceptionHandler.SILENT_IN_DEBUG, this);
                } catch (Exception e) {
                    ExceptionHandler.handleException(e,
                        ExceptionHandler.SILENT_IN_RELEASE, this);
                }

                // finally clear object reference from shape
                shape.setUserData(null);
                shape = null;
                try {

                    // remove shape from collection
                    iterator.remove();
                } catch (ConcurrentModificationException e) {
                    ExceptionHandler.handleException(e,
                        ExceptionHandler.SILENT_IN_RELEASE, this);

                    // reset iterator
                    shapes.remove(shape);
                    iterator = shapes.iterator();
                }
            }
        }
        return shapes.isEmpty();
    }

    /**
     * Sets the capabilities of the specified shape. Includes setting the shape object
     * reference to the given <code>AbstractObject</code> .
     *
     * @param shape The shape to set capabilities.
     * @param object The object the shape belongs to.
     */
    public static void setCapabilities(Shape3D shape, Object object) {

        // reference object from the shape
        shape.setUserData(object);

        // defined set of capabilites
        shape.setCapability(Shape3D.ALLOW_APPEARANCE_READ);
        shape.setCapability(Shape3D.ALLOW_APPEARANCE_WRITE);
        shape.setCapability(Shape3D.ALLOW_BOUNDS_READ);
        try {
            PickTool.setCapabilities(shape, PickTool.INTERSECT_FULL);
        } catch (Exception e) {

            // empty
            // :NOTE: this happens i.e. with objects having shared memory shared geometry.
            //   this geometry is already life and so you cannot change attributes.
            //   if you need a different behavior, create your own (unshared) geometry
            //   this exception will always happen for shared geometry attribute
            //   setting. don't provide any output here.
            ExceptionHandler.handleException(e, ExceptionHandler.SILENT_IN_DEBUG);
        }
    }
}
