/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.j3d;

import java.util.ConcurrentModificationException;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;

import org.srs3d.viewer.objects.AbstractManager;
import org.srs3d.viewer.objects.AbstractObject;
import org.srs3d.viewer.util.ExceptionHandler;

/**
 * Manages all <code>Spawner</code> objects is the system. The class provides additional
 * methods for retrieving and manipulating the appropriate <code>Spawner</code> s.
 *
 * @author Karsten Klein
 *
 * @created November 26, 2000
 */
public class SpawnerManager extends AbstractManager {
    private Map spawnerMap = new Hashtable();

    /**
     * Constructor description.
     *
     * @param contextData Parameter description.
     */
    public SpawnerManager(ContextData contextData) {
        super(contextData);
    }

    /**
     * Returns the <code>Spawner</code> of the given objects. object.
     *
     * @param object Object identifying the <code>Spawner</code> .
     *
     * @return The <code>Spawner</code> object, or <code>null</code> if not registered.
     */
    public Spawner getSpawner(AbstractObject object) {
        Spawner spawner = (Spawner) spawnerMap.get(object);
        return spawner;
    }

    /**
     * Registers a <code>Spawner</code> to an object.
     *
     * @param object Object.
     * @param spawner <code>Spawner</code> to associate with the object.
     */
    public void register(AbstractObject object, Spawner spawner) {
        spawnerMap.put(object, spawner);
    }

    /**
     * Clears the manager. All Spawners are removed from the system.
     */
    public void clear() {
        HashSet set = new HashSet(spawnerMap.values());
        Iterator iterator = set.iterator();
        Spawner spawner;
        while (iterator.hasNext()) {
            spawner = (Spawner) iterator.next();
            try {
                iterator.remove();
            } catch (ConcurrentModificationException e) {
                ExceptionHandler.handleException(e,
                    ExceptionHandler.SILENT_IN_RELEASE);

                // restart iteration
                iterator = set.iterator();
            }
            spawner.cleanup(getContextData());
        }
    }

    /**
     * The <code>Spawner</code> object will be removed and cleaned up for finalization.
     *
     * @param object Description of parameter.
     */
    public void remove(AbstractObject object) {
        Spawner spawner = (Spawner) spawnerMap.remove(object);
        if (spawner != null) {
            spawner.cleanup(getContextData());
        }
    }
}
