/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.j3d;

import javax.media.j3d.BoundingSphere;
import javax.media.j3d.Bounds;
import javax.media.j3d.BranchGroup;
import javax.media.j3d.Group;
import javax.media.j3d.Transform3D;
import javax.media.j3d.TransformGroup;
import javax.vecmath.Point3d;
import javax.vecmath.Tuple3f;
import javax.vecmath.Vector3f;

import org.srs3d.viewer.j3d.behaviors.RotateBehavior;
import org.srs3d.viewer.j3d.behaviors.RotateXYBehavior;
import org.srs3d.viewer.j3d.commands.ComputeCenterCommand;
import org.srs3d.viewer.objects.AbstractObject;

/**
 * @author Karsten Klein
 *
 * @created November 14, 2000
 */
public class Transform {

    /** Description of the field */
    public final BranchGroup entry;

    /** Description of the field */
    public final TransformGroup transformGroup;

    /** Description of the field. */
    private boolean enabled = false;
    private RotateXYBehavior rotateXYBehavior = null;
    private RotateBehavior rotateBehavior = null;
//    private KinectRotateBehavior rotateBehavior = null;
    
    /**
     * <code>Transform</code> constructor.
     */
    public Transform() {
        entry = new BranchGroup();

        //    entry.setCapability( BranchGroup.ALLOW_DETACH );
        //    entry.setCapability( BranchGroup.ALLOW_CHILDREN_READ );
        //    entry.setCapability( BranchGroup.ALLOW_CHILDREN_WRITE );
        transformGroup = new TransformGroup();

        //    transformGroup.setCapability( TransformGroup.ALLOW_TRANSFORM_WRITE );
        //    transformGroup.setCapability( TransformGroup.ALLOW_TRANSFORM_READ );
        transformGroup.setCapability(TransformGroup.ALLOW_CHILDREN_EXTEND);
        entry.addChild(transformGroup);
    }

    /**
     * Sets the <code>Transform</code> attribute of the <code>Transform</code> object.
     *
     * @param translation The new <code>Transform</code> value.
     */
    public void setTransform(Tuple3f translation) {
        Transform3D transform = new Transform3D();
        transform.setTranslation(new Vector3f(translation));
        setTransform(transform);
    }

    /**
     * Sets the <code>Transform</code> attribute of the <code>Transform</code> object.
     *
     * @param transform The new <code>Transform</code> value.
     */
    public void setTransform(Transform3D transform) {
        transformGroup.setTransform(transform);
    }

    /**
     * Sets the <code>enabled</code> attribute of the <code>Transform</code> object.
     *
     * @param enabled The new <code>enabled</code> value.
     */
    public void setEnabled(boolean enabled) {
        this.enabled = enabled;
        if (rotateBehavior != null) {
            rotateBehavior.setEnable(enabled);
        }
        if (rotateXYBehavior != null) {
            rotateXYBehavior.setEnable(enabled);
        }
    }

    /**
     * Gets the <code>Transform</code> attribute of the <code>Transform</code> object.
     *
     * @return The <code>Transform</code> value.
     */
    public Transform3D getTransform() {
        Transform3D transform = new Transform3D();
        transformGroup.getTransform(transform);
        return transform;
    }

    /**
     * Gets the <code>enabled</code> attribute of the <code>Transform</code> object.
     *
     * @return The <code>enabled</code> value.
     */
    public boolean isEnabled() {
        return enabled;
    }

    /**
     * Description of the method.
     *
     * @param context Description of parameter.
     * @param object Description of parameter.
     */
    public void attachRotateBehaviors(ContextData contextData,
        AbstractObject object) {
        BranchGroup branchGroup = new BranchGroup();
        attachRotateXYBehavior(contextData, object, branchGroup);
        attachRotateBehavior(contextData, object, branchGroup);
        transformGroup.addChild(branchGroup);
    }

    /**
     * Description of the method.
     *
     * @param context Description of parameter.
     * @param object Description of parameter.
     */
    public void attachRotateXYBehavior(ContextData contextData,
        AbstractObject object, Group group) {
        if (rotateXYBehavior == null) {
            rotateXYBehavior = new RotateXYBehavior(contextData, object);
            rotateXYBehavior.setSchedulingBounds(computeBounds(contextData,
                    object));
            group.addChild(rotateXYBehavior);
            rotateXYBehavior.setEnable(enabled);
        }
    }

    /**
     * Description of the method.
     *
     * @param context Description of parameter.
     * @param object Description of parameter.
     */
    public void attachRotateBehavior(ContextData contextData,
        AbstractObject object, Group group) {
        if (rotateBehavior == null) {
            rotateBehavior = new RotateBehavior(contextData, object);
            rotateBehavior.setSchedulingBounds(computeBounds(contextData, object));
            group.addChild(rotateBehavior);
            rotateBehavior.setEnable(enabled);
        }
    }

    private Bounds computeBounds(ContextData contextData, AbstractObject object) {

        //    :NOTE: uncomment for adaptive bounds
        ComputeCenterCommand computeCenterCommand =
            new ComputeCenterCommand(contextData);
        contextData.getStrategyManager().execute(object, computeCenterCommand);
        return new BoundingSphere(new Point3d(),
            computeCenterCommand.getExtend().length());

        //    return new BoundingSphere( new Point3d( computeCenterCommand.getCenter() ), computeCenterCommand.getExtend().length() / 2 );
        //    return new BoundingSphere( new Point3d(), Double.MAX_VALUE );
    }

    /**
     * Method description.
     *
     * @return Return description.
     */
    public BranchGroup getEntry() {
        return entry;
    }
}
