/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.j3d.behaviors;

import java.awt.event.MouseEvent;

import javax.media.j3d.WakeupOnAWTEvent;
import javax.media.j3d.WakeupOr;
import javax.vecmath.AxisAngle4f;
import javax.vecmath.Matrix3f;

import org.srs3d.viewer.j3d.ContextData;
import org.srs3d.viewer.j3d.commands.RotateCommand;
import org.srs3d.viewer.objects.AbstractObject;

/**
 * The class <code>RotateBehavior</code> represents a specialized
 * <code>MouseBehavior</code> .
 *
 * @author Karsten Klein, 11/2000
 *
 * @created March 20, 2001
 * @since 1.0
 */
public class RotateXYBehavior extends MouseBehavior {
    private AbstractObject object = null;

    /** Integers used for saving the mouse positions */
    private int lastX;

    /** Integers used for saving the mouse positions */
    private int x;
    private int lastY;
    private int y;

    /**
     * <code>RotateXYBehavior</code> contructor.
     *
     * @param context Description of parameter.
     * @param object Description of parameter.
     */
    public RotateXYBehavior(ContextData contextData, AbstractObject object) {
        super(contextData);
        this.object = object;
    }

    /**
     * Initializes the wakeup conditions for the behavior.
     */
    public void initialize() {
        WakeupOnAWTEvent[] conditions = new WakeupOnAWTEvent[2];

        // set wakeup condition (waiting for further dragging events)
        conditions[0] = new WakeupOnAWTEvent(MouseEvent.MOUSE_DRAGGED);

        // set wakeup condition (waiting first click)
        conditions[1] = new WakeupOnAWTEvent(MouseEvent.MOUSE_PRESSED);
        wakeupOn(new WakeupOr(conditions));
    }

    /**
     * Processes the mouseEvent. This will directly result in a modification of the
     * viewingPlatform in the activated <code>Context</code> , if the according
     * conditions are met by the mouse event.
     *
     * @param mouseEvent Description of parameter.
     */
    public void processStimulus(MouseEvent mouseEvent) {
        lastX = x;
        lastY = y;
        x = mouseEvent.getX();
        y = mouseEvent.getY();
        if (checkCriteria(mouseEvent)) {
            AxisAngle4f axisAngleX =
                new AxisAngle4f(0, 1, 0, 4.0f / 180 * (x - lastX));
            AxisAngle4f axisAngleY =
                new AxisAngle4f(1, 0, 0, 4.0f / 180 * (y - lastY));
            Matrix3f matrix = new Matrix3f();
            matrix.set(axisAngleX);
            Matrix3f dummy = new Matrix3f();
            dummy.set(axisAngleY);
            matrix.mul(dummy);
            ContextData contextData = getContextData();
            RotateCommand rotateCommand = new RotateCommand(contextData);
            rotateCommand.setMatrix(matrix);
            contextData.getStrategyManager().execute(object, rotateCommand);
        }
    }

    /**
     * This method checks the mouseEvent (fail early, fail fast)
     *
     * @param mouseEvent Description of parameter.
     *
     * @return Description of the returned value.
     */
    protected boolean checkCriteria(MouseEvent mouseEvent) {
        if (mouseEvent.getID() != MouseEvent.MOUSE_DRAGGED) {
            return false;
        }
        if ((mouseEvent.getModifiers() & MouseEvent.BUTTON3_MASK) != 0) {
            return false;
        }
        if (mouseEvent.isShiftDown()) {
            return false;
        }
        if (mouseEvent.isControlDown()) {
            return false;
        }
        if (mouseEvent.isAltGraphDown()) {
            return false;
        }
        mouseEvent.consume();
        return true;
    }
}
