/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.objects;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.Map;

import org.srs3d.viewer.util.Log;

/**
 * The AbstractObject class defines a set of methods that implement default access and
 * cleanup functions.
 *
 * @author Karsten Klein, 02/2001
 * @version March 20, 2001
 */
public abstract class AbstractObject {
    private static final Log log = new Log(AbstractObject.class);

    /**
     * Default object retrieving method. This methods collect all the objects that are
     * directly coupled with the concrete instance.
     *
     * @param collection the collection the objects are added to.
     */
    public void getAllObjects(Collection collection) {
    }

    /**
     * Default updating method. The object builts its transient data.
     */
    public void update() {
    }

    /**
     * Explicitly clears a objects references. Implementations have to delete any
     * circular references. This implementation is kept very general and uses the object
     * retrieving with <code>getAllObjects( Collection )</code> . Note that you have to
     * call <code>super.cleanup()</code> before the reference reset.
     */
    public void cleanup() {
    }

    /**
     * Description of the method
     *
     * @param level Description of parameter
     */
    public void dump(int level) {
        String string = "";
        for (int i = 0; i < level; i++) {
            string += " ";
        }
        log.info(string + this);
        ArrayList vector = new ArrayList();
        getAllObjects(vector);
        Iterator iterator = vector.iterator();
        while (iterator.hasNext()) {
            ((AbstractObject) iterator.next()).dump(level + 1);
        }
    }

    /**
     * Small helper method that performs the null check before clearing the speicified
     * map.
     *
     * @param map Map to clean up.
     */
    public static final void cleanup(Map map) {
        if (map != null) {
            map.clear();
        }
    }

    /**
     * Small helper method to clean up a collection of object. The methods performs the
     * null check before clearing the collection.
     *
     * @param collection Collection to clean up.
     */
    public static final void cleanup(Collection collection) {
        if (collection != null) {
            collection.clear();
        }
    }

    /**
     * Method to explicitly cleanup the hierarchy subtree headed by the specified object.
     *
     * @param object Object to start the cleanup with.
     */
    public static void clearDeep(AbstractObject object) {
        ArrayList children = new ArrayList();
        object.getAllObjects(children);
        Iterator iterator = children.iterator();
        while (iterator.hasNext()) {
            clearDeep((AbstractObject) iterator.next());
        }
        children.clear();
        children = null;
        object.cleanup();
    }
}
