/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.objects;

import java.util.Collection;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;

import org.srs3d.viewer.util.Log;

/**
 * The <code>AssociatioManager</code> manages directional associations.
 *
 * @author Karsten Klein, 03/2001
 * @version March 20, 2001
 */
public final class AssociationManager {
    private static final Log log = new Log(AssociationManager.class);
    private static boolean isVerbose = false;

    /**
     * The implementation uses a map that maps object class affiliation to a second
     * hashTable that stores the associations.
     */
    private Map classMap = new Hashtable();
    private Class cachedClass = null;
    private Map cachedMap = null;

    /**
     * Gets the <code>objects</code> attribute of the <code>AssociationManager </code>
     * object.
     *
     * @param objects Description of parameter.
     */
    public void getObjects(Collection objects) {
        Iterator iterator = classMap.values().iterator();
        Map map;
        while (iterator.hasNext()) {
            map = (Map) iterator.next();
            objects.addAll(map.keySet());
        }
    }

    /**
     * Method description.
     *
     * @param keys Parameter description.
     * @param collection Parameter description.
     */
    public void getAssociations(Collection keys, Collection collection) {
        Iterator iterator = keys.iterator();
        while (iterator.hasNext()) {
            getAssociations(iterator.next(), collection);
        }
    }

    /**
     * Retrieves all association with the specified key.
     *
     * @param collection the found objects are accumulated in this collection.
     * @param key Description of parameter.
     */
    public void getAssociations(Object key, Collection collection) {
        Collection set = getAssociatedSet(key, false);
        if (set != null) {
            Iterator iterator = set.iterator();
            Object object;
            while (iterator.hasNext()) {
                object = iterator.next();
                if (!collection.contains(object)) {

                    // recursion
                    getAssociations(object, collection);
                    collection.add(object);
                }
            }
        }
    }

    /**
     * Maps an object onto it's class map.
     *
     * @param object the object that identifies the class map entry
     * @param create boolean flag indicating, whether a entry should be created, if it
     *        doesn't exits yet.
     *
     * @return <code>Map</code> - found or created <code>Map</code> , can be
     *         <code>null</code>
     */
    public Map getObjectMap(Object object, boolean create) {

        // :OPTIMIZED: by caching the last found map; this caching should produce
        //   good results at least for retrieving the maps for objects of the same
        //   class which is considered as a frequent case operating on a hierarchy.
        Class objectClass = object.getClass();

        // :FIXME: the caching doesn't work properly; testcase superposition
        //   (1adj->12asA); single protein 1kcw
        if (cachedClass != objectClass) {

            // try to get an existing key to association set map
            cachedMap = (Map) classMap.get(objectClass);
            cachedClass = objectClass;
        }
        if (cachedMap == null && create) {

            // add a new hash map to the class map
            cachedMap = new Hashtable();
            classMap.put(objectClass, cachedMap);
        }
        return cachedMap;
    }

    /**
     * Retrieves the <code>Collection</code> of objects that were associatied with
     * <code>object</code> .
     *
     * @param object object specifies the source of the associations.
     * @param create boolean flag indicating, whether a entry should be created, if it
     *        doesn't exits yet.
     *
     * @return <code>Collection</code> - the returned Collection contains all the objects
     *         associatied with <code>object</code> .
     */
    public Collection getAssociatedSet(Object object, boolean create) {
        Map objectMap = getObjectMap(object, create);

        // the object map might be null if create was false
        if (objectMap != null) {
            Collection associationsHash = (Collection) objectMap.get(object);
            if (create && associationsHash == null) {

                // add a new set if no set is existing for the current key
                // :NOTE: we use a hash set, because we don't want to accumulate registration
                associationsHash = new HashSet();
                objectMap.put(object, associationsHash);
            }
            return associationsHash;
        } else {
            return null;
        }
    }

    /**
     * Prints all entries of the <code>AssociationManager</code> to the out stream of the
     * system.
     */
    public void dump() {
        String string = new String();
        Iterator classIterator = classMap.values().iterator();
        Map objectMap;
        while (classIterator.hasNext()) {
            objectMap = (Map) classIterator.next();
            log.info("  " + objectMap);
        }
    }

    /**
     * Register an associations between key and object
     *
     * @param key the associate object will be used as a key.
     * @param object object to register.
     */
    public void register(Object key, Object object) {
        Collection associationsHash = getAssociatedSet(key, true);

        // store key to object association
        associationsHash.add(object);
    }

    /**
     * Removes all associations of the object from the list.
     *
     * @param object the object identifies the associations that should be removed.
     * @param collection Description of parameter.
     */
    public void remove(Object object, Collection collection) {
        Map objectMap = getObjectMap(object, false);
        if (objectMap != null) {

            // remove the whole set
            objectMap.remove(object);

            // iterate over all the associations and delete the other direction
            Iterator iterator = collection.iterator();
            Collection set;
            Object related;
            while (iterator.hasNext()) {
                related = iterator.next();
                set = getAssociatedSet(related, false);
                if (set != null) {
                    set.remove(object);
                    if (set.isEmpty()) {
                        objectMap.remove(related);
                    }
                }
            }
            if (objectMap.isEmpty()) {
                objectMap.clear();
            }
        }
    }

    /**
     * Description of the method.
     */
    public void clear() {
        Iterator classIterator = classMap.values().iterator();
        Iterator setIterator;
        Map objectMap;
        Collection collection;
        while (classIterator.hasNext()) {
            objectMap = (Map) classIterator.next();
            setIterator = objectMap.values().iterator();
            while (setIterator.hasNext()) {
                ((Collection) setIterator.next()).clear();
            }
            objectMap.clear();
        }
        classMap.clear();
        cachedClass = null;
        cachedMap = null;
    }

    /**
     * Description of the method.
     */
    void removeEmptyAssociationSets() {
        Iterator classIterator = classMap.values().iterator();
        Iterator setIterator;
        Map objectMap;
        Collection collection;
        while (classIterator.hasNext()) {
            objectMap = (Map) classIterator.next();
            setIterator = objectMap.values().iterator();
            while (setIterator.hasNext()) {
                collection = (Collection) setIterator.next();
                if (collection.isEmpty()) {
                    setIterator.remove();
                }
            }
            if (objectMap.isEmpty()) {
                classIterator.remove();
            }
        }
        cachedClass = null;
        cachedMap = null;
    }
}
