/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.objects;

import java.util.Hashtable;
import java.util.Map;

import org.srs3d.viewer.util.ExceptionHandler;

/**
 * Abstract <code>Attribute</code> class. This class implements a framework to save
 * instances over the whole system. The user is encouraged to instantiate concrete
 * Attribute classes using the static <code>getInstance() </code>method. Defined
 * customized subclasses of <code>Attribute</code> must adhere to some requirements to
 * assure effective instance managment. Note: the class implements the
 * ImmutableAttribute interface, which is actually a trick. Remove this and the default
 * implementation of getImmutable() after all attribute classes care for themselves.
 *
 * @author Karsten Klein
 * @version April 03, 2001
 */
public abstract class Attribute {

    /** This map registers the single instance attributes. */
    private static Map singleInstances = new Hashtable();

    /**
     * <code>Attribute</code> constructor. Unfortunaltey the constructor must be defined
     * public to assure customized objects outside this package can be used.
     */
    public Attribute() {
    }

    /**
     * The default implementation returns <code>true</code> in any case. The subclasses
     * with specific attributes MUST override this method. The reason for this default
     * behavior is that most attributes are considered to be singleInstances. Otherwise
     * we would have to implement a method returning <code>true</code> for every single
     * instance attribute, while the attributes with specific members have to implement
     * the method anyway.
     *
     * @param attribute Attribute to compare this instance to.
     *
     * @return Boolean value indicating equal content.
     */
    public abstract boolean isEqual(Attribute attribute);

    /**
     * Gets an immutable version of the attribute. This default implementation creates a
     * copy of the attribute and gives full access.
     *
     * @return The <code>immutable</code> value.
     */
    public abstract Attribute.Immutable getImmutable();

    /**
     * If an <code>Attribute</code> is found a flag <code>Attribute</code> that can be
     * shared by all objects that use it, the public constructor has to call
     * <code>registerSingleInstance()</code> method. The instance will be the only VM
     * instance of the object.
     */
    public void registerSingleInstance() {
        singleInstances.put(getClass(), this);
    }

    /**
     * This method is meant to copy the <code>Attribute</code> instance. The
     * implementation, whether it's a shallow or a deep copy depends on the implementor
     * of this method in the concrete subclass. Elements of the attribute are to be
     * changed must be deep copied.
     *
     * @return Copied instatnce of <code>this</code><code>Attribute</code> instance.
     */
    public Attribute copy() {
        return getInstance(this.getClass());
    }

    /**
     * Gets an instance of the concrete <code>Attribute</code> class. This instance can
     * be a system wide instance that is used as positive flag, a instance from a
     * special set of instances, or a specialized instance.
     *
     * @param attributeClass Class of the new <code>Attribute</code> instance.
     *
     * @return Appropriate instance of the class.
     */
    public static Attribute getInstance(Class attributeClass) {
        Attribute attribute = (Attribute) singleInstances.get(attributeClass);
        if (attribute == null) {
            try {
                attribute = (Attribute) attributeClass.newInstance();
            } catch (Exception e) {
                ExceptionHandler.handleException(e,
                    ExceptionHandler.SILENT_IN_RELEASE);
            }
        }
        return attribute;
    }

    /**
     * Method description.
     *
     * @param out Parameter description.
     *
     * @throws java.io.IOException Exception description.
     */
    public void stream(java.io.OutputStream out) throws java.io.IOException {
        out.write(("  " + getClass().toString() + "\n").getBytes());
    }

    /**
     * Description of the class.
     *
     * @author fries, LION bioscience AG
     * @version December 14, 2001
     */
    public class Immutable {

        /**
         * Gets the <code>copy</code> attribute of the <code>Immutable</code> object.
         *
         * @return The <code>copy</code> value.
         */
        public Attribute getCopy() {
            return Attribute.this.copy();
        }
    }
}
