/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.objects;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

/**
 * An ObjectContainer - is in principle a folder that- organizes objects. Every
 * application has an ObjectContainer instance at the top-level of the object-hierarchy.
 * This is essential for all kinds of methods executed on the hierarchy.
 *
 * @author Karsten Klein
 * @version March 20, 2001
 */
public class ObjectContainer extends AbstractObject {

    /** Note that we use a array list here to keep track of the object ordering */
    private Collection objects = new ArrayList();
    private String name = null;

    /**
     * Sets the <code>Name</code> attribute of the <code>ObjectContainer</code> object
     *
     * @param name The new <code>Name</code> value
     */
    public void setName(String name) {
        this.name = new String(name);
    }

    /**
     * Gets the <code>Iterator</code> attribute of the <code>ObjectContainer
     * </code>object
     *
     * @return The <code>Iterator</code> value
     */
    public Iterator getIterator() {
        return objects.iterator();
    }

    /**
     * Gets the <code>AllObjects</code> attribute of the <code>ObjectContainer
     * </code>object
     *
     * @param collection Description of parameter
     */
    public void getAllObjects(Collection collection) {
        if (objects != null) {
            collection.addAll(objects);
        }
    }

    /**
     * Gets the <code>Name</code> attribute of the <code>ObjectContainer</code> object
     *
     * @return The <code>Name</code> value
     */
    public String getName() {
        if (name != null) {
            return new String(name);
        } else {
            return new String();
        }
    }

    /**
     * Gets the <code>size</code> attribute of the <code>ObjectContainer</code> object.
     *
     * @return The <code>size</code> value.
     */
    public int getSize() {
        return objects.size();
    }

    /**
     * Gets the <code>Objects</code> attribute of the <code>ObjectContainer </code>object
     *
     * @return The <code>Objects</code> value
     */
    public Collection getObjects() {
        return objects;
    }

    /**
     * Adds a feature to the <code>Object</code> attribute of the
     * <code>ObjectContainer</code> object
     *
     * @param object The feature to be added to the <code>Object</code> attribute
     */
    public void addObject(AbstractObject object) {
        if (!objects.contains(object)) {
            objects.add(object);
        }
    }

    /**
     * Adds a <code>All</code> object to the <code>ObjectContainer</code> object.
     *
     * @param objects The <code>All</code> object to be added.
     */
    public void addAll(Collection objects) {
        this.objects.addAll(objects);
    }

    /**
     * Description of the method
     *
     * @param objectContainer Description of parameter.
     */
    public void merge(ObjectContainer objectContainer) {
        objects.addAll(objectContainer.getObjects());
    }

    /**
     * Description of the method
     */
    public void cleanup() {
        super.cleanup();
        cleanup(objects);
        objects = null;
    }

    /**
     * Description of the method
     */
    public void update() {
        Iterator iterator = getIterator();
        while (iterator.hasNext()) {
            ((AbstractObject) iterator.next()).update();
        }
    }

    /**
     * Description of the method
     *
     * @param object Description of parameter
     */
    public void removeObject(AbstractObject object) {
        objects.remove(object);
        object.cleanup();
    }

    /**
     * Description of the method
     *
     * @return Description of the returned value
     */
    public String toString() {
        if (getName().length() > 0) {
            return getName();
        }
        return "Object Container";
    }

    /**
     * Clears the object container. Note that no cleanup of the container objects and the
     * container object itself is done.
     */
    public void clear() {
        objects.clear();
    }
}
