/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.objects;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;

import org.srs3d.viewer.util.Chronometer;
import org.srs3d.viewer.util.Log;

/**
 * This <code>ObjectManager</code> class is capable to efficiently register and retrieve
 * associations in between <code>Object</code> s. Note that all kinds of associations
 * can be stored in between objects of different classes (subclasses of
 * <code>Object</code>). Associations in between instances of the same class are
 * restricted, because they are stored in a single hash table. Associations are
 * directional!
 *
 * @author Karsten Klein
 * @version February 20, 2001
 */
public final class ObjectManager {
    private static final Log log = new Log(ObjectManager.class);
    private static final boolean isVerbose = false;

    /** Stores all associations down the hierarchy */
    private AssociationManager downAssociationManager =
        new AssociationManager();

    /** Stores all associations up the hierarchy */
    private AssociationManager upAssociationManager = new AssociationManager();

    /**
     * Retrieves all associations specified by the key.
     *
     * @param key the object that specifies the association.
     * @param collection the found objects are accumulated in this collection.
     */
    public void getAssociations(Object key, Collection collection) {
        Chronometer c = Chronometer.getChronometer("ObjectManager");
        c.start();
        downAssociationManager.getAssociations(key, collection);
        upAssociationManager.getAssociations(key, collection);
        c.stop("getAssociations( Object, Collection )");
    }

    /**
     * Retrieves all associations specified by the key.
     *
     * @param key the object that specifies the association.
     *
     * @return <code>Collection</code> - the found objects are accumulated and returned
     *         in the a collection.
     */
    public Collection getAssociations(Object key) {
        Chronometer c = Chronometer.getChronometer("ObjectManager");
        c.start();
        Collection collection = new HashSet();
        getAssociations(key, collection);
        c.stop("getAssociations( Object )");
        return collection;
    }

    /**
     * Retrieves all downward associations specified by the key.
     *
     * @param key the object that specifies the association.
     * @param collection the found objects are accumulated in this collection.
     */
    public void getDownAssociations(Object key, Collection collection) {
        Chronometer c = Chronometer.getChronometer("ObjectManager");
        c.start();
        downAssociationManager.getAssociations(key, collection);
        c.stop("getDownAssociations( Object, Collection )");
    }

    /**
     * Gets the <code>downAssociations</code> attribute of the <code>ObjectManager</code>
     * object.
     *
     * @param keys Description of parameter.
     * @param collection Description of parameter.
     */
    public void getDownAssociations(Collection keys, Collection collection) {
        Chronometer c = Chronometer.getChronometer("ObjectManager");
        c.start();
        downAssociationManager.getAssociations(keys, collection);
        c.stop("getDownAssociations( Collection, Collection )");
    }

    /**
     * Method description.
     *
     * @param keys Parameter description.
     * @param collection Parameter description.
     */
    public void getUpAssociations(Collection keys, Collection collection) {
        Chronometer c = Chronometer.getChronometer("ObjectManager");
        c.start();
        upAssociationManager.getAssociations(keys, collection);
        c.stop("getUpAssociations( Collection, Collection )");
    }

    /**
     * Retrieves all upward associations specified by the key.
     *
     * @param collection the found objects are accumulated in this collection.
     * @param object Description of parameter.
     */
    public void getDirectUpAssociations(Object object, Collection collection) {
        Chronometer c = Chronometer.getChronometer("ObjectManager");
        c.start();
        Collection upSet = upAssociationManager.getAssociatedSet(object, false);
        if (upSet != null) {
            collection.addAll(upSet);
        }
        c.stop("getDirectUpAssociations( Object, Collection )");
    }

    /**
     * Gets the <code>directUpAssociations</code> attribute of the
     * <code>ObjectManager</code> object.
     *
     * @param keys Description of parameter.
     * @param collection Description of parameter.
     */
    public void getDirectUpAssociations(Collection keys, Collection collection) {
        Chronometer c = Chronometer.getChronometer("ObjectManager");
        c.start();
        Iterator iterator = keys.iterator();
        while (iterator.hasNext()) {
            getDirectUpAssociations(iterator.next(), collection);
        }
        c.stop("getAssociations( Collection, Collection )");
    }

    /**
     * Retrieves all direct downward associations specified by the key.
     *
     * @param collection the found objects are accumulated in this collection.
     * @param object Description of parameter.
     */
    public void getDirectDownAssociations(Object object, Collection collection) {
        Chronometer c = Chronometer.getChronometer("ObjectManager");
        c.start();
        Collection downSet =
            downAssociationManager.getAssociatedSet(object, false);
        if (downSet != null) {
            collection.addAll(downSet);
        }
        c.stop("getDirectDownAssociations( Object, Collection )");
    }

    /**
     * Gets the <code>directDownAssociations</code> attribute of the
     * <code>ObjectManager</code> object.
     *
     * @param keys Description of parameter.
     * @param collection Description of parameter.
     */
    public void getDirectDownAssociations(Collection keys, Collection collection) {
        Chronometer c = Chronometer.getChronometer("ObjectManager");
        c.start();
        Iterator iterator = keys.iterator();
        while (iterator.hasNext()) {
            getDirectDownAssociations(iterator.next(), collection);
        }
        c.stop("getDirectDownAssociations( Collection, Collection");
    }

    /**
     * Retrieves all direct upward associations specified by the key.
     *
     * @param key the object that specifies the association.
     * @param collection the found objects are accumulated in this collection.
     */
    public void getUpAssociations(Object key, Collection collection) {
        Chronometer c = Chronometer.getChronometer("ObjectManager");
        c.start();
        upAssociationManager.getAssociations(key, collection);
        c.stop("getUpAssociations( Object, Collection");
    }

    /**
     * Traverses the association upwards and downwards from the resulting set
     *
     * @param key source object.
     * @param collection destination to store the related objects in.
     */
    public void getAllRelated(Object key, Collection collection) {
        Chronometer c = Chronometer.getChronometer("ObjectManager");
        c.start();
        Collection upAssociations = new HashSet();
        getAssociations(key, upAssociations);
        upAssociations.add(key);
        Collection collapsed = new HashSet();
        collapseUp(upAssociations, collapsed);
        getDownAssociations(collapsed, collection);
        collection.addAll(collapsed);
        c.stop("getAllRelated( Object, Collection");
    }

    /**
     * Gets the <code>allRelated</code> attribute of the <code>ObjectManager </code>
     * object.
     *
     * @param keys Description of parameter.
     * @param collection Description of parameter.
     */
    public void getAllRelated(Collection keys, Collection collection) {
        Chronometer c = Chronometer.getChronometer("ObjectManager");
        c.start();
        Iterator iterator = keys.iterator();
        while (iterator.hasNext()) {
            getAllRelated(iterator.next(), collection);
        }
        c.stop("getAllRelated( Collection, Collection");
    }

    /**
     * Gets the <code>roots</code> attribute of the <code>ObjectManager</code> object.
     *
     * @param objects Description of parameter.
     * @param roots Description of parameter.
     */
    public void getRoots(Collection objects, Collection roots) {
        Collection set = new HashSet(objects.size());
        collapseUp(objects, set);
        if (set.size() == 1) {
            roots.addAll(set);
        } else {
            Iterator iterator = set.iterator();
            HashSet parents = new HashSet();
            Object object;

            // collect all parents
            while (iterator.hasNext()) {
                object = iterator.next();
                set = upAssociationManager.getAssociatedSet(object, false);
                if (set != null) {
                    parents.addAll(set);
                } else {
                    parents.add(object);
                }
            }
            if (!parents.isEmpty()) {
                if (parents.size() == 1) {
                    roots.addAll(parents);
                } else {
                    if (parents.containsAll(objects)) {
                        roots.addAll(parents);
                    } else {
                        getRoots(parents, roots);
                    }
                }
            }
        }
    }

    /**
     * Gets the <code>objects</code> attribute of the <code>ObjectManager</code> object.
     *
     * @param roots Description of parameter.
     */
    public void getObjects(Collection roots) {
        upAssociationManager.getObjects(roots);
        downAssociationManager.getObjects(roots);
    }

    /**
     * Registers an association between the two objects.
     *
     * @param parent the parental object.
     * @param child the child object.
     */
    public void register(Object parent, Object child) {
        if (child != null && parent != null) {
            upAssociationManager.register(child, parent);
            downAssociationManager.register(parent, child);
        } else {
            throw new RuntimeException(this + ": attempt to register " + child +
                " to " + parent + "detected!");
        }
    }

    /**
     * Registers a directional association between the two objects.
     *
     * @param parent the parental object.
     * @param child the child object.
     */
    public void registerUp(Object parent, Object child) {
        if (child != null && parent != null) {
            upAssociationManager.register(child, parent);
        } else {
            throw new RuntimeException(this + ": attempt to register " + child +
                " to " + parent + "detected!");
        }
    }

    /**
     * Registers a directional association between the two objects.
     *
     * @param parent the parental object.
     * @param child the child object.
     */
    public void registerDown(Object parent, Object child) {
        if (child != null && parent != null) {
            downAssociationManager.register(parent, child);
        } else {
            throw new RuntimeException(this + ": attempt to register " + child +
                " to " + parent + "detected!");
        }
    }

    /**
     * Removes all associations, where object is involved.
     *
     * @param object specifies the associations that are tobe removed.
     */
    public void remove(Object object) {
        Collection related = new ArrayList();
        Collection set;
        set = downAssociationManager.getAssociatedSet(object, false);
        if (set != null) {
            related.addAll(set);
        }
        set = upAssociationManager.getAssociatedSet(object, false);
        if (set != null) {
            related.addAll(set);
        }
        downAssociationManager.remove(object, related);
        upAssociationManager.remove(object, related);

        // :NOTE: this should be done explicitly, otherwise it will slow down
        //    the manager.
        // downAssociationManager.removeEmptyAssociationSets();
        // upAssociationManager.removeEmptyAssociationSets();
    }

    /**
     * Removes all associations, where object is involved.
     *
     * @param parent Description of parameter.
     * @param child Description of parameter.
     */
    public void remove(Object parent, Object child) {
        removeDown(parent, child);
        removeUp(parent, child);
    }

    /**
     * Description of the method.
     *
     * @param parent Description of parameter.
     * @param child Description of parameter.
     */
    public void removeUp(Object parent, Object child) {
        Collection related = new ArrayList();
        related.add(parent);
        upAssociationManager.remove(child, related);
    }

    /**
     * Description of the method.
     *
     * @param parent Description of parameter.
     * @param child Description of parameter.
     */
    public void removeDown(Object parent, Object child) {
        Collection related = new ArrayList();
        related.add(child);
        downAssociationManager.remove(parent, related);
    }

    /**
     * Prints the content of the manager to the systems out stream
     */
    public void dump() {
        log.info("ObjectManager dumping data: ");
        log.info("Upward linking data: ");
        upAssociationManager.dump();
        log.info("Downward linking data: ");
        downAssociationManager.dump();
    }

    /**
     * Removes all associations from the manager.
     */
    public void clear() {
        upAssociationManager.clear();
        downAssociationManager.clear();
    }

    /**
     * Description of the method.
     *
     * @param objects Description of parameter.
     * @param collapsed Description of parameter.
     */
    public void collapseUp(Collection objects, Collection collapsed) {
        Chronometer c = Chronometer.getChronometer("ObjectManager");
        c.start();
        if (objects.size() > 1) {
            Collection set = new HashSet(objects.size());
            collapsed.addAll(objects);
            getDownAssociations(objects, set);
            collapsed.removeAll(set);
        } else {
            collapsed.addAll(objects);
        }
        c.stop("collapseUp( Collection, Collection )");
    }

    /**
     * Method description.
     *
     * @param objects Parameter description.
     * @param collapsed Parameter description.
     */
    public void collapseUpExtended(Collection objects, Collection collapsed) {
        collapseUpExtended(objects, collapsed, true);
    }

    /**
     * Description of the method.
     *
     * @param objects Description of parameter.
     * @param collapsed Description of parameter.
     */
    public void collapseUpExtended(Collection objects, Collection collapsed,
        boolean collapseUp) {
        Chronometer c = Chronometer.getChronometer("ObjectManager");
        c.start();
        Collection set = new HashSet();
        Collection downSet;
        Collection roots = new HashSet();

        // get all objects that have down associated objects
        getDirectUpAssociations(objects, roots);

        // same as collapse up, but more efficient since we use the set collection
        // several times
        if (collapseUp && objects.size() > 1) {
            collapsed.addAll(objects);
            getDownAssociations(objects, set);
            collapsed.removeAll(set);
        } else {
            collapsed.addAll(objects);
        }
        Iterator iterator = roots.iterator();
        Object object;
        boolean modified = false;

        // collect all parents
        while (iterator.hasNext()) {
            object = iterator.next();
            downSet = downAssociationManager.getAssociatedSet(object, false);
            if (downSet != null) {
                if (collapsed.containsAll(downSet)) {
                    collapsed.removeAll(downSet);
                    if (!collapsed.contains(object)) {
                        collapsed.add(object);
                        modified = true;
                    }
                }
            }
        }

        // recurse on modifications
        if (modified) {
            set.clear();
            collapseUpExtended(collapsed, set, false);
            collapsed.clear();
            collapsed.addAll(set);
        }
        c.stop("collapseUpExtended( Collection, Collection ) )");
    }

    /**
     * Method description.
     *
     * @param objects Parameter description.
     * @param collapsed Parameter description.
     * @param collapseUp Parameter description.
     */
    public void collapseUpExtended_(Collection objects, Collection collapsed,
        boolean collapseUp) {
        Chronometer c = Chronometer.getChronometer("ObjectManager");
        c.start();
        Collection set = new HashSet(objects.size());
        Collection roots = new HashSet(objects.size());

        // get all objects that have down associated objects
        getDirectUpAssociations(objects, roots);

        // same as collapse up, but more efficient since we use the set collection
        // several times
        if (collapseUp && objects.size() > 1) {
            collapsed.addAll(objects);
            getDownAssociations(objects, set);
            collapsed.removeAll(set);
        } else {
            collapsed.addAll(objects);
        }
        Iterator iterator = roots.iterator();
        Object object;
        boolean modified = false;

        // collect all parents
        while (iterator.hasNext()) {
            object = iterator.next();
            set.clear();
            getDirectDownAssociations(object, set);
            if (collapsed.containsAll(set)) {
                collapsed.removeAll(set);
                if (!collapsed.contains(object)) {
                    collapsed.add(object);
                    modified = true;
                }
            }
        }

        // recurse on modifications
        if (modified) {
            set.clear();
            collapseUpExtended(collapsed, set, false);
            collapsed.clear();
            collapsed.addAll(set);
        }
        c.stop("collapseUpExtended( Collection, Collection ) )");
    }

    /**
     * Filters out the leafs of the in the objects subhierarchy.
     *
     * @param objects Description of parameter.
     * @param collapsed Description of parameter.
     */
    public void collapseDown(Collection objects, Collection collapsed) {
        Chronometer c = Chronometer.getChronometer("ObjectManager");
        c.start();
        Collection set = new HashSet(objects.size());
        collapsed.addAll(objects);
        getUpAssociations(objects, set);
        collapsed.removeAll(set);
        c.stop("collapseDown( Collection, Collection");
    }

    /**
     * Extracts all elements from the given collection that are not castable to the
     * specified class. Note that castable is a weaker condition than instance of.
     *
     * @param collection the collection to process.
     * @param extractClass class specifying the extraction.
     */
    public static void extract(Collection collection, Class extractClass) {
        Chronometer c = Chronometer.getChronometer("ObjectManager");
        c.start();
        Iterator iterator = collection.iterator();
        Class objectClass;
        while (iterator.hasNext()) {
            objectClass = iterator.next().getClass();
            if (!extractClass.isAssignableFrom(objectClass)) {
                iterator.remove();
            }
        }
        c.stop("extract( Collection, Class )");
    }

    /**
     * Extracts all elements from the given collection that are not castable to the
     * specified class. Note that castable is a weaker condition than instance of.
     *
     * @param collection the collection to process.
     * @param classes Description of parameter.
     */
    public static void extract(Collection collection, Collection classes) {
        Chronometer c = Chronometer.getChronometer("ObjectManager");
        c.start();
        Collection extracted = new ArrayList();
        Collection tmp = new ArrayList();
        Iterator iterator = classes.iterator();
        Class objectClass;
        while (iterator.hasNext()) {
            objectClass = (Class) iterator.next();
            tmp.clear();
            tmp.addAll(collection);
            extract(tmp, objectClass);
            extracted.addAll(tmp);
        }
        collection.clear();
        collection.addAll(extracted);
        c.stop("extract( Collection, Collection )");
    }

    /**
     * Method description.
     *
     * @param collection Parameter description.
     * @param classes Parameter description.
     */
    public static void filter(Collection collection, Collection classes) {
        Chronometer c = Chronometer.getChronometer("ObjectManager");
        c.start();
        Iterator iterator = collection.iterator();
        Object object;
        while (iterator.hasNext()) {
            object = iterator.next();
            if (classes.contains(object.getClass())) {
                iterator.remove();
            }
        }
        c.stop("filter( Collection, Collection");
    }

    /**
     * Filters all elements from the given collection that are castable to the specified
     * class. Note that castable is a weaker condition than instance of.
     *
     * @param collection the collection to process.
     * @param filterClass class specifying the filter.
     */
    public static void filter(Collection collection, Class filterClass) {
        Chronometer c = Chronometer.getChronometer("ObjectManager");
        c.start();
        Iterator iterator = collection.iterator();
        Class objectClass;
        while (iterator.hasNext()) {
            objectClass = iterator.next().getClass();
            if (objectClass == filterClass) {
                iterator.remove();
            } else {
                if (filterClass.isAssignableFrom(objectClass)) {
                    iterator.remove();
                }
            }
        }
        c.stop("filter( Collection, Class");
    }

    /**
     * Description of the method.
     *
     * @param collection Description of parameter.
     * @param filtered Description of parameter.
     * @param filterClass Description of parameter.
     */
    public static void extract(Collection collection, Collection filtered,
        Class filterClass) {
        Chronometer c = Chronometer.getChronometer("ObjectManager");
        c.start();
        Iterator iterator = collection.iterator();
        Object object;
        Class objectClass;
        while (iterator.hasNext()) {
            object = iterator.next();
            objectClass = object.getClass();
            if (objectClass == filterClass) {
                filtered.add(object);
            } else if (filterClass.isAssignableFrom(objectClass)) {
                filtered.add(object);
            }
        }
        c.stop("extract( Collection, Collection, filterClass )");
    }

    /**
     * Method description.
     *
     * @param setA Parameter description.
     * @param setB Parameter description.
     * @param intersection Parameter description.
     */
    public static void exclude(Collection setA, Collection setB,
        Collection intersection) {
        Chronometer c = Chronometer.getChronometer("ObjectManager");
        c.start();
        Iterator iterator = setA.iterator();
        Object object;
        while (iterator.hasNext()) {
            object = iterator.next();
            if (!setB.contains(object)) {
                intersection.add(object);
            }
        }
        c.stop("exclude( Collection, Collection, Collection )");
    }

    /**
     * Method description.
     *
     * @param keys Parameter description.
     * @param collection Parameter description.
     */
    public void getAssociations(Collection keys, Collection collection) {
        Chronometer c = Chronometer.getChronometer("ObjectManager");
        c.start();
        if (isVerbose) {
            log.debug("retrieving associations for " + keys + "(" +
                keys.getClass() + ")");
        }
        downAssociationManager.getAssociations(keys, collection);
        upAssociationManager.getAssociations(keys, collection);
        c.stop("getAssociations( Collection, Collection )");
    }
}
