/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.objects;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.StringTokenizer;
import java.util.Vector;

import org.srs3d.viewer.bioatlas.operations.CreateAnnotationOperation;
import org.srs3d.viewer.j3d.Context;
import org.srs3d.viewer.j3d.ContextData;
import org.srs3d.viewer.j3d.operations.TestOperation;
import org.srs3d.viewer.objects.visitors.ObjectIdentifier;
import org.srs3d.viewer.util.Log;
import org.srs3d.viewer.util.XmlTag;

/**
 * @author Karsten Klein
 * @version May 28, 2001
 */
public class Operation {
    private static final Log log = new Log(Operation.class);
    private AbstractObject object = null;
    private Collection objects = null;
    private String id = null;
    private ContextData contextData = null;
    private boolean isSerializeable = true;
    private boolean consumed = false;

    /**
     * <code>Operation</code> constructor.
     *
     * @param context Description of parameter.
     * @param id Description of parameter.
     * @param object Description of parameter.
     */
    public Operation(Context context, String id, AbstractObject object) {
        setContext(context);
        setId(id);
        setObject(object);
    }

    /**
     * Sets the <code>object</code> attribute of the <code>Operation</code> object.
     *
     * @param object The new <code>object</code> value.
     */
    public void setObject(AbstractObject object) {
        this.object = object;
    }

    /**
     * Sets the <code>objects</code> attribute of the <code>Operation</code> object.
     *
     * @param objects The new <code>objects</code> value.
     */
    public void setObjects(Collection objects) {
        this.objects = objects;
    }

    /**
     * Sets the <code>id</code> attribute of the <code>Operation</code> object.
     *
     * @param id The new <code>id</code> value.
     */
    public void setId(String id) {
        this.id = id;
    }

    /**
     * Sets the <code>context</code> attribute of the <code>Operation</code> object.
     *
     * @param context The new <code>context</code> value.
     */
    public void setContext(Context context) {
        if (context != null) {
            this.contextData = context.getContextData();
        } else {
            this.contextData = null;
        }
    }

    /**
     * Sets the <code>consumed</code> attribute of the <code>Operation</code> object.
     *
     * @param consumed The new <code>consumed</code> value.
     */
    public void setConsumed(boolean consumed) {
        this.consumed = consumed;
    }

    /**
     * Sets the <code>serializable</code> attribute of the <code>Operation</code> object.
     *
     * @param isSerializeable The new <code>serializable</code> value.
     */
    public void setSerializable(boolean isSerializeable) {
        this.isSerializeable = isSerializeable;
    }

    /**
     * Gets the <code>object</code> attribute of the <code>Operation</code> object.
     *
     * @return The <code>object</code> value.
     */
    public AbstractObject getObject() {
        return object;
    }

    /**
     * Gets the <code>objects</code> attribute of the <code>Operation</code> object.
     *
     * @return The <code>objects</code> value.
     */
    public Collection getObjects() {
        return objects;
    }

    /**
     * Gets the <code>id</code> attribute of the <code>Operation</code> object.
     *
     * @return The <code>id</code> value.
     */
    public String getId() {
        return id;
    }

    /**
     * Gets the <code>context</code> attribute of the <code>Operation</code> object.
     *
     * @return The <code>context</code> value.
     */
    public Context getContext() {
        return contextData.getContext();
    }

    /**
     * Gets the <code>consumed</code> attribute of the <code>Operation</code> object.
     *
     * @return The <code>consumed</code> value.
     */
    public boolean isConsumed() {
        return consumed;
    }

    /**
     * Description of the method.
     *
     * @param id Description of parameter.
     *
     * @return Description of the returned value.
     */
    public boolean is(String id) {
        return this.id.equalsIgnoreCase(id);
    }

    /**
     * Gets the <code>serializeable</code> attribute of the <code>Operation</code>
     * object.
     *
     * @return The <code>serializeable</code> value.
     */
    public boolean isSerializeable() {
        return isSerializeable;
    }

    /**
     * Description of the method.
     *
     * @return Description of the returned value.
     */
    public String toString() {
        return id + "(on object " + object + ")";
    }

    /**
     * Method description.
     *
     * @param contextData Parameter description.
     *
     * @return Return description.
     */
    public XmlTag serialize(ContextData contextData) {
        XmlTag xmlTag = new XmlTag("operation");

        // set operation id
        xmlTag.setAttribute("id", getId());
        boolean isSuccess = serializeAttributes(contextData, xmlTag);
        if (isSuccess) {
            return xmlTag;
        }
        return null;
    }

    /**
     * Gets the <code>idList</code> attribute of the <code>Operation</code> class.
     *
     * @param contextData Description of parameter.
     * @param collection Description of parameter.
     *
     * @return The <code>idList</code> value.
     */
    public static String getIdList(ContextData contextData,
        Collection collection) {
        String idList = "";
        Iterator iterator = collection.iterator();
        ObjectIdentifier objectIdentifier = new ObjectIdentifier(contextData);
        AbstractObject object;
        while (iterator.hasNext()) {
            object = (AbstractObject) iterator.next();
            objectIdentifier.setObject(object);
            objectIdentifier.setObjectId(null);
            objectIdentifier.setStrategy(ObjectIdentifier.ASSOCIATION_STRATEGY);
            objectIdentifier.visit(contextData.getObjectContainer().getObjects());
            if (objectIdentifier.getObjectId() == null) {
                objectIdentifier.setStrategy(ObjectIdentifier.LOCALIZATION_STRATEGY);
                objectIdentifier.visit(contextData.getObjectContainer()
                                                  .getObjects());
            }
            if (objectIdentifier.getObjectId() != null) {
                if (idList.length() > 0) {
                    idList += ", ";
                }
                idList += objectIdentifier.getObjectId();
            } else {
                log.error("cannot identify object " + object);

                // the method failes if at least one object can not be identified
                return null;
            }
        }
        return idList;
    }

    /**
     * Description of the method.
     *
     * @param template Description of parameter.
     * @param objects Description of parameter.
     *
     * @return Description of the returned value.
     */
    public static Collection createOperations(Operation template,
        Collection objects) {
        Vector vector = new Vector();
        Iterator iterator = objects.iterator();
        Operation operation;
        while (iterator.hasNext()) {
            operation =
                new Operation(template.getContext(), template.getId(), null);
            operation.setObject((AbstractObject) iterator.next());
            vector.add(operation);
        }
        return vector;
    }

    /**
     * Method description.
     *
     * @param contextData Parameter description.
     * @param tag Parameter description.
     *
     * @return Return description.
     */
    public static Collection deserialize(ContextData contextData, XmlTag tag) {
        boolean isSuccess = true;
        Collection operations = new ArrayList(5);
        if (tag.getName().equalsIgnoreCase("operation")) {
            Operation operation;
            String value;
            operation =
                new Operation(contextData.getContext(),
                    "UPDATE-" + tag.getAttribute("id"), null);
            operation.setSerializable(false);
            operations.add(operation);
            String subId = tag.getAttribute("subId");
            if (subId != null) {

                // :FIXME: general mechanism needed
                if (subId.equalsIgnoreCase("CreateAnnotation")) {
                    operation =
                        new CreateAnnotationOperation(contextData.getContext(),
                            tag.getAttribute("id"), null, null, null);
                } else if (subId.equalsIgnoreCase("Test")) {
                    operation =
                        new TestOperation(contextData.getContext(),
                            tag.getAttribute("id"));
                } else {
                    log.error("unknown sub id " + subId);
                    operation =
                        new Operation(contextData.getContext(),
                            tag.getAttribute("id"), null);
                }
            } else {
                operation =
                    new Operation(contextData.getContext(),
                        tag.getAttribute("id"), null);
            }
            isSuccess = operation.deserializeAttributes(contextData, tag);
            if (isSuccess) {
                operations.add(operation);
            } else {
                return null;
            }
        }
        if (isSuccess) {
            return operations;
        }
        return null;
    }

    /**
     * Converts the specified string to objects. In the string the objects identifiers
     * are separated by commas.
     *
     * @param contextData Description of parameter.
     * @param idList Description of parameter.
     * @param objects Description of parameter.
     *
     * @return Description of the returned value.
     */
    private static boolean identifyObjects(ContextData contextData,
        String idList, Collection objects) {
        StringTokenizer tokenizer = new StringTokenizer(idList, ",");
        ObjectIdentifier objectIdentifier = new ObjectIdentifier(contextData);
        objectIdentifier.setMode(ObjectIdentifier.ID_TO_OBJECT);
        while (tokenizer.hasMoreTokens()) {
            objectIdentifier.setObjectId(tokenizer.nextToken().trim());
            objectIdentifier.setObject(null);
            objectIdentifier.visit(contextData.getObjectContainer().getObjects());
            if (objectIdentifier.getObject() != null) {
                objects.add(objectIdentifier.getObject());
            } else {
                log.error("cannot identify object " +
                    objectIdentifier.getObjectId());
                return false;
            }
        }
        return true;
    }

    /**
     * Method description.
     *
     * @param contextData Parameter description.
     * @param xmlTag Parameter description.
     *
     * @return Return description.
     */
    public boolean deserializeAttributes(ContextData contextData, XmlTag xmlTag) {
        boolean isSuccess = true;
        String value = xmlTag.getAttribute("object");
        if (value != null) {
            ArrayList objects = new ArrayList();
            if (identifyObjects(contextData, value, objects)) {
                if (!objects.isEmpty()) {
                    setObject((AbstractObject) objects.get(0));
                }
            } else {
                isSuccess = false;
            }
        }
        if (isSuccess) {
            XmlTag objectsTag = xmlTag.getTag("objects");
            if (objectsTag != null) {
                value = objectsTag.getValue();
                if (value != null) {
                    ArrayList objects = new ArrayList();
                    if (identifyObjects(contextData, value, objects)) {
                        setObjects(objects);
                    } else {
                        isSuccess = false;
                    }
                }
            }
        }
        return isSuccess;
    }

    /**
     * Method description.
     *
     * @param contextData Parameter description.
     * @param xmlTag Parameter description.
     *
     * @return Return description.
     */
    public boolean serializeAttributes(ContextData contextData, XmlTag xmlTag) {
        String idList;
        boolean isSuccess = true;
        if (getObject() != null) {
            Collection objects = new ArrayList(1);
            objects.add(getObject());
            idList = getIdList(contextData, objects);
            if (idList != null) {
                xmlTag.setAttribute("object", idList);
            } else {
                isSuccess = false;
            }
        }
        if (isSuccess && getObjects() != null && !getObjects().isEmpty()) {
            idList = getIdList(contextData, getObjects());
            if (idList != null) {
                XmlTag objectsTag = new XmlTag("objects");
                objectsTag.setValue(idList);
                xmlTag.getSubtags().add(objectsTag);
            } else {
                isSuccess = false;
            }
        }
        return isSuccess;
    }
}
