/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.objects;

import java.util.Hashtable;
import java.util.Iterator;

/**
 * The <code>ObjectState</code> stores <code>Attribute</code> instances. Note that it
 * only can keep one kind of each <code>Attribute</code> at a time. The motivation in
 * this kind of storing attributes for objects is that of memory consumption. Most of
 * the objects don't have certain attributes and therefore shouldn't even allocate a
 * link that is <code>null</code> most of the time or through the whole livetime of the
 * object. We assume a default <code>State </code>that can be inquired through the
 * <code>StateManager</code> class. Moreover, this class provides a very lean interface
 * to access all attributes of the object. If you want to store two objects of the same
 * <code>Attribute </code>class you have to implement a new <code>Attribute</code> class
 * that encapsulates one or both of the objects.
 *
 * @author Karsten Klein
 * @version April 03, 2001
 *
 * @see org.srs3d.viewer.object.Attribute
 */
public final class State extends Hashtable {
    public static int copyCounter = 0;

    /**
     * Sets the <code>Attribute</code> of the <code>State</code> object.
     *
     * @param attribute The new <code>Attribute</code> value.
     */
    public void setAttribute(Attribute attribute) {
        put(attribute.getClass(), attribute);
    }

    /**
     * Gets a <code>Attribute</code> of the <code>State</code> object.
     *
     * @param attributeClass Attribute class.
     *
     * @return The <code>Attribute</code> instance if available. <code>null</code> if
     *         not.
     */
    public Attribute getAttribute(Class attributeClass) {
        return (Attribute) get(attributeClass);
    }

    /**
     * Description of the method.
     *
     * @param attributeClass Specifies the attribute to remove.
     */
    public void removeAttribute(Class attributeClass) {
        remove(attributeClass);
    }

    /**
     * Checks the presence of the specified <code>Attribute</code>.
     *
     * @param attributeClass Class specifiying the attribute type.
     *
     * @return Boolean that indicates the presence of the attribute.
     */
    public boolean hasAttribute(Class attributeClass) {
        return containsKey(attributeClass);
    }

    /**
     * Checks whether the state has <code>Attributes</code> at all.
     *
     * @return Boolean value indicating the existence of attributes.
     */
    public boolean hasAttributes() {
        return !isEmpty();
    }

    /**
     * Clears all <code>Attribute</code> instances from the state.
     */
    public void clear() {
        super.clear();
    }

    /**
     * Copies the state object. Every attribute of the instance will be copied to the
     * returned <code>State</code> instance.
     *
     * @return Copied state instance.
     */
    public State copy() {
        copyCounter++;
        State state = new State();

        // copy all the attributes
        Iterator iterator = values().iterator();
        Attribute attribute;
        while (iterator.hasNext()) {
            attribute = (Attribute) iterator.next();

            // use the attributes copy method
            state.setAttribute(attribute.copy());
        }
        return state;
    }

    /**
     * Method description.
     *
     * @param state Parameter description.
     *
     * @return Return description.
     */
    public boolean isEqual(State state) {
        if (this == state) {
            return true;
        } else {
            if (size() == state.size()) {

                // check the attributes
                Iterator iterator = values().iterator();
                Attribute attribute;
                while (iterator.hasNext()) {
                    attribute = (Attribute) iterator.next();
                    Attribute stateAttribute =
                        state.getAttribute(attribute.getClass());
                    if (stateAttribute == null) {
                        return false;
                    } else {
                        if (!attribute.isEqual(stateAttribute)) {
                            return false;
                        }
                    }
                }
                return true;
            }
        }
        return false;
    }

    /**
     * Method description.
     *
     * @return Return description.
     */
    public String toString() {
        return values().toString();
    }

    /**
     * Method description.
     *
     * @return Return description.
     */
    public Immutable getImmutable() {
        return new Immutable();
    }

    /**
     * Class description.
     *
     * @author Karsten Klein
     */
    public class Immutable {

        /**
         * Method description.
         *
         * @param objectClass Parameter description.
         *
         * @return Return description.
         */
        public Attribute.Immutable getAttribute(Class objectClass) {
            Attribute attribute = State.this.getAttribute(objectClass);
            if (attribute != null) {
                return State.this.getAttribute(objectClass).getImmutable();
            }
            return null;
        }

        /**
         * Method description.
         *
         * @param objectClass Parameter description.
         *
         * @return Return description.
         */
        public boolean hasAttribute(Class objectClass) {
            return State.this.hasAttribute(objectClass);
        }

        /**
         * Method description.
         *
         * @return Return description.
         */
        public State getCopy() {
            return State.this.copy();
        }

        /**
         * Method description.
         *
         * @param out Parameter description.
         * @param objectId Parameter description.
         *
         * @throws java.io.IOException Exception description.
         */
        public void stream(java.io.OutputStream out, String objectId)
            throws java.io.IOException {
            out.write(("<state> object=\"" + objectId + "\"\n").getBytes());
            java.util.Iterator iterator = values().iterator();
            while (iterator.hasNext()) {
                ((Attribute) iterator.next()).stream(out);
            }
            out.write("</state>\n".getBytes());
        }
    }
}
