/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.objects;

import java.util.Collection;
import java.util.Hashtable;
import java.util.Map;

import org.srs3d.viewer.j3d.ContextData;

/**
 * The <code>StateManager</code> is used to store <code>AbstractObject</code> to
 * <code>State</code> associations in an very effective way. As long as the state
 * instance doesn't differ from the default state we simple do NOT store the objects
 * state. The default state is inquired using the <code>StatePrototypeManager</code> .
 * In the case that no prototype is available an empty <code>State</code> instance is
 * assumed as default state. Note that mapping back to the default states may consume
 * more time at a first glance, but we can really benefit from a dramatically reduced
 * size the prototype state and the non-default state sets.
 *
 * @author Karsten Klein
 * @version April 03, 2001
 */
public class StateManager extends AbstractManager {

    /** Maps the objects to their <code>State</code> instances. */
    private Map stateMap = new Hashtable();

    /**
     * <code>StateManager</code> constructor.
     *
     * @param contextData Description of parameter.
     */
    public StateManager(ContextData contextData) {
        super(contextData);
    }

    /**
     * Gets the <code>State</code> instance associated with the specified object. If the
     * object is in the default state and thus has no <code>State</code> a empty state
     * is created.
     *
     * @param object Object specifying the state association.
     *
     * @return <code>State</code> instance of the object.
     */
    public State getState(AbstractObject object) {
        State state = (State) stateMap.get(object);
        if (state == null) {

            // get a state prototype if available, no need to copy!
            state =
                getContextData().getStatePrototypeManager().getState(object);

            // nothing helped
            if (state == null) {

                // create a new empty state
                state = new State();
            }
        } else {

            // testing
            //      countCopy( object );
            state = state.copy();
        }
        return state;
    }

    /**
     * Method description.
     *
     * @param object Parameter description.
     *
     * @return Return description.
     */
    public State.Immutable getImmutableState(AbstractObject object) {
        State state = (State) stateMap.get(object);
        if (state == null) {
            State.Immutable immutableState;

            // get a state prototype if available, no need to copy!
            immutableState =
                getContextData().getStatePrototypeManager().getImmutableState(object);

            // nothing helped
            if (immutableState == null) {

                // create a new empty state
                state = new State();
            } else {
                return immutableState;
            }
        }
        return state.getImmutable();
    }

    /**
     * Gets the <code>Objects</code> attribute of the <code>StateManager</code> object.
     *
     * @return The <code>Objects</code> value.
     */
    public Collection getObjects() {
        return stateMap.keySet();
    }

    /**
     * Registers a new object with a state, or registers a state change of the object.
     *
     * @param object Object instance.
     * @param state State instance to associate with the specified object.
     */
    public void register(AbstractObject object, State state) {

        // states that match the shared default state are not registered and will
        // be removed
        if (!isDefaultState(object, state)) {
            stateMap.put(object, state);
        } else {
            stateMap.remove(object);
        }
    }

    /**
     * Description of the method.
     */
    public void clear() {
        stateMap.clear();
    }

    /**
     * Checks whether the specified state matches the default state of the given object.
     *
     * @param object Object the state should be checked for.
     * @param state <code>State</code> instance to check.
     *
     * @return The <code>DefaultState</code> value.
     */
    private boolean isDefaultState(AbstractObject object, State state) {
        boolean flag = false;
        State defaultState =
            getContextData().getStatePrototypeManager().getState(object);

        // if there is no prototype state, the default state is empty
        if (defaultState == null) {

            // if the proposed state doesn't contain any data, it's considered as default.
            // it will be properly recreated on demand.
            if (!state.hasAttributes()) {
                flag = true;
            }
        } else {
            flag = defaultState.isEqual(state);
        }
        return flag;
    }

    /**
     * Method description.
     *
     * @param object Parameter description.
     */
    public void remove(AbstractObject object) {
        stateMap.remove(object);
    }
}
