/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.objects;

import java.util.ArrayList;
import java.util.Collection;
import java.util.ConcurrentModificationException;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;

import org.srs3d.viewer.j3d.ContextData;
import org.srs3d.viewer.j3d.attributes.Expanded;
import org.srs3d.viewer.util.ExceptionHandler;

/**
 * This <code>StrategyManager</code> class is an associative registry of
 * <code>AbstractObject</code> classes to default <code>State</code> objects.
 *
 * @author Karsten Klein
 * @version April 05, 2001
 */
public class StrategyManager extends AbstractManager implements Computation {

    /** Association store the classes stategy objects. */
    private Map strategyMap = new Hashtable();
    private transient Strategy cachedStrategy = null;
    private transient Class cachedObjectClass = null;
    private int hit = 0;
    private int miss = 0;
    private transient String description = "";
    private transient float progress = 0;
    private transient boolean canCancel = true;
    private transient boolean isCanceled = false;

    /**
     * <code>StrategyManager</code> constructor.
     *
     * @param contextData Description of parameter.
     */
    public StrategyManager(ContextData contextData) {
        super(contextData);
    }

    /**
     * Returns a <code>State</code> prototype if available (register first!!).
     *
     * @param object Object to retrieve the state for.
     *
     * @return The new <code>State</code> instance, or <code>null</code> .
     */
    public Strategy getStrategy(AbstractObject object) {
        Strategy strategy;
        if (object.getClass() == cachedObjectClass) {
            strategy = cachedStrategy;
            hit++;
        } else {
            strategy = (Strategy) strategyMap.get(object.getClass());
            cachedObjectClass = object.getClass();
            miss++;
        }
        if (strategy == null) {

            // can be replaced by a default strategy using StrategyProtoypeFactory
            strategy =
                new org.srs3d.viewer.j3d.objects.strategies.AbstractStrategy();
        }
        cachedStrategy = strategy;
        return strategy;
    }

    /**
     * Registers the specified class with a prototype <code>State</code> .
     *
     * @param objectClass Object class.
     * @param strategy Description of parameter.
     */
    public void register(Class objectClass, Strategy strategy) {
        if (strategy != null) {
            strategyMap.put(objectClass, strategy);
        }
    }

    /**
     * Description of the method.
     */
    public void clear() {
        strategyMap.clear();
    }

    /**
     * Description of the method.
     *
     * @param object Description of parameter.
     * @param command Description of parameter.
     */
    public final void propagate(AbstractObject object, Command command) {

        // store parent
        AbstractObject parent = command.getParent();
        execute(object, command);
        StateManager stateManager = getContextData().getStateManager();
        State.Immutable state = stateManager.getImmutableState(object);
        if (state.hasAttribute(Expanded.class)) {
            Collection objects = new ArrayList();
            object.getAllObjects(objects);
            Iterator iterator = objects.iterator();
            AbstractObject child;
            while (iterator.hasNext()) {
                child = (AbstractObject) iterator.next();
                state = stateManager.getImmutableState(child);
                command.setParent(object);
                if (state.hasAttribute(Expanded.class)) {
                    propagate(child, command);
                } else {
                    execute(child, command);
                }
            }
        }

        // restore old parent
        command.setParent(parent);
    }

    /**
     * Description of the method.
     *
     * @param objects Description of parameter.
     * @param command Description of parameter.
     * @param objectClass Description of parameter.
     */
    public final void propagate(Collection objects, Command command,
        Class objectClass) {
        Iterator iterator = objects.iterator();
        float progress = 0;
        float deltaProgress = (float) objects.size() / 100;
        while (iterator.hasNext()) {
            this.progress = (int) progress;
            progress += deltaProgress;
            propagate((AbstractObject) iterator.next(), command, objectClass);
        }
        this.progress = 0;
    }

    /**
     * Description of the method.
     *
     * @param object Description of parameter.
     * @param command Description of parameter.
     * @param objectClass Description of parameter.
     */
    public final void propagate(AbstractObject object, Command command,
        Class objectClass) {

        // store parent
        AbstractObject parent = command.getParent();
        execute(object, command);
        if (objectClass == null ||
              !objectClass.isAssignableFrom(object.getClass())) {
            Collection objects = new ArrayList();
            object.getAllObjects(objects);
            Iterator iterator = objects.iterator();
            AbstractObject child;
            while (iterator.hasNext()) {
                child = (AbstractObject) iterator.next();
                command.setParent(object);
                if (objectClass == null ||
                      !objectClass.isAssignableFrom(child.getClass())) {
                    propagate(child, command, objectClass);
                } else {
                    execute(child, command);
                }
            }
        }

        // restore old parent
        command.setParent(parent);
    }

    /**
     * Description of the method.
     *
     * @param objects Description of parameter.
     * @param command Description of parameter.
     */
    public final void propagate(Collection objects, Command command) {
        propagate(objects, command, true);
    }

    /**
     * Description of the method.
     *
     * @param objects Description of parameter.
     * @param command Description of parameter.
     * @param collapseUp Description of parameter.
     */
    public final void propagate(Collection objects, Command command,
        boolean collapseUp) {
        Collection set;
        if (collapseUp) {
            set = new HashSet();
            getContextData().getObjectManager().collapseUp(objects, set);
        } else {
            set = objects;
        }
        Iterator iterator = set.iterator();
        AbstractObject object;
        float progress = 0;
        float deltaProgress = (float) objects.size() / 100;
        while (iterator.hasNext()) {
            this.progress = (int) progress;
            progress += deltaProgress;
            object = (AbstractObject) iterator.next();
            command.setObject(object);
            propagate(object, command);
        }
        this.progress = 0;
    }

    /**
     * Executes a commands on the specified object.
     *
     * @param object Target object.
     * @param command Command to execute.
     */
    public final void execute(AbstractObject object, Command command) {
        if (!isCanceled) {
            description = "Executing on " + object;
            Strategy strategy = getStrategy(object);
            command.setObject(object);
            try {
                strategy.execute(command);
            } catch (Exception e) {
                ExceptionHandler.handleException(e,
                    ExceptionHandler.SILENT_IN_DEBUG, this);
            }
        }
    }

    /**
     * Executes a series of commands on the specified object.
     *
     * @param object Target object.
     * @param commands Commands to execute. The collection type specifies the order of
     *        the commands.
     */
    public final void execute(AbstractObject object, Collection commands) {
        Strategy strategy = getStrategy(object);
        Iterator iterator = commands.iterator();
        Command command;
        while (iterator.hasNext()) {
            command = (Command) iterator.next();
            command.setObject(object);
            strategy.execute(command);
        }
    }

    /**
     * Method description.
     *
     * @param objects Parameter description.
     * @param commands Parameter description.
     */
    public final void execute(Collection objects, Collection commands) {
        Iterator iterator = objects.iterator();
        float deltaProgress = objects.size() / 100;
        float progress = 0;
        while (iterator.hasNext()) {
            this.progress = (int) progress;
            progress += deltaProgress;
            execute((AbstractObject) iterator.next(), commands);
        }
        this.progress = 0;
    }

    /**
     * Description of the method.
     *
     * @param objects Description of parameter.
     * @param command Description of parameter.
     */
    public void execute(Collection objects, Command command) {
        Iterator iterator = objects.iterator();
        AbstractObject object;
        float deltaProgress = objects.size() / 100;
        float progress = 0;
        while (iterator.hasNext()) {
            this.progress = (int) progress;
            try {
                object = (AbstractObject) iterator.next();
                execute(object, command);
            } catch (ConcurrentModificationException e) {
                ExceptionHandler.handleException(e,
                    ExceptionHandler.SILENT_IN_DEBUG, this);
                iterator = objects.iterator();
            }
            progress += deltaProgress;
        }
        this.progress = 0;
    }

    /**
     * Description of the method.
     *
     * @param object Description of parameter.
     * @param objects Description of parameter.
     * @param stateManager Description of parameter.
     */
    private static final void collectAllObjects(AbstractObject object,
        Vector objects, StateManager stateManager) {
        State.Immutable state = stateManager.getImmutableState(object);
        if (state.hasAttribute(Expanded.class)) {
            Collection set = new ArrayList();
            object.getAllObjects(set);
            Iterator iterator = set.iterator();
            while (iterator.hasNext()) {
                collectAllObjects((AbstractObject) iterator.next(), objects,
                    stateManager);
            }
            objects.addAll(set);
        }
    }

    /**
     * Description of the method.
     *
     * @return Description of the returned value.
     */
    public boolean hasFinished() {
        return false;
    }

    /**
     * Description of the method.
     */
    public void cancel() {
        if (canCancel()) {
            isCanceled = true;
        }
    }

    /**
     * Description of the method.
     *
     * @return Description of the returned value.
     */
    public boolean canCancel() {
        return canCancel;
    }

    /**
     * Method description.
     *
     * @return Return description.
     */
    public String getDescription() {
        return description;
    }

    /**
     * Method description.
     *
     * @return Return description.
     */
    public int getProgress() {
        return (int) progress;
    }

    /**
     * Method description.
     */
    public void resetComputation() {
        isCanceled = false;
    }

    /**
     * Method description.
     *
     * @param canCancel Parameter description.
     */
    public void setCancelEnable(boolean canCancel) {
        this.canCancel = canCancel;
    }
}
