/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.objects.visitors;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;

import org.srs3d.viewer.objects.AbstractObject;
import org.srs3d.viewer.objects.ObjectFilter;
import org.srs3d.viewer.util.Log;

/**
 * This class provides the methods to collect all the <code>Objects</code> with specified
 * properties. Providing <code>null</code> as filter will collect the full hierarchy.
 *
 * @author Karsten Klein
 * @version January 20, 2001
 */
public class ObjectCollector extends AbstractVisitor {
    private static final Log log = new Log(ObjectCollector.class);

    /** Collect all mode. */
    public static final int COLLECT_ALL = 1;
    public static final int COLLECT_FLAT = 2;

    /** Find one mode. The first object that applies to the filter will stop the search. */
    public static final int COLLECT_ONE = 4;
    private Collection objects = new HashSet();
    private ObjectFilter objectFilter;
    private int mode = COLLECT_ALL;

    /**
     * <code>ObjectCollector</code> constructor.
     *
     * @param objectFilter Description of parameter.
     */
    public ObjectCollector(ObjectFilter objectFilter) {
        setObjectFilter(objectFilter);
    }

    /**
     * Sets the <code>objectFilter</code> attribute of the <code>ObjectCollector </code>
     * object.
     *
     * @param objectFilter The new <code>objectFilter</code> value.
     */
    public final void setObjectFilter(ObjectFilter objectFilter) {
        this.objectFilter = objectFilter;
    }

    /**
     * Sets the <code>mode</code> attribute of the <code>ObjectCollector</code> object.
     *
     * @param mode The new <code>mode</code> value.
     */
    public final void setMode(int mode) {
        this.mode = mode;
        if ((mode & COLLECT_ONE) != 0 && !objects.isEmpty()) {
            log.warn(
                "setting COLLECT_ONE policy one already filled object list.");
        }
    }

    /**
     * Sets the <code>objects</code> attribute of the <code>ObjectCollector</code>
     * object.
     *
     * @param objects The new <code>objects</code> value.
     */
    public final void setObjects(Collection objects) {
        if ((mode & COLLECT_ONE) != 0 && !objects.isEmpty()) {
            log.warn("setting the objects with COLLECT_ONE policy.");
        }
        this.objects = objects;
    }

    /**
     * Gets the <code>objects</code> attribute of the <code>ObjectCollector </code>
     * object.
     *
     * @return The <code>objects</code> value.
     */
    public final Collection getObjects() {
        return objects;
    }

    /**
     * Description of the method.
     *
     * @param object Description of parameter.
     */
    public final void visit(AbstractObject object) {
        boolean isVisiting = true;

        // this stops the collection if the collector is supposed to collect one
        // object.
        if ((mode & COLLECT_ONE) != 0) {
            if (!objects.isEmpty()) {
                isVisiting = false;
            }
        }
        if (isVisiting && object != null) {
            if (objectFilter == null || objectFilter.filter(object)) {
                objects.add(object);
            }
            if ((mode & COLLECT_FLAT) == 0) {

                // apply filter to the object
                if (objectFilter == null || objectFilter.preFilter(object)) {
                    ArrayList children = new ArrayList();
                    object.getAllObjects(children);
                    visit(children);
                }
            }
        }
    }

    /**
     * Description of the method.
     */
    public final void clearObjects() {
        objects.clear();
    }

    /**
     * Sets the mode of the object collector
     *
     * @return Description of the returned value.
     */
    public final boolean hasObjects() {
        return !objects.isEmpty();
    }
}
