/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.objects.visitors;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;

import org.srs3d.viewer.objects.AbstractObject;
import org.srs3d.viewer.objects.ObjectManager;

/**
 * Localizes the object in the object hierarchy, by accumulating all objects on the way
 * down. Note that the localizer stops the search after it has found the specified
 * search object. For collecting objects with the same properties use the class
 * <code>ObjectCollector</code> .
 *
 * @author Karsten Klein
 * @version March 22, 2001
 */
public final class ObjectLocalizer extends AbstractVisitor {

    /** Terminates the search after the first occurance of the object. */
    public static final int SEARCH_FIRST = 1;

    /** Checks the full hierarchy. */
    public static final int SEARCH_ALL = 2;
    private AbstractObject searchObject;
    private Collection objects = new ArrayList();
    private Collection filterClasses = null;
    private int searchMode = SEARCH_ALL;
    private AbstractObject cachedObject = null;
    private HashSet cachedObjectsChildren = null;

    /**
     * Sets the <code>searchObject</code> attribute of the <code>ObjectLocalizer </code>
     * object.
     *
     * @param searchObject The new <code>searchObject</code> value.
     */
    public void setSearchObject(AbstractObject searchObject) {
        this.searchObject = searchObject;
    }

    /**
     * Sets the <code>searchMode</code> attribute of the <code>ObjectLocalizer </code>
     * object.
     *
     * @param searchMode The new <code>searchMode</code> value.
     */
    public void setSearchMode(int searchMode) {
        this.searchMode = searchMode;
    }

    /**
     * Sets the <code>objects</code> attribute of the <code>ObjectLocalizer</code>
     * object.
     *
     * @param objects The new <code>objects</code> value.
     */
    public void setObjects(Collection objects) {
        this.objects = objects;
    }

    /**
     * Gets the <code>objects</code> attribute of the <code>ObjectLocalizer </code>
     * object.
     *
     * @return The <code>objects</code> value.
     */
    public Collection getObjects() {
        return objects;
    }

    /**
     * Localizes the searchobject in the specified objects hierarchy. This method is used
     * as wrapper for localize() to adhere to the visitor interface
     *
     * @param object Specifies the object for start the search.
     */
    public void visit(AbstractObject object) {
        localize(object);
    }

    /**
     * Localizes the searchobject in the specified objects hierarchy.
     *
     * @param object Specifies the object for start the search.
     *
     * @return Description of the returned value.
     */
    public final boolean localize(AbstractObject object) {
        boolean found = false;
        if (filterClasses != null) {
            if (filterClasses.contains(object.getClass())) {
                return false;
            }
        }
        if (object != searchObject) {
            HashSet children = new HashSet(20);
            object.getAllObjects(children);
            if (children.contains(searchObject)) {

                // direct parent of the searchObject is cached
                cachedObject = object;
                cachedObjectsChildren = children;
                found = true;
            } else {
                Iterator iterator = children.iterator();

                // :OPTIMIZATION: the while block is doubled and optimized. This saves
                //   us 2 checks in the first and 1 check in the second case.
                if (searchMode == SEARCH_ALL) {
                    while (iterator.hasNext()) {
                        found |= localize((AbstractObject) iterator.next());
                    }
                } else {
                    while (!found && iterator.hasNext()) {
                        found = localize((AbstractObject) iterator.next());
                    }
                }
            }
            if (found) {
                if (!objects.contains(searchObject)) {
                    objects.add(searchObject);
                }
                objects.add(object);
            }
        } else {

            // no caching
            // only happens one of the search entry points is the search object itself
            found = true;
            objects.add(searchObject);
        }
        return found;
    }

    /**
     * localizes the search object in the hierarchy of a collection of objects.
     *
     * @param collection Collection of <code>AbstractObjects</code> instances to be
     *        visitied.
     */
    public void localize(Collection collection) {
        boolean found = false;
        if (filterClasses != null) {
            ObjectManager.filter(collection, filterClasses);
        }
        Iterator iterator = collection.iterator();

        // :OPTIMIZATION: the while block is doubled and optimized. This saves
        //   us 2 checks in the first and 1 check in the second case.
        if (searchMode == SEARCH_ALL) {
            while (iterator.hasNext()) {
                localize((AbstractObject) iterator.next());
            }
        } else {
            while (!found && iterator.hasNext()) {
                found = localize((AbstractObject) iterator.next());
            }
        }
    }

    /**
     * Description of the method.
     *
     * @param searchObjects Description of parameter.
     * @param objects Description of parameter.
     */
    public void localize(Collection searchObjects, Collection objects) {
        if (filterClasses != null) {
            ObjectManager.filter(objects, filterClasses);
        }
        Iterator iterator = searchObjects.iterator();
        while (iterator.hasNext()) {
            setSearchObject((AbstractObject) iterator.next());

            // :OPTIMIZATION: we store the last direct parent in the cachedObject
            //   member and its children in the appropriate set. If a cached object
            //   is set we check the children for inclusion of the search object.
            //   If the search object is a child of the cachedObject we are already
            //   finished.
            if (cachedObject != null) {

                // use caching
                if (cachedObjectsChildren.contains(searchObject)) {

                    // cache hit: done, the object is already in the list
                } else {

                    // reset cache and use normal localization procedure
                    cachedObject = null;
                    cachedObjectsChildren = null;
                    localize(objects);
                }
            } else {

                // use normal localization procedure
                localize(objects);
            }
        }
    }

    /**
     * Description of the method.
     *
     * @param searchObjects Description of parameter.
     * @param object Description of parameter.
     */
    public void localize(Collection searchObjects, AbstractObject object) {
        Iterator iterator = searchObjects.iterator();
        while (iterator.hasNext()) {
            setSearchObject((AbstractObject) iterator.next());
            if (cachedObject != null) {

                // use caching
                if (cachedObjectsChildren.contains(searchObject)) {

                    // cache hit: done, the object is already in the list
                } else {

                    // reset cache and use normal localization procedure
                    cachedObject = null;
                    cachedObjectsChildren = null;
                    localize(object);
                }
            } else {

                // use normal localization procedure
                localize(object);
            }
        }
    }

    /**
     * Visits the specified <code>AbstractObject</code> seraching for all objects in the
     * specified collection.
     *
     * @param object Description of parameter.
     * @param searchObjects Description of parameter.
     */
    public void visit(AbstractObject object, Collection searchObjects) {
        Iterator iterator = searchObjects.iterator();
        while (iterator.hasNext()) {
            setSearchObject((AbstractObject) iterator.next());
            visit(object);
        }
    }

    /**
     * Method description.
     *
     * @param filterClasses Parameter description.
     */
    public void setFilterClasses(Collection filterClasses) {
        this.filterClasses = filterClasses;
    }
}
