/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.sequence.constraints;

import javax.vecmath.Point3d;
import javax.vecmath.Point3f;
import javax.vecmath.Vector3d;

import org.srs3d.viewer.annotation.attributes.LayoutPosition;
import org.srs3d.viewer.annotation.constraints.PlaneAnnotationConstraint;
import org.srs3d.viewer.annotation.contexts.AbstractAnnotationContext;
import org.srs3d.viewer.annotation.contexts.AnnotationContextData;
import org.srs3d.viewer.annotation.objects.Alignment;
import org.srs3d.viewer.objects.State;

/**
 * Limits the sequenceview to the current alignmentlength.
 *
 * @author Christian Zofka
 *
 * @created August 17, 2001
 * @reviewed Karsten Fries, LION bioscience AG
 * @reviewed September 21, 2001
 */
public class PlaneSequenceConstraint extends PlaneAnnotationConstraint {

    // stores values for optimization
    private Alignment currentAlignment;
    private float currentViewLength;

    /**
     * <code>PlaneAnnotationConstraint</code> constructor.
     *
     * @param context Description of parameter.
     */
    public PlaneSequenceConstraint(AbstractAnnotationContext context) {
        super(context);
    }

    /**
     * sets the contraints if the alignment or windowsize(viewdistance) have changed
     */
    public void update() {
        AnnotationContextData contextData =
            (AnnotationContextData) getContext().getContextData();
        Alignment alignment = contextData.getActiveBoxAlignment();
        Point3f center = new Point3f(getContext().getViewingPlatformPosition());
        float viewLength = center.z * 0.80f;
        if (alignment != currentAlignment || viewLength != currentViewLength) {
            currentAlignment = alignment;
            currentViewLength = viewLength;
            State.Immutable state =
                contextData.getStateManager().getImmutableState(alignment);
            LayoutPosition.Immutable layoutPosition =
                (LayoutPosition.Immutable) state.getAttribute(LayoutPosition.class);
            if (layoutPosition != null) {
                Point3f position = layoutPosition.getPosition();
                float alignmentLength = alignment.getLength();
                float leftLimit;
                float rightLimit;
                if (alignmentLength < viewLength) {

                    // fixedposition at center of alignment0
                    float x = position.x + alignmentLength / 2.0f;
                    leftLimit = x;
                    rightLimit = x;
                } else {

                    // scroll to leftLimit & rightLimit allowed
                    leftLimit = position.x + viewLength / 2.0f;
                    rightLimit =
                        position.x + alignmentLength - viewLength / 2.0f;

                    // pdb text handling
                    //          leftLimit -= 5;
                }
                Point3d leftAnchor = new Point3d(leftLimit, 0, 0);
                Vector3d leftNormal = new Vector3d(-1, 0, 0);
                Point3d rightAnchor = new Point3d(rightLimit, 0, 0);
                Vector3d rightNormal = new Vector3d(1, 0, 0);
                setLeft(leftAnchor, leftNormal);
                setRight(rightAnchor, rightNormal);
            }
        }
    }
}
