/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.sequence.contexts;

import java.awt.GraphicsConfiguration;

import javax.media.j3d.BoundingSphere;
import javax.media.j3d.Transform3D;
import javax.vecmath.Matrix4f;
import javax.vecmath.Point3d;
import javax.vecmath.Vector3f;

import org.srs3d.viewer.annotation.contexts.AbstractAnnotationContext;
import org.srs3d.viewer.annotation.contexts.AnnotationContextData;
import org.srs3d.viewer.bioatlas.Parameter;
import org.srs3d.viewer.j3d.behaviors.MouseOverBehavior;
import org.srs3d.viewer.j3d.behaviors.SelectZoomBehavior;
import org.srs3d.viewer.j3d.behaviors.UpdateBehavior;
import org.srs3d.viewer.j3d.operations.ZoomBoxOperation;
import org.srs3d.viewer.objects.Operation;

/**
 * Description of the class.
 *
 * @author Christian Zofka
 *
 * @created August 13, 2001
 */
public class SequenceContext extends AbstractAnnotationContext {

    /**
     * <code>SequenceContext</code> contructor.
     */
    public SequenceContext() {
        super();
    }

    /**
     * Constructs a context with the appropriate <code>GraphicsContext</code>
     *
     * @param graphicsConfiguration Description of parameter
     */
    public SequenceContext(GraphicsConfiguration graphicsConfiguration) {
        super(graphicsConfiguration);
    }

    /**
     * Constructs a context with the appropriate <code>GraphicsContext</code> and offset
     * screen flag
     *
     * @param graphicsConfiguration Description of parameter
     * @param offScreen Description of parameter
     */
    public SequenceContext(GraphicsConfiguration graphicsConfiguration,
        boolean offScreen) {
        super(graphicsConfiguration, offScreen);
    }

    /**
     * Sets the <code>ViewingPlatformTransform</code> attribute of the
     * <code>Context</code> object.
     *
     * @param transform The new <code>ViewingPlatformTransform</code> value.
     */
    public void setViewingPlatformTransform(Transform3D transform) {
        if (getSceneTransform() != null) {

            // modifies the transform
            Matrix4f matrix = new Matrix4f();
            transform.get(matrix);
            Transform3D transformS = new Transform3D();
            getSceneTransform().getTransform(transformS);
            Matrix4f matrixScene = new Matrix4f();
            transformS.get(matrixScene);
            if (matrix.m23 != 0) {
                matrixScene.m00 = 350 / matrix.m23;
            }
            matrixScene.m11 =
                (float) getView().getFieldOfView() / (float) getWidth() * 1800;
            transformS.set(matrixScene);
            matrix.m23 = 350;
            matrix.m03 *= matrixScene.m00;
            transform.set(matrix);
            getSceneTransform().setTransform(transformS);
            usesInternalTransformation = true;
        }
        super.setViewingPlatformTransform(transform);
    }

    /**
     * Description of the method.
     */
    public void update() {
        final AnnotationContextData contextData =
            (AnnotationContextData) getContextData();
        final Point3d point = getViewingPlatformPosition();
        point.z =
            (float) getWidth() * (float) getView().getFieldOfView() / Parameter.sequenceLengthFactor;
        UpdateBehavior.Callback callback =
            new UpdateBehavior.Callback() {
                public void execute() {
                    setViewingPlatformPosition(point);
                }
            };
        if (isUpdateBehaviorRunning()) {
            addUpdateCallback(callback);
        } else {
            callback.execute();
        }

        // because of several 0.0 updates at startup,
        // the update with a value could be filtered by the dispatcher
        // if the updates follows too fast
        if (point.z > 0.0f && contextData.getZoomBox() != null) {
            int length = contextData.getActiveBoxAlignment().getLength();
            if (length < point.z * org.srs3d.viewer.bioatlas.Parameter.aspectScaleX) {

                // smaller zoomBox to fit the alignment
                contextData.getZoomBox().setWidth((float) length);
            } else {
                contextData.getZoomBox().setWidth((float) point.z * org.srs3d.viewer.bioatlas.Parameter.aspectScaleX);
            }
            Operation operation =
                new ZoomBoxOperation(this, "MOVE_ZOOMBOX",
                    contextData.getZoomBox());
            contextData.getDispatcher().dispatch(operation);
        }
    }

    /**
     * Method description.
     */
    public void setup() {
        super.setup();
        setupLights(true);

        // create a bounding sphere for the whole bunch of behaviors
        BoundingSphere bounds =
            new BoundingSphere(new Point3d(0.0, 0.0, 0.0), Double.MAX_VALUE);
        SelectZoomBehavior zoomer = new SelectZoomBehavior(getContextData());
        zoomer.setSchedulingBounds(bounds);
        getBehaviorLevel().addChild(zoomer);
        MouseOverBehavior mouseOverBehavior =
            new MouseOverBehavior(getContextData());
        mouseOverBehavior.setEnable(true);
        mouseOverBehavior.setSchedulingBounds(bounds);
        getBehaviorLevel().addChild(mouseOverBehavior);
    }

    /**
     * Method description.
     *
     * @return Return description.
     */
    public boolean isTimeoutExecutionEnabled() {
        return true;
    }

    /**
     * Method description.
     */
    public void adjustClipping() {
        setNeedsClippingAdjustment(false);
        Vector3f vector = new Vector3f(getViewingPlatformPosition());
        double back = 500;
        getView().setBackClipDistance(back);
        getView().setFrontClipDistance(1.2);
    }
}
