/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.sequence.creators;

import java.awt.Color;
import java.util.Collection;
import java.util.Iterator;

import javax.media.j3d.Appearance;
import javax.media.j3d.BranchGroup;
import javax.media.j3d.Geometry;
import javax.media.j3d.ImageComponent2D;
import javax.media.j3d.RenderingAttributes;
import javax.media.j3d.Shape3D;
import javax.media.j3d.TransparencyAttributes;
import javax.vecmath.Point3f;
import javax.vecmath.Vector3f;

import org.srs3d.viewer.annotation.attributes.LayoutPosition;
import org.srs3d.viewer.annotation.objects.ChainAnnotation;
import org.srs3d.viewer.annotation.objects.Segment;
import org.srs3d.viewer.bioatlas.Parameter;
import org.srs3d.viewer.bioatlas.objects.Residue;
import org.srs3d.viewer.j3d.AppearanceHelper;
import org.srs3d.viewer.j3d.ContextData;
import org.srs3d.viewer.j3d.creators.AbstractGeometryCreator;
import org.srs3d.viewer.objects.AbstractObject;
import org.srs3d.viewer.objects.Attribute;
import org.srs3d.viewer.objects.State;
import org.srs3d.viewer.util.Log;

/**
 * Performs layouting and explicit registration for a <code>ChainAnnotation</code> .
 *
 * @author Christian Zofka
 * @author Karsten Klein
 *
 * @created July 31, 2001
 * @reviewed September 21, 2001
 */
public class ChainAnnotationGeometryCreator extends AbstractGeometryCreator {
    private static final Log log =
        new Log(ChainAnnotationGeometryCreator.class);
    private static transient Appearance appearance = null;
    private static final Color opaqueWhite = new Color(1.0f, 1.0f, 1.0f, 1.0f);
    private Vector3f position = new Vector3f();

    /**
     * Description of the method
     *
     * @param object Description of parameter
     * @param branchGroup Description of parameter
     */
    public void create(AbstractObject object, BranchGroup branchGroup) {
        ChainAnnotation chainAnnotation = (ChainAnnotation) object;
        create(chainAnnotation, branchGroup);
    }

    /**
     * Description of the method.
     *
     * @param contextData Description of parameter.
     * @param object Description of parameter.
     */
    public void modifyAttributes(ContextData contextData, AbstractObject object) {
        super.modifyAttributes(contextData, object);
        State.Immutable state =
            getContextData().getStateManager().getImmutableState(object);
        LayoutPosition.Immutable layoutPosition =
            (LayoutPosition.Immutable) state.getAttribute(LayoutPosition.class);
        if (layoutPosition != null) {
            position.set(layoutPosition.getPosition());
        }
        position.y -= 0.5f * Parameter.chainAnnotationHeight;
    }

    /**
     * Description of the method.
     *
     * @param branchGroup Description of parameter.
     * @param chainAnnotation Description of parameter.
     */
    private void create(ChainAnnotation chainAnnotation, BranchGroup branchGroup) {

        // create segment layout
        createBackgroundGeometry(chainAnnotation, branchGroup);
        Collection sequence = chainAnnotation.getGapSequence();
        Residue residue;
        int index = 0;
        if (sequence != null && !sequence.isEmpty()) {
            Iterator iterator = sequence.iterator();
            index = 0;
            Shape3D shape = new Shape3D();
            String letterCode;
            while (iterator.hasNext()) {
                residue = (Residue) iterator.next();
                if (residue != null) {
                    letterCode = residue.getTemplate().getSingleLetterCode();
                    shape.addGeometry(createLetter(letterCode, index));
                }
                index++;
            }
            if (appearance == null) {
                appearance = new Appearance();
                AppearanceHelper.setDefaults(appearance);
                RenderingAttributes renderingAttributes =
                    new RenderingAttributes();
                renderingAttributes.setDepthBufferEnable(false);
                appearance.setRenderingAttributes(renderingAttributes);
                TransparencyAttributes transparencyAttributes =
                    new TransparencyAttributes(TransparencyAttributes.BLENDED,
                        0.0f);
                appearance.setTransparencyAttributes(transparencyAttributes);
            }
            shape.setAppearance(appearance);
            shape.setPickable(false);
            branchGroup.addChild(shape);
        }
    }

    /**
     * Description of the method.
     *
     * @param string Description of parameter.
     * @param index Description of parameter.
     *
     * @return Description of the returned value.
     */
    private Geometry createLetter(String string, float offset) {
        ImageComponent2D image =
            getContextData().getLabelManager().getLabel(string, opaqueWhite,
                true, null);
        javax.media.j3d.Raster raster = new javax.media.j3d.Raster();
        raster.setImage(image);
        raster.setSize(image.getWidth(), image.getHeight());
        float x = Parameter.aspectScaleX * Parameter.residueSequenceWidth / 2;
        x = offset * Parameter.residueSequenceWidth - x;

        // :FIXME: the y value is not general (zooming, font dependend)
        Point3f textPosition = new Point3f(x, 1.5f, 0.0f);
        textPosition.add(position);
        raster.setPosition(textPosition);
        return raster;
    }

    /**
     * Method description.
     *
     * @param chainAnnotation Parameter description.
     * @param branchGroup Parameter description.
     */
    public void createBackgroundGeometry(ChainAnnotation chainAnnotation,
        BranchGroup branchGroup) {
        if (chainAnnotation != null) {
            Point3f position = new Point3f();
            Iterator iterator;
            State.Immutable immutableState;
            LayoutPosition.Immutable immutableLayoutPosition;
            State state;
            LayoutPosition layoutPosition;
            immutableState =
                getContextData().getStateManager().getImmutableState(chainAnnotation);
            immutableLayoutPosition =
                (LayoutPosition.Immutable) immutableState.getAttribute(LayoutPosition.class);
            if (immutableLayoutPosition != null) {
                position.set(immutableLayoutPosition.getPosition());
            }
            position.y -= 0.5f * Parameter.chainAnnotationHeight;

            // iterate over segments annotations and do layout stuff
            iterator = chainAnnotation.getSegments().iterator();
            Segment segment;
            while (iterator.hasNext()) {
                segment = (Segment) iterator.next();

                // modify layout position attribute of chain annotation state
                state = getContextData().getStateManager().getState(segment);
                layoutPosition =
                    (LayoutPosition) Attribute.getInstance(LayoutPosition.class);
                layoutPosition.setPosition(position);
                state.setAttribute(layoutPosition);
                getContextData().getStateManager().register(segment, state);
            }
        } else {
            log.error("chain annotation is null. No geometry created.");
        }
    }
}
