/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.sequence.creators;

import javax.media.j3d.BranchGroup;
import javax.media.j3d.Shape3D;
import javax.vecmath.Matrix3f;
import javax.vecmath.Point3f;
import javax.vecmath.TexCoord2f;
import javax.vecmath.Tuple3f;
import javax.vecmath.Vector3f;

import org.srs3d.viewer.annotation.attributes.LayoutPosition;
import org.srs3d.viewer.annotation.objects.Segment;
import org.srs3d.viewer.bioatlas.Parameter;
import org.srs3d.viewer.bioatlas.objects.Helix;
import org.srs3d.viewer.bioatlas.objects.Section;
import org.srs3d.viewer.bioatlas.objects.Strand;
import org.srs3d.viewer.bioatlas.objects.Turn;
import org.srs3d.viewer.j3d.ContextData;
import org.srs3d.viewer.j3d.ShapeManager;
import org.srs3d.viewer.j3d.creators.AbstractGeometryCreator;
import org.srs3d.viewer.j3d.geometry.GeometryHelper;
import org.srs3d.viewer.objects.AbstractObject;
import org.srs3d.viewer.objects.State;
import org.srs3d.viewer.vecmath.Constants;
import org.srs3d.viewer.vecmath.Contour;
import org.srs3d.viewer.vecmath.SweepingSurface;
import org.srs3d.viewer.vecmath.TexCoordArray2f;
import org.srs3d.viewer.vecmath.VectorArray3f;

/**
 * Description of the class
 *
 * @author Christian Zofka
 *
 * @created July 08, 2001
 */
public class SegmentGeometryCreator extends AbstractGeometryCreator {
    private static final Matrix3f rotation =
        new Matrix3f(0, 0, 1, 0, 1, 0, 1, 0, 0);
    private static final Matrix3f rotation90 =
        new Matrix3f(0, 0, 1, 0, 1, 0, 1, 0, 0);
    private static final Matrix3f identity =
        new Matrix3f(1, 0, 0, 0, 1, 0, 0, 0, 1);
    private Vector3f position = new Vector3f();
    private int segmentMode = Parameter.segmentCylinder;

    /**
     * Description of the method
     *
     * @param object Description of parameter
     * @param branchGroup Description of parameter
     */
    public void create(AbstractObject object, BranchGroup branchGroup) {
        create((Segment) object, branchGroup);
    }

    /**
     * Description of the method
     *
     * @param segment Description of parameter
     * @param branchGroup Description of parameter
     */
    public void create(Segment segment, BranchGroup branchGroup) {
        if ((segmentMode & Parameter.segmentCylinder) != 0) {
            createCylinderGeometry(segment, branchGroup);
        }
    }

    /**
     * Description of the method.
     *
     * @param object Description of parameter.
     * @param contextData Description of parameter.
     */
    public void modifyAttributes(ContextData contextData, AbstractObject object) {
        super.modifyAttributes(contextData, object);
        State.Immutable state =
            getContextData().getStateManager().getImmutableState(object);
        LayoutPosition.Immutable layoutPosition =
            (LayoutPosition.Immutable) state.getAttribute(LayoutPosition.class);
        if (layoutPosition != null) {
            position.set(layoutPosition.getPosition());
        }
    }

    /**
     * Description of the method
     *
     * @param segment Description of parameter
     * @param branchGroup Description of parameter
     */
    private void createCylinderGeometry(Segment segment, BranchGroup branchGroup) {
        Shape3D shape = null;
        int startIndex = segment.getStartIndex();
        int endIndex = segment.getEndIndex() + 1;
        Vector3f position = new Vector3f(this.position);
        position.x -= 0.5f;
        if (startIndex < endIndex) {
            if (segment.getSubchain() != null) {
                if (segment.getSubchain().getCode() == Strand.CODE) {

                    //          shape = createPlaneStrand( segment );
                    if (segment.getEndResidue() == segment.getSubchain()
                                                            .getEndResidue()) {
                        shape =
                            createCylinderArrow(segment, startIndex, endIndex,
                                Parameter.strandSequenceSize, position);
                    } else {
                        shape =
                            createCylinder(segment, startIndex, endIndex,
                                Parameter.strandSequenceSize, position);
                    }
                } else {
                    float size = Parameter.coilSequenceSize;
                    if (segment.getSubchain().getCode() == Helix.CODE) {
                        size = Parameter.helixSequenceSize;
                    } else if (segment.getSubchain().getCode() == Section.CODE) {
                        size = Parameter.helixSequenceSize;
                    } else if (segment.getSubchain().getCode() == Turn.CODE) {
                        size = Parameter.turnSequenceSize;
                    }
                    shape =
                        createCylinder(segment, startIndex, endIndex, size,
                            position);
                }
                if (shape != null) {
                    ShapeManager.setCapabilities(shape, segment);
                    getContextData().getShapeManager().register(segment, shape);
                    branchGroup.addChild(shape);
                }
            }
        }
    }

    /**
     * Description of the method
     *
     * @param segment Description of parameter
     *
     * @return
     */
    private Shape3D createCoil(Segment segment) {
        Shape3D shape = new Shape3D();
        Shape3D tempShape = null;
        int startIndex = segment.getStartIndex();
        int endIndex = segment.getEndIndex() + 1;
        tempShape =
            createCylinder(segment, startIndex, endIndex,
                Parameter.coilAnnotationSize, position);
        shape.addGeometry(tempShape.getGeometry(1));
        return shape;
    }

    /**
     * Description of the method.
     *
     * @param segment Description of parameter.
     *
     * @return Description of the returned value.
     */
    private Shape3D createPlaneStrand(Segment segment) {
        if (segment.getEndResidue() == segment.getSubchain().getEndResidue()) {
            return createPlaneStrandPike(segment);
        }
        Shape3D shape = null;
        int startIndex = segment.getStartIndex();
        int endIndex = segment.getEndIndex() + 1;
        Contour contour = Contour.getRectangleContour();
        contour.setClosed(false);
        SweepingSurface sweepingSurface = new SweepingSurface(contour);
        sweepingSurface.setTextured(true);
        VectorArray3f scale = new VectorArray3f(2);
        scale.setAt(0, Parameter.strandAnnotationDepth,
            Parameter.strandAnnotationSize, 1);
        scale.setAt(1, Parameter.strandAnnotationDepth,
            Parameter.strandAnnotationSize, 1);
        VectorArray3f translation = new VectorArray3f(2);
        translation.setAt(0, startIndex, 0, 0);
        translation.setAt(1, endIndex, 0, 0);
        TexCoord2f textureCoordinate = new TexCoord2f();
        float width = endIndex - startIndex;
        float widthPowerOf2 =
            Constants.computeNextHigherPowerOf2(segment.getSubchain().getLength());
        float deltaTexture = 1.0f / widthPowerOf2;
        float offset =
            deltaTexture * segment.getSubchain().computeResidueIndex(segment.getInitialResidue());
        textureCoordinate.x = offset;
        translation.getAt(0).add(position);
        translation.getAt(1).add(position);
        sweepingSurface.addFrame(rotation90, translation.getAt(0),
            scale.getAt(0), false, textureCoordinate);
        textureCoordinate.x = offset + deltaTexture * width;
        sweepingSurface.addFrame(rotation90, translation.getAt(1),
            scale.getAt(1), false, textureCoordinate);
        shape = sweepingSurface.getFullGeometry(0, 1, false);
        return shape;
    }

    /**
     * Description of the method.
     *
     * @param segment Description of parameter.
     *
     * @return Description of the returned value.
     */
    private Shape3D createPlaneStrandPike(Segment segment) {
        Shape3D shape = null;
        int startIndex = segment.getStartIndex();
        int endIndex = segment.getEndIndex() + 1;
        int length = endIndex - startIndex;
        int coneSize = Parameter.maxConeSize;
        Contour contour = Contour.getRectangleContour();
        contour.setClosed(true);
        SweepingSurface sweepingSurface = new SweepingSurface(contour);
        int start = 0;
        float width = length;
        float widthPowerOf2 =
            Constants.computeNextHigherPowerOf2(segment.getSubchain().getLength());
        float deltaTexture = 1.0f / widthPowerOf2;
        float offset =
            deltaTexture * segment.getSubchain().computeResidueIndex(segment.getInitialResidue());
        if (length == 1) {
            coneSize = 1;
            length = 0;
            start = 2;
        } else if (length <= coneSize) {
            coneSize = length - 1;
        }
        endIndex -= coneSize;
        VectorArray3f scale = new VectorArray3f(4);
        scale.setAt(0, Parameter.strandAnnotationDepth,
            Parameter.strandAnnotationSize, 1);
        scale.setAt(1, Parameter.strandAnnotationDepth,
            Parameter.strandAnnotationSize, 1);
        scale.setAt(2, Parameter.strandAnnotationDepth,
            Parameter.strandAnnotationConeSize, 1);
        scale.setAt(3, Parameter.strandAnnotationDepth,
            Parameter.coilAnnotationSize, 1);
        VectorArray3f translation = new VectorArray3f(4);
        TexCoord2f textureCoordinate = new TexCoord2f();
        TexCoordArray2f textureCoordinates = new TexCoordArray2f(4);
        textureCoordinate.x = offset;
        textureCoordinates.setAt(0, textureCoordinate);
        textureCoordinate.x = offset + deltaTexture * (endIndex - startIndex);
        textureCoordinates.setAt(1, textureCoordinate);
        textureCoordinates.setAt(2, textureCoordinate);
        textureCoordinate.x =
            offset + deltaTexture * (endIndex - startIndex + coneSize);
        textureCoordinates.setAt(3, textureCoordinate);
        translation.setAt(0, startIndex, 0, 0);
        translation.setAt(1, endIndex, 0, 0);
        translation.setAt(2, endIndex, 0, 0);
        translation.setAt(3, endIndex + coneSize, 0, 0);
        sweepingSurface.setTextured(true);
        for (int i = start; i < 4; i++) {
            translation.getAt(i).add(position);
            sweepingSurface.addFrame(rotation90, translation.getAt(i),
                scale.getAt(i), true, textureCoordinates.getAt(i));
        }

        // normally we don't have to artifically destinguish the two cases, but
        // it seems that the getFullGeometry method does not work porperly and produdes
        // degenerated triangles.
        if (start == 2) {
            shape =
                sweepingSurface.getFullGeometry(0,
                    sweepingSurface.getFrameCount() - 1, false);
        } else {
            shape = sweepingSurface.getFullGeometry(0, 1, false);
            GeometryHelper.copyGeometry(sweepingSurface.getFullGeometry(2, 3,
                    false), shape);
        }
        return shape;
    }

    /**
     * Draws the front part of a cylinder without bottom and top geometry. Texturing
     * coodrinates are set lineary approproate to the y position in (segment aligned) in
     * space;
     *
     * @param startIndex Description of parameter.
     * @param endIndex Description of parameter.
     * @param radius Description of parameter.
     * @param position Description of parameter.
     *
     * @return Description of the returned value.
     */
    public static Shape3D createCylinder(Segment segment, float startIndex,
        float endIndex, float radius, Tuple3f position) {
        Contour contour = getPlanarContour();
        contour.setClosed(false);
        float width = endIndex - startIndex;
        float widthPowerOf2 =
            Constants.computeNextHigherPowerOf2(segment.getSubchain().getLength());
        float deltaTexture = 1.0f / widthPowerOf2;
        float offset =
            deltaTexture * segment.getSubchain().computeResidueIndex(segment.getInitialResidue());
        SweepingSurface sweepingSurface = new SweepingSurface(contour);
        sweepingSurface.setTextured(true);
        TexCoord2f textureCoordinate = new TexCoord2f(offset, 0);
        VectorArray3f scale = new VectorArray3f(2);
        scale.setAt(0, radius, radius, 1);
        scale.setAt(1, radius, radius, 1);
        VectorArray3f translation = new VectorArray3f(2);
        translation.setAt(0, startIndex, 0, 0);
        translation.setAt(1, endIndex, 0, 0);
        translation.getAt(0).add(position);
        sweepingSurface.addFrame(identity, translation.getAt(0),
            scale.getAt(0), false, textureCoordinate);
        translation.getAt(1).add(position);

        // :NOTE: the endIndex has already been raised by 1
        textureCoordinate.x = offset + deltaTexture * width;
        sweepingSurface.addFrame(identity, translation.getAt(1),
            scale.getAt(1), false, textureCoordinate);
        return sweepingSurface.getFullGeometry(0, 1, false);
    }

    /**
     * Method description.
     *
     * @param segment Parameter description.
     * @param startIndex Parameter description.
     * @param endIndex Parameter description.
     * @param radius Parameter description.
     * @param position Parameter description.
     *
     * @return Return description.
     */
    public static Shape3D createCylinderArrow(Segment segment,
        float startIndex, float endIndex, float radius, Tuple3f position) {
        Contour contour = getPlanarContour();
        contour.setClosed(false);
        float width = endIndex - startIndex;
        float widthPowerOf2 =
            Constants.computeNextHigherPowerOf2(segment.getSubchain().getLength());
        float deltaTexture = 1.0f / widthPowerOf2;
        float offset =
            deltaTexture * segment.getSubchain().computeResidueIndex(segment.getInitialResidue());
        SweepingSurface sweepingSurface = new SweepingSurface(contour);
        sweepingSurface.setTextured(true);
        TexCoord2f textureCoordinate = new TexCoord2f(offset, 0);
        VectorArray3f scale = new VectorArray3f(4);
        scale.setAt(0, radius, radius, 1);
        scale.setAt(1, radius, radius, 1);
        scale.setAt(2, radius * 1.5f, radius * 1.5f, 1);
        scale.setAt(3, Parameter.coilSequenceSize, Parameter.coilSequenceSize, 1);
        VectorArray3f translation = new VectorArray3f(4);
        translation.setAt(0, startIndex, 0, 0);
        translation.setAt(1, endIndex - 1, 0, 0);
        translation.setAt(2, endIndex - 1, 0, 0);
        translation.setAt(3, endIndex, 0, 0);
        translation.getAt(0).add(position);
        translation.getAt(1).add(position);
        translation.getAt(2).add(position);
        translation.getAt(3).add(position);
        sweepingSurface.addFrame(identity, translation.getAt(0),
            scale.getAt(0), false, textureCoordinate);

        // :NOTE: the endIndex has already been raised by 1
        textureCoordinate.x = offset + deltaTexture * (width - 1);
        sweepingSurface.addFrame(identity, translation.getAt(1),
            scale.getAt(1), false, textureCoordinate);
        sweepingSurface.addFrame(identity, translation.getAt(2),
            scale.getAt(2), false, textureCoordinate);
        textureCoordinate.x = offset + deltaTexture * width;
        sweepingSurface.addFrame(identity, translation.getAt(3),
            scale.getAt(3), false, textureCoordinate);
        if (width > 1) {
            return sweepingSurface.getFullGeometry(0, 3, false);
        } else {
            return sweepingSurface.getFullGeometry(2, 3, false);
        }
    }

    private static Contour getPlanarContour() {
        Contour contour = new Contour(2);
        contour.getCoordinates().setAt(0, new Point3f(0, -1, 0));
        contour.getCoordinates().setAt(1, new Point3f(0, 1, 0));
        contour.getNormals().setUniform(new Vector3f(0, 0, 1));
        return contour;
    }
}
