/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.sequence.dispatcher;

import java.awt.event.KeyEvent;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Vector;

import javax.vecmath.Point3d;
import javax.vecmath.Point3f;
import javax.vecmath.Vector3d;

import org.srs3d.viewer.annotation.attributes.LayoutPosition;
import org.srs3d.viewer.annotation.contexts.AnnotationContextData;
import org.srs3d.viewer.annotation.filters.ChainAnnotationChainFilter;
import org.srs3d.viewer.annotation.filters.ResidueSearchFilter;
import org.srs3d.viewer.annotation.filters.StringSearchFilter;
import org.srs3d.viewer.annotation.objects.Alignment;
import org.srs3d.viewer.annotation.objects.AnnotationContainer;
import org.srs3d.viewer.annotation.objects.ChainAnnotation;
import org.srs3d.viewer.bioatlas.Application;
import org.srs3d.viewer.bioatlas.objects.Chain;
import org.srs3d.viewer.bioatlas.objects.Residue;
import org.srs3d.viewer.bioatlas.operations.SequenceSearchOperation;
import org.srs3d.viewer.interfaces.ApplicationContext;
import org.srs3d.viewer.interfaces.SequencePublisher;
import org.srs3d.viewer.j3d.Context;
import org.srs3d.viewer.j3d.ContextData;
import org.srs3d.viewer.j3d.DispatchManager;
import org.srs3d.viewer.j3d.Selection;
import org.srs3d.viewer.j3d.behaviors.KeyHandler;
import org.srs3d.viewer.j3d.behaviors.UpdateBehavior;
import org.srs3d.viewer.j3d.commands.ColorCommand;
import org.srs3d.viewer.j3d.commands.ExpandCommand;
import org.srs3d.viewer.j3d.commands.SpawnCommand;
import org.srs3d.viewer.j3d.commands.VisibleCommand;
import org.srs3d.viewer.j3d.objects.Rectangle;
import org.srs3d.viewer.j3d.operations.KeyStrokeOperation;
import org.srs3d.viewer.j3d.operations.ZoomBoxOperation;
import org.srs3d.viewer.objects.AbstractObject;
import org.srs3d.viewer.objects.ObjectManager;
import org.srs3d.viewer.objects.Operation;
import org.srs3d.viewer.objects.State;
import org.srs3d.viewer.objects.StateManager;
import org.srs3d.viewer.objects.filters.ObjectClassFilter;
import org.srs3d.viewer.objects.visitors.ObjectCollector;
import org.srs3d.viewer.objects.visitors.ObjectLocalizer;
import org.srs3d.viewer.util.ExceptionHandler;

/**
 * Description of the class.
 *
 * @author Christian Zofka
 *
 * @created August 01, 2001
 */
public class SequenceUpDispatcher
    extends org.srs3d.viewer.annotation.dispatcher.AbstractUpDispatcher {

    /** Description of the field. */
    public AnnotationContainer annotationContainer = null;
    private boolean isFocusLocked = false;
    private String publishedSequence = "";

    // used for the sequence text searching facility
    private String searchString = new String();
    private long searchTimestamp = -1;

    // triggers the zoomBox update communication to the other contextDispatchers
    private transient long zoomBoxUpdateTime = -Integer.MAX_VALUE;
    private Collection cachedChainAnnotations = null;
    private Thread textSearchThread = null;

    /**
     * Handles all incoming operations.
     *
     * @param operation operation.
     * @param context current context.
     */
    public void process(final Context context, final Operation operation) {
        final ContextData contextData = context.getContextData();
        if (operation.is("MOVE_OVERVIEW")) {
            return;
        }
        if (operation.is("MOUSEOVER")) {
            processMouseOver(context.getContextData(), operation);
            return;
        }
        if (operation.is("INTERSECT")) {
            processIntersect(contextData, operation);
            return;
        }
        if (operation.is("KEY_INPUT")) {
            processKeyInput(contextData, operation);
            return;
        }
        if (operation.is("SEARCH_SEQUENCE")) {
            processSearchString(contextData, operation);
        }
        if (operation.is("KEYSTROKE")) {
            processKeystroke(contextData, operation);
            return;
        }
        if (operation.is("UPDATE_COLORING")) {
            processUpdateColoring(contextData, operation);
            return;
        }
        if (operation.is("ZOOM")) {
            if (operation.getContext() == context) {
                DispatchManager.runDispatch(context, operation);
            }
            return;
        }
        if (operation.is("COPY")) {
            processCopy(contextData, operation);
        } else if (operation.is("PASTE")) {
            processPaste(contextData, operation);
        } else if (operation.is("TRANSFER_SELECTION")) {
            processTransferSelection(contextData, operation);
            setRangeAnchor(contextData, null);
            publishSequences(contextData);
        } else if (operation.is("FORWARD_SELECTION")) {
            processTransferSelection(contextData, operation);
            operation.setConsumed(false);
            publishSequences(contextData);
        } else if (operation.is("SELECT")) {
            forwardOperation(contextData, operation);
        } else if (operation.is("SELECT_CONTROL")) {
            forwardOperation(contextData, operation);
        } else if (operation.is("SELECT_SHIFT")) {
            forwardOperation(contextData, operation);
        } else if (operation.is("SELECT_SHIFT_CONTROL")) {
            forwardOperation(contextData, operation);
        } else if (operation.is("SELECT_ALT")) {
            forwardOperation(contextData, operation);
        } else if (operation.is("SELECT_ALT_CTRL")) {
            forwardOperation(contextData, operation);
        } else if (operation.is("MOVE_ZOOMBOX")) {
            if (!isFocusLocked) {
                processPanMove(contextData, operation);
            }
            isFocusLocked = false;
            if (operation.getContext() == contextData.getContext()) {
                if (operation.is("MOVE_ZOOMBOX")) {
                    ((AnnotationContextData) contextData).getZoomBoxThread()
                     .schedule(contextData.getContext(), operation);

                    // this is the same as
                    // DispatchManager.dispatch( contextData.getContext(), operation );
                    // only scheduled in the appropriate thread
                }
            }
        } else if (operation.is("SWITCH_ALIGNMENT")) {
            processSwitchAlignment(contextData, operation);
            context.waitForUpdate();
        }
    }

    /**
     * Centers the selection.
     *
     * @param contextData Description of parameter.
     */
    public void selectionZoom(final AnnotationContextData contextData,
        final ContextData sourceContextData, final AbstractObject priorityObject) {
        if (!isFocusLocked) {
            contextData.getContext().addUpdateCallback(new UpdateBehavior.Callback() {
                    public void execute() {
                        Collection set = null;
                        ContextData referenceContextData =
                            ((AnnotationContextData) contextData).getReferenceContextData();
                        Selection selection =
                            referenceContextData.getSelectionManager()
                                                .getSelection();
                        boolean failed = priorityObject == null;
                        if (!failed) {
                            set = new HashSet();
                            set.add(priorityObject);
                            contextData.getReferenceContextData()
                                       .getObjectManager().getDownAssociations(priorityObject,
                                set);

                            // needed to map from atoms as priority objects to residues
                            HashSet residues = new HashSet();
                            ObjectManager.extract(set, residues, Residue.class);
                            HashSet directUp = new HashSet(set.size());
                            while (!failed && residues.isEmpty()) {
                                directUp.clear();
                                referenceContextData.getObjectManager()
                                                    .getDirectUpAssociations(set,
                                    directUp);
                                if (directUp.isEmpty()) {
                                    failed = true;
                                }
                                set.addAll(directUp);
                                ObjectManager.extract(set, residues,
                                    Residue.class);
                            }
                            directUp.clear();
                            contextData.getReferenceContextData()
                                       .getObjectManager().getDownAssociations(set,
                                directUp);
                            set.addAll(directUp);
                        }
                        if (failed) {
                            set = new HashSet(selection);
                            referenceContextData.getObjectManager()
                                                .getAssociations(selection, set);
                        }

                        // the active alignment can also be determined by the priority object
                        Alignment alignment =
                            contextData.getActiveBoxAlignment();
                        float position =
                            annotationContainer.getPosition(alignment);
                        StateManager stateManager =
                            contextData.getStateManager();
                        State.Immutable state =
                            stateManager.getImmutableState(alignment);
                        LayoutPosition.Immutable layoutPosition =
                            (LayoutPosition.Immutable) state.getAttribute(LayoutPosition.class);
                        if (layoutPosition != null) {
                            position = layoutPosition.getPosition().x;
                        }
                        int extend = (int) contextData.getZoomBox().getWidth();
                        int center = alignment.getLength();
                        int tempCenter = center;
                        ChainAnnotation chainAnnotation;
                        Iterator iterator =
                            alignment.getChainAnnotations().iterator();
                        while (iterator.hasNext()) {
                            chainAnnotation = (ChainAnnotation) iterator.next();
                            if (chainAnnotation != null) {
                                tempCenter =
                                    computeSelectionCenter(chainAnnotation.getGapSequence(),
                                        set, extend, center + extend);
                            }
                            if (tempCenter < center) {
                                center = tempCenter;
                            }
                        }
                        contextData.getZoomBox().getCoordinate().x =
                            position + center;
                        ZoomBoxOperation operation =
                            new ZoomBoxOperation(contextData.getContext(),
                                "MOVE_ZOOMBOX", contextData.getZoomBox());
                        dispatch(operation);
                    }
                });
        }

        // the lock is released after each try
        isFocusLocked = false;
    }

    /**
     * Finds the current alignment (switches if neccessary); moves viewport to selection
     * (starting left & centered)
     *
     * @param contextData Description of parameter.
     *
     * @return Description of the returned value.
     */
    public Alignment findLeftAlignment(AnnotationContextData contextData,
        AbstractObject priorityObject) {
        if (annotationContainer != null) {
            ContextData referenceContextData =
                ((AnnotationContextData) contextData).getReferenceContextData();
            ObjectManager objectManager = contextData.getObjectManager();
            Collection selection =
                referenceContextData.getSelectionManager().getSelection();
            Collection set = null;
            if (priorityObject != null) {
                set = new HashSet();
                referenceContextData.getObjectManager().getAssociations(priorityObject,
                    set);
                set.add(priorityObject);
            } else {
                if (selection != null && !selection.isEmpty()) {

                    // center to associated chains
                    set = new HashSet(selection);

                    // we are first filtering the aligments on chain level
                    referenceContextData.getObjectManager().getAssociations(selection,
                        set);
                }
            }
            if (set != null) {
                ObjectManager.extract(set, Chain.class);
                ChainAnnotationChainFilter chainFilter =
                    new ChainAnnotationChainFilter();
                ObjectCollector objectCollector =
                    new ObjectCollector(chainFilter);
                Object object;

                // search for those annotations, that contain parts of the chain
                Iterator iterator = set.iterator();
                while (iterator.hasNext()) {
                    chainFilter.setChain((Chain) iterator.next());
                    objectCollector.visit(annotationContainer);
                }

                // we have collected chainAnnotation that potentially contain part of
                // the selection
                set = objectCollector.getObjects();

                // now we have to filter those that really are part of the selection
                ResidueSearchFilter residueFilter = new ResidueSearchFilter();
                objectCollector = new ObjectCollector(residueFilter);

                // extract all residues from the selection
                Collection residues = new HashSet(selection);
                referenceContextData.getObjectManager().getAssociations(selection,
                    residues);
                ObjectManager.extract(residues, Residue.class);

                // search in the preprocessed set of ChainAnnotations for those
                // that contain at least on of the residues
                iterator = residues.iterator();
                Residue residue;
                while (iterator.hasNext()) {
                    residue = (Residue) iterator.next();
                    residueFilter.setResidue(residue);
                    objectCollector.visit(set);

                    // we remove all ChainAnnotation where we already know they are selected
                    // we don't have to revisit them
                    set.removeAll(objectCollector.getObjects());
                }
                set = objectCollector.getObjects();

                // look up the first alignment
                ObjectLocalizer localizer = new ObjectLocalizer();
                localizer.setSearchMode(ObjectLocalizer.SEARCH_FIRST);
                localizer.localize(set, annotationContainer);
                set = localizer.getObjects();
                ObjectManager.extract(set, Alignment.class);
                if (!set.isEmpty()) {
                    Alignment alignment = null;

                    // here is an intrinsic ordering, because getAlignments() returns
                    // a vector
                    iterator = annotationContainer.getAlignments().iterator();
                    while (iterator.hasNext() && !set.contains(alignment)) {
                        alignment = (Alignment) iterator.next();
                    }
                    return alignment;
                }
            }
        }
        return null;
    }

    /**
     * Method description.
     *
     * @param contextData Parameter description.
     * @param operation Parameter description.
     */
    public void processKeystroke(final ContextData contextData,
        final Operation operation) {
        operation.setConsumed(true);
        Context context = contextData.getContext();
        KeyStrokeOperation keyStrokeOperation = (KeyStrokeOperation) operation;
        KeyEvent keyEvent = (KeyEvent) keyStrokeOperation.getInputEvent();
        if (keyEvent != null) {
            int keyCode = keyEvent.getKeyCode();
            char keyChar = keyEvent.getKeyChar();
            int keyModifiers = keyEvent.getModifiers();
            boolean isAltDown = (keyModifiers & KeyEvent.ALT_MASK) != 0;
            boolean isShiftDown = (keyModifiers & KeyEvent.SHIFT_MASK) != 0;
            boolean isControlDown = (keyModifiers & KeyEvent.CTRL_MASK) != 0;
            boolean clearSelection = false;
            long delayTime = System.currentTimeMillis() - searchTimestamp;
            if (delayTime > 3000) {
                searchString = "";
            }
            if (keyCode == KeyEvent.VK_ESCAPE) {
                searchString = "";
                clearSelection = true;
            } else if (keyCode == KeyEvent.VK_BACK_SPACE) {
                if (searchString.length() > 0) {
                    searchString =
                        searchString.substring(0, searchString.length() - 1);
                    processTextSearch(contextData);
                } else {
                    java.awt.Toolkit.getDefaultToolkit().beep();
                    searchString = "";
                    clearSelection = true;
                }
            } else if (!isControlDown) {
                if (!isAltDown) {
                    if ((keyCode >= KeyEvent.VK_A && keyCode <= KeyEvent.VK_Z) ||
                          keyChar == '*' || keyChar == '?') {
                        searchString += keyChar;
                        processTextSearch(contextData);
                    }
                    if (!isShiftDown) {
                        if (keyChar == ',') {
                            processPan(contextData, -1.0f);
                        } else if (keyChar == '.') {
                            processPan(contextData, 1.0f);
                        }
                    }
                }
            } else {
                if (keyChar == ',') {
                    processPan(contextData, -10.0f);
                } else if (keyChar == '.') {
                    processPan(contextData, 10.0f);
                }
            }
            if (clearSelection) {
                Operation op =
                    new Operation(context, "REQUEST_SELECTION", null);
                op.setObjects(new HashSet());
                op.setSerializable(true);
                DispatchManager.dispatch(context, op);
            }
        }
        searchTimestamp = System.currentTimeMillis();
    }

    /**
     * Description of the method.
     *
     * @param set Description of parameter.
     */
    public void filterContextObjects(Collection set) {
        ObjectManager.filter(set, ChainAnnotation.class);
        ObjectManager.filter(set, Alignment.class);
        ObjectManager.filter(set, AnnotationContainer.class);
    }

    /**
     * Description of the method.
     *
     * @param contextData Description of parameter.
     * @param operation Description of parameter.
     */
    public void processCopy(ContextData contextData, Operation operation) {
        String string = createSequenceReport(contextData);
        java.awt.Toolkit.getDefaultToolkit().getSystemClipboard().setContents(new java.awt.datatransfer.StringSelection(
                string), null);
    }

    /**
     * Description of the method. Sets the <code>rangeAnchor</code> attribute of the
     * <code>SequenceUpDispatcher</code> object.
     *
     * @param contextData Description of parameter.
     * @param object The new <code>rangeAnchor</code> value.
     */
    protected void setRangeAnchor(ContextData contextData, AbstractObject object) {
        setAnchor(object);
        getPreviousSelection().clear();
        getPreviousTransferSelection().clear();
    }

    protected void selectRange(ContextData contextData,
        AbstractObject startObject, AbstractObject endObject) {

        // empty; range selection is performed by 3d view
    }

    /**
     * Description of the method.
     *
     * @param contextData Description of parameter.
     * @param operation Description of parameter.
     */
    protected void processTransferSelection(final ContextData contextData,
        final Operation operation) {
        final AnnotationContextData annotationContextData =
            (AnnotationContextData) contextData;
        contextData.getContext().addUpdateCallback(new UpdateBehavior.Callback() {
                public void execute() {

                    // update selection
                    processUpdateColoring(contextData, null);

                    // adjust alignment
                    Alignment alignment =
                        findLeftAlignment(annotationContextData,
                            operation.getObject());
                    boolean adaptZoomToSelection = true;
                    if (alignment == null) {
                        alignment =
                            annotationContextData.getActiveBoxAlignment();
                        adaptZoomToSelection = false;
                    }
                    if (alignment != null) {
                        if (alignment != annotationContextData.getActiveBoxAlignment()) {
                            Context context = contextData.getContext();
                            Operation switchOperation =
                                new Operation(context, "SWITCH_ALIGNMENT",
                                    alignment);
                            switchOperation.setSerializable(false);

                            // :TEST: was dispatch only
                            DispatchManager.runDispatch(context, switchOperation);
                            processSwitchAlignment(annotationContextData,
                                switchOperation);
                        }
                        if (adaptZoomToSelection) {
                            selectionZoom(annotationContextData,
                                operation.getContext().getContextData(),
                                operation.getObject());
                        }
                    }
                }
            });
    }

    /**
     * Description of the method.
     *
     * @param step Description of parameter.
     * @param contextData Description of parameter.
     */
    protected void processPan(ContextData contextData, float step) {
        Rectangle zoomBox = ((AnnotationContextData) contextData).getZoomBox();
        zoomBox.getCoordinate().x += step;
        Operation operation =
            new ZoomBoxOperation(contextData.getContext(), "MOVE_ZOOMBOX",
                zoomBox);
        process(contextData.getContext(), operation);
    }

    /**
     * sets the viewingPlatform and checks for the right contrains.
     *
     * @param operation operation with updated zoomBox that contains the new
     *        centerPosition.
     * @param contextData Description of parameter.
     */
    private void processPanMove(final ContextData contextData,
        final Operation operation) {
        UpdateBehavior.Callback callback =
            new UpdateBehavior.Callback() {
                public void execute() {
                    Rectangle zoomBox = (Rectangle) operation.getObject();
                    if (zoomBox != null) {
                        Point3d point =
                            contextData.getContext().getViewingPlatformPosition();
                        float x =
                            limitXRange((AnnotationContextData) contextData,
                                zoomBox.getCoordinate().x);
                        contextData.getContext().setViewingPlatformPosition(new Vector3d(
                                x, point.y, point.z));
                    }
                }
            };
        if (contextData.getContext().isUpdateBehaviorRunning()) {
            contextData.getContext().addUpdateCallback(callback);
        } else {
            callback.execute();
        }
    }

    /**
     * Replaces the current alignment with the new one (hides, collapse and unregisters
     * the current one).
     *
     * @param operation SWITCH_ALIGNMENT operation.
     * @param contextData Description of parameter.
     */
    private void processSwitchAlignment(final ContextData contextData,
        final Operation operation) {
        contextData.getContext().addUpdateCallback(new UpdateBehavior.Callback() {
                public void execute() {
                    Alignment alignment = (Alignment) operation.getObject();
                    Alignment oldAlignment =
                        ((AnnotationContextData) contextData).getActiveBoxAlignment();
                    if (oldAlignment != alignment) {
                        Collection alignments =
                            ((AnnotationContextData) contextData).getAnnotationContainer()
                             .getAlignments();
                        VisibleCommand visibleCommand =
                            new VisibleCommand(contextData, false);
                        ExpandCommand expandCommand =
                            new ExpandCommand(contextData, false);

                        // some cleanup
                        contextData.getObjectManager().clear();
                        contextData.getColorSchemeManager().clear();
                        contextData.getAppearanceManager().clear();
                        contextData.getShapeManager().clear();
                        contextData.getStateManager().clear();

                        // hide all alignments (by collapsing)
                        contextData.getStrategyManager().execute(alignments,
                            expandCommand);

                        // except the current one
                        ((AnnotationContextData) contextData).setActiveBoxAlignment(alignment);
                        expandCommand =
                            new ExpandCommand(contextData, true, false);
                        contextData.getStrategyManager().execute(alignment,
                            expandCommand);
                        Collection spawners = new HashSet();
                        spawners.add(annotationContainer);
                        SpawnCommand.searchSpawners(contextData, spawners);
                        SpawnCommand spawn = new SpawnCommand(contextData);
                        contextData.getStrategyManager().execute(annotationContainer,
                            spawn);
                        ColorCommand colorCommand =
                            new ColorCommand(contextData,
                                contextData.getColorSchemeBucket());
                        colorCommand.setRespawning(false);
                        contextData.getStrategyManager().execute(contextData.getShapeManager()
                                                                            .getObjects(),
                            colorCommand);
                    }
                }
            });
    }

    private void processTextSearch(final ContextData contextData) {
        if (textSearchThread != null) {
            textSearchThread.interrupt();
        } else {
            textSearchThread =
                new Thread(new Runnable() {
                        public void run() {
                            try {
                                Thread.sleep(500);
                                processTextSearch_(contextData);
                                textSearchThread = null;
                            } catch (InterruptedException e) {

                                // :SILENT EXCEPTION:
                                run();
                            }
                        }
                    });
            textSearchThread.start();
        }
        contextData.getAppletStub().getAppletContext().showStatus("Search String: " +
            searchString);
    }

    /**
     * Description of the method.
     *
     * @param contextData Description of parameter.
     */
    private void processTextSearch_(final ContextData contextData) {
        contextData.getContext().addUpdateCallback(new UpdateBehavior.Callback() {
                public void execute() {
                    try {
                        StringSearchFilter filter =
                            new StringSearchFilter(searchString);
                        Collection foundSequence = new HashSet();
                        ChainAnnotation chainAnnotation;
                        if (cachedChainAnnotations == null) {
                            cachedChainAnnotations = new ArrayList(10);
                            ObjectCollector objectCollector =
                                new ObjectCollector(new ObjectClassFilter(
                                        ChainAnnotation.class));
                            objectCollector.setObjects(cachedChainAnnotations);
                            objectCollector.visit(contextData.getObjectContainer()
                                                             .getObjects());
                        }
                        Iterator iterator = cachedChainAnnotations.iterator();
                        while (iterator.hasNext()) {
                            chainAnnotation = (ChainAnnotation) iterator.next();
                            foundSequence.addAll(filter.filter(
                                    chainAnnotation.getGapSequence()));
                        }
                        Operation operation =
                            new Operation(contextData.getContext(),
                                "REQUEST_SELECTION", null);
                        operation.setSerializable(true);
                        publishedSequence = searchString;
                        if (foundSequence != null &&
                              (!foundSequence.isEmpty() ||
                              searchString.length() == 0)) {
                            HashSet collapsed =
                                new HashSet(foundSequence.size());
                            ((AnnotationContextData) contextData).getReferenceContextData()
                             .getObjectManager().collapseUpExtended(foundSequence,
                                collapsed);
                            operation.setObjects(collapsed);
                            DispatchManager.runDispatch(contextData.getContext(),
                                operation);
                        } else {
                            java.awt.Toolkit.getDefaultToolkit().beep();
                        }
                    } catch (Exception e) {
                        ExceptionHandler.handleException(e,
                            ExceptionHandler.VISIBLE_IN_RELEASE, this);
                    }
                }
            });
    }

    private ApplicationContext getApplicationContext(ContextData contextData) {
        ContextData referenceContextData =
            ((AnnotationContextData) contextData).getReferenceContextData();
        Application application =
            ((Application) referenceContextData.getProperty("Application"));
        if (application != null) {
            return application.getApplicationContext();
        }
        return null;
    }

    private void publishSequences(ContextData contextData) {
        ApplicationContext applicationContext =
            getApplicationContext(contextData);
        if (applicationContext != null) {
            SequencePublisher sequencePublisher =
                (SequencePublisher) applicationContext.get("SequencePublisher");
            if (sequencePublisher != null) {
                String sequences = createSequenceString(contextData);
                if (!sequences.equals(publishedSequence)) {
                    if (sequences.indexOf("-") == -1) {
                        publishedSequence = sequences;
                        sequencePublisher.publish(sequences);
                    }
                }
            }
        }
    }

    /**
     * Description of the method.
     *
     * @param contextData Description of parameter.
     *
     * @return Description of the returned value.
     */
    private String createSequenceString(ContextData contextData) {
        Vector reportSequences = new Vector();
        Iterator chainIterator =
            ((AnnotationContextData) contextData).getActiveBoxAlignment()
             .getChainAnnotations().iterator();
        ChainAnnotation chainAnnotation;
        boolean hasSelection = false;
        ContextData referenceContextData =
            ((AnnotationContextData) contextData).getReferenceContextData();
        Collection selection =
            referenceContextData.getSelectionManager().getSelection();
        ObjectCollector collector =
            new ObjectCollector(new ObjectClassFilter(Residue.class));
        collector.visit(selection);
        selection = collector.getObjects();
        while (chainIterator.hasNext()) {
            chainAnnotation = (ChainAnnotation) chainIterator.next();
            Iterator iterator = chainAnnotation.getGapSequence().iterator();
            String sequence = new String();
            String selected = new String();
            Residue residue;
            hasSelection = false;
            while (iterator.hasNext()) {
                residue = (Residue) iterator.next();
                if (residue == null) {
                    sequence += "-";
                    selected += "-";
                } else {
                    sequence += residue.getTemplate().getSingleLetterCode();
                    if (selection.contains(residue)) {
                        selected += residue.getTemplate().getSingleLetterCode();
                        hasSelection = true;
                    } else {
                        selected += "-";
                    }
                }
            }
            if (hasSelection) {
                reportSequences.add(selected);
            }
        }
        int sequenceCount = reportSequences.size();
        Iterator iterator = reportSequences.iterator();
        String sequence = "";
        String dummy;
        int index;
        while (iterator.hasNext()) {
            dummy = (String) iterator.next();
            index = 0;
            while (dummy.charAt(index) == '-' && index < dummy.length()) {
                index++;
            }
            dummy = dummy.substring(index);
            index = dummy.length() - 1;
            while (dummy.charAt(index) == '-' && index > 0) {
                index--;
            }
            dummy = dummy.substring(0, index + 1);
            sequence += dummy;
            if (iterator.hasNext()) {
                sequence += "|";
            }
        }
        return sequence;
    }

    /**
     * Description of the method.
     *
     * @param contextData Description of parameter.
     *
     * @return Description of the returned value.
     */
    private String createSequenceReport(ContextData contextData) {
        String report =
            new String("CLUSTAL format of " +
                org.srs3d.viewer.bioatlas.Parameter.applicationName +
                " selection\n\n");
        String name;
        Vector reportSequences = new Vector();
        Vector reportSequenceNames = new Vector();
        Iterator chainIterator =
            ((AnnotationContextData) contextData).getActiveBoxAlignment()
             .getChainAnnotations().iterator();
        ChainAnnotation chainAnnotation;
        boolean hasSelection = false;
        int maxOffset = 0;
        ContextData referenceContextData =
            ((AnnotationContextData) contextData).getReferenceContextData();
        Collection selection =
            referenceContextData.getSelectionManager().getSelection();
        ObjectCollector collector =
            new ObjectCollector(new ObjectClassFilter(Residue.class));
        collector.visit(selection);
        selection = collector.getObjects();
        while (chainIterator.hasNext()) {
            chainAnnotation = (ChainAnnotation) chainIterator.next();
            Iterator iterator = chainAnnotation.getGapSequence().iterator();
            String sequence = new String();
            String selected = new String();
            Residue residue;
            hasSelection = false;
            while (iterator.hasNext()) {
                residue = (Residue) iterator.next();
                if (residue == null) {
                    sequence += "-";
                    selected += "-";
                } else {
                    sequence += residue.getTemplate().getSingleLetterCode();
                    if (selection.contains(residue)) {
                        selected += residue.getTemplate().getSingleLetterCode();
                        hasSelection = true;
                    } else {
                        selected += "-";
                    }
                }
            }
            name = chainAnnotation.getName();
            reportSequenceNames.add(name);
            reportSequences.add(sequence);
            if (hasSelection) {
                name += "_SELECTED";
                reportSequenceNames.add(name);
                reportSequences.add(selected);
            }
            if (name.length() > maxOffset) {
                maxOffset = name.length();
            }
        }
        int sequenceCount = reportSequences.size();
        Iterator iterator = reportSequences.iterator();
        String sequence;
        String dummy;
        int size;
        boolean modified = true;
        while (modified) {
            size = reportSequences.size();
            modified = false;
            for (int i = 0; i < size; i++) {
                sequence = (String) reportSequences.elementAt(i);
                if (sequence.length() > 60) {
                    name = (String) reportSequenceNames.elementAt(i);
                    dummy = sequence.substring(60);
                    sequence = sequence.substring(0, 60);
                    reportSequences.setElementAt(sequence, i);
                    modified = true;
                    reportSequences.add(dummy);
                    reportSequenceNames.add(name);
                }
            }
        }
        for (int i = 0; i < reportSequences.size(); i++) {
            if (i > 0 && (i % sequenceCount) == 0) {
                report += "\n";
            }
            name = (String) reportSequenceNames.elementAt(i);
            name += " ";
            sequence = (String) reportSequences.elementAt(i);
            report += name;
            for (int j = name.length(); j <= maxOffset; j++) {
                report += " ";
            }
            report += sequence + "\n";
        }
        return report;
    }

    /**
     * Mainly PlaneAnnotationConstraint is doing it: checks the viewport center so it
     * can't move over the limits of the alignment; centers it when alignmentlength is
     * smaller than viewportlength
     *
     * @param currentX current center x-position of the viewPort.
     * @param contextData Description of parameter.
     *
     * @return new x position.
     */
    public static float limitXRange(AnnotationContextData contextData,
        float currentX) {
        float x = currentX;
        Point3f center =
            new Point3f(contextData.getContext().getViewingPlatformPosition());
        Point3f position = null;
        State.Immutable state =
            contextData.getStateManager().getImmutableState(contextData.getActiveBoxAlignment());
        LayoutPosition.Immutable layoutPosition =
            (LayoutPosition.Immutable) state.getAttribute(LayoutPosition.class);
        if (layoutPosition != null) {
            position = layoutPosition.getPosition();
            float alignmentLength =
                contextData.getActiveBoxAlignment().getLength();
            float viewLength =
                center.z * org.srs3d.viewer.bioatlas.Parameter.aspectScaleX;
            if (alignmentLength < viewLength) {
                x = position.x + alignmentLength / 2.0f;
            } else {
                float viewHalf = viewLength / 2.0f;
                float end = position.x + alignmentLength;
                float lowBorder = position.x + viewHalf;
                if (x < lowBorder) {
                    x = lowBorder;
                } else {
                    float highBorder = end - viewHalf;
                    if (x > highBorder) {
                        x = highBorder;
                    }
                }
            }
        }
        return x;
    }

    /**
     * Calculates the center of the selection; starting from the left all selected
     * residues that fit in width are taken for the calculation
     *
     * @param residues gapSequence from the chainAnnotation.
     * @param selection selected residues.
     * @param width width of the viewingPlatform(=zoomBoxWidth).
     * @param limit max. searchrange (residuesNumber).
     *
     * @return center of the selection.
     */
    private static int computeSelectionCenter(Collection residues,
        Collection selection, int width, int limit) {
        int start = limit;
        int end = limit;
        boolean finishChain = false;
        int residueCount = -1;
        Residue residue;
        Iterator iterator = residues.iterator();
        while (iterator.hasNext() && !finishChain) {
            residue = (Residue) iterator.next();
            residueCount++;
            if (residueCount > limit || residueCount - start > width) {
                finishChain = true;
            }
            if (!finishChain && selection.contains(residue)) {
                if (start == limit) {
                    start = residueCount;
                    end = start;
                } else {
                    end = residueCount;
                }
            }
        }
        return start + (end - start) / 2;
    }

    /**
     * Method description.
     *
     * @param contextData Parameter description.
     * @param operation Parameter description.
     */
    public void processPaste(ContextData contextData, Operation operation) {
        java.awt.datatransfer.Transferable transferable =
            java.awt.Toolkit.getDefaultToolkit().getSystemClipboard()
                            .getContents(null);
        if (transferable != null) {
            try {
                String string =
                    (String) transferable.getTransferData(java.awt.datatransfer.DataFlavor.stringFlavor);

                // :TODO: remove '-'s
                searchString = string;
                processTextSearch(contextData);

                //        selectionZoom( (AnnotationContextData) contextData, null, null );
                // recopy to the clipboard (should prevent clipboard lock)
                java.awt.Toolkit.getDefaultToolkit().getSystemClipboard()
                                .setContents(new java.awt.datatransfer.StringSelection(
                        string), null);
            } catch (java.awt.datatransfer.UnsupportedFlavorException e) {

                // empty
                ExceptionHandler.handleException(e,
                    ExceptionHandler.SILENT_IN_DEBUG, this);
            } catch (java.io.IOException e) {

                // empty
                ExceptionHandler.handleException(e,
                    ExceptionHandler.SILENT_IN_DEBUG, this);
            }
        }
    }

    /**
     * Method description.
     *
     * @param contextData Parameter description.
     * @param operation Parameter description.
     */
    public void processMouseOver(final ContextData contextData,
        Operation operation) {
        AbstractObject object = operation.getObject();
        if (object != null) {
            showStatus(contextData, object.toString());
        } else {
            showStatus(contextData, "");
        }
        KeyHandler.setCursor(contextData, "Select", false);
    }

    protected void forwardOperation(ContextData contextData, Operation operation) {
        isFocusLocked = true;
        if (operation.getObjects() == null || operation.getObjects().isEmpty()) {
            if (operation.getObject() == null) {

                // don't lock the context if nothing is transfered
                // click off selection in sequence context and click secondary structure
                // in annotation context: focus didn't adjust
                isFocusLocked = false;
            }
        }
        super.forwardOperation(contextData, operation);
    }

    /**
     * Method description.
     */
    public void cleanup() {
        if (cachedChainAnnotations != null) {
            cachedChainAnnotations.clear();
            cachedChainAnnotations = null;
        }
    }

    private void processSearchString(ContextData contextData,
        Operation operation) {
        SequenceSearchOperation searchOperation =
            (SequenceSearchOperation) operation;
        searchString = searchOperation.getSequence();
        processTextSearch(contextData);
    }
}
