/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.sequence.modules;

import java.awt.event.ActionEvent;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;

import javax.media.j3d.BoundingSphere;
import javax.media.j3d.BranchGroup;
import javax.vecmath.Point3d;
import javax.vecmath.Point3f;

import org.srs3d.viewer.annotation.attributes.AlignmentLayout;
import org.srs3d.viewer.annotation.attributes.LayoutPosition;
import org.srs3d.viewer.annotation.colorschemes.HomologyColorScheme;
import org.srs3d.viewer.annotation.contexts.AnnotationContext;
import org.srs3d.viewer.annotation.contexts.AnnotationContextData;
import org.srs3d.viewer.annotation.modules.AnnotationModule;
import org.srs3d.viewer.annotation.objects.Alignment;
import org.srs3d.viewer.annotation.objects.AnnotationContainer;
import org.srs3d.viewer.annotation.visitors.GapAnnotationCreator;
import org.srs3d.viewer.bioatlas.Parameter;
import org.srs3d.viewer.bioatlas.modules.AbstractModule;
import org.srs3d.viewer.bioatlas.objects.Layer;
import org.srs3d.viewer.j3d.Context;
import org.srs3d.viewer.j3d.ContextData;
import org.srs3d.viewer.j3d.Selection;
import org.srs3d.viewer.j3d.behaviors.SequencePanXYBehavior;
import org.srs3d.viewer.j3d.commands.ExpandCommand;
import org.srs3d.viewer.j3d.commands.SpawnCommand;
import org.srs3d.viewer.j3d.objects.Rectangle;
import org.srs3d.viewer.objects.ObjectContainer;
import org.srs3d.viewer.objects.ObjectManager;
import org.srs3d.viewer.objects.State;
import org.srs3d.viewer.sequence.constraints.PlaneSequenceConstraint;
import org.srs3d.viewer.sequence.contexts.SequenceContext;
import org.srs3d.viewer.sequence.dispatcher.SequenceUpDispatcher;
import org.srs3d.viewer.util.Log;

/**
 * Description of the class.
 *
 * @author Karsten Klein
 *
 * @created May 14, 2001
 */
public class SequenceModule extends AbstractModule {
    private static final Log log = new Log(SequenceModule.class);
    private AnnotationContext annotationContext = null;
    private SequenceContext sequenceContext = null;
    private Context context = null;

    /**
     * <code>AnnotationModule</code> constructor.
     *
     * @param name Description of parameter.
     * @param context Description of parameter.
     * @param sequenceContext Description of parameter.
     * @param annotationContext Description of parameter.
     */
    public SequenceModule(String name, Context context,
        AnnotationContext annotationContext, SequenceContext sequenceContext) {
        super(name, context.getContextData());
        this.context = context;
        this.annotationContext = annotationContext;
        this.sequenceContext = sequenceContext;
    }

    /**
     * Description of the method.
     *
     * @param e Description of parameter.
     */
    public void react(ActionEvent e) {
        ContextData contextData = context.getContextData();
        Selection selection = contextData.getSelectionManager().getSelection();
        Collection layers = new HashSet();
        contextData.getObjectManager().getUpAssociations(selection, layers);
        layers.addAll(selection);
        ObjectManager.extract(layers, Layer.class);
        ObjectContainer localContainer = new ObjectContainer();
        localContainer.addAll(layers);

        // remove all from current annotation container
        ObjectContainer.clearDeep(sequenceContext.getContextData()
                                                 .getObjectContainer());
        sequenceContext.clearContent();
        visualizeResidueInformation(localContainer, annotationContext,
            sequenceContext, context);
    }

    /**
     * Description of the method.
     */
    public void cleanup() {
        super.cleanup();
        context = null;
        sequenceContext = null;
        annotationContext = null;
    }

    /**
     * Description of the method.
     *
     * @param objectContainer Description of parameter.
     * @param context Description of parameter.
     * @param annotationContext Description of parameter.
     * @param sequenceContext Description of parameter.
     */
    public static void visualizeResidueInformation(
        ObjectContainer objectContainer,
        final AnnotationContext annotationContext,
        final SequenceContext sequenceContext, Context context) {
        AnnotationContainer annotationContainer = null;
        final AnnotationContextData sequenceContextData =
            (AnnotationContextData) sequenceContext.getContextData();
        ObjectContainer sequenceContainer =
            sequenceContextData.getObjectContainer();
        if (annotationContext != null) {
            AnnotationContextData annotationContextData =
                (AnnotationContextData) annotationContext.getContextData();
            annotationContainer =
                annotationContextData.getAnnotationContainer();
        }
        if (annotationContainer == null) {
            GapAnnotationCreator annotationCreator = new GapAnnotationCreator();
            annotationCreator.visit(objectContainer);
            annotationContainer = annotationCreator.getAnnotationContainer();
            if (annotationContainer != null &&
                  annotationContainer.getAlignmentMap() != null) {
                HomologyColorScheme.register(context.getContextData(),
                    annotationContainer);
            }
        }
        if (annotationContainer != null) {
            sequenceContainer.addObject(annotationContainer);
            sequenceContext.setViewingPlatformPosition(new Point3d());
            if (annotationContainer.getAlignmentCount() != 0) {

                // iterate over alignments and do layout stuff
                Iterator iterator =
                    annotationContainer.getAlignments().iterator();
                Alignment alignment;
                ExpandCommand expandCommand =
                    new ExpandCommand(sequenceContextData, false);
                boolean first = true;
                while (iterator.hasNext()) {
                    alignment = (Alignment) iterator.next();
                    if (!first) {
                        sequenceContextData.getStrategyManager().execute(alignment,
                            expandCommand);
                    } else {
                        first = false;
                    }
                }
                annotationContainer.update();
                ((SequenceUpDispatcher) sequenceContextData.getDispatcher()).annotationContainer =
                    annotationContainer;
                final AnnotationContainer ac = annotationContainer;

                // :TEST: initialization problem
                context.addUpdateCallback(new org.srs3d.viewer.j3d.behaviors.UpdateBehavior.Callback() {
                        public void execute() {
                            SpawnCommand spawn =
                                new SpawnCommand(sequenceContextData);
                            spawn.setBranchGroup(sequenceContext.getSceneLevel());
                            sequenceContextData.getStrategyManager().execute(ac,
                                spawn);
                            initializeZoomBox(annotationContext,
                                sequenceContext, ac);
                            BoundingSphere bounds =
                                new BoundingSphere(new Point3d(0.0, 0.0, 0.0),
                                    Double.MAX_VALUE);
                            SequencePanXYBehavior panner =
                                new SequencePanXYBehavior(sequenceContextData);
                            panner.setSchedulingBounds(bounds);
                            panner.setSpeedy(0.0f);
                            PlaneSequenceConstraint constraint =
                                new PlaneSequenceConstraint(sequenceContext);
                            panner.addConstraint(constraint);
                            BranchGroup branchGroup = new BranchGroup();
                            branchGroup.addChild(panner);
                            sequenceContext.getBehaviorLevel().addChild(branchGroup);
                        }
                    });
            }
        } else {
            log.error("no sequence created.");
        }
    }

    /**
     * creates zoomBox; needed even without annotationContext
     *
     * @param annotationContext annotationContext
     * @param sequenceContext sequenceContext
     * @param annotationContainer same annotationContainer is used for both contexts
     */
    private static void initializeZoomBox(AnnotationContext annotationContext,
        SequenceContext sequenceContext, AnnotationContainer annotationContainer) {
        if (annotationContext != null) {
            AnnotationContextData sequenceContextData =
                (AnnotationContextData) sequenceContext.getContextData();

            // get 1. alignment in annotationContainer
            Iterator iterator = annotationContainer.getAlignments().iterator();
            Alignment alignment = (Alignment) iterator.next();
            Point3f position = new Point3f();
            State.Immutable state;
            LayoutPosition.Immutable layoutPosition;
            state =
                annotationContext.getContextData().getStateManager()
                                 .getImmutableState(alignment);
            layoutPosition =
                (LayoutPosition.Immutable) state.getAttribute(LayoutPosition.class);
            AlignmentLayout alignmentLayout =
                (AlignmentLayout) state.getCopy().getAttribute(AlignmentLayout.class);
            if (layoutPosition != null) {
                position.add(layoutPosition.getPosition());
            }

            // creates zoomBoxObject as Rectangle
            Rectangle zoomBox = new Rectangle();
            zoomBox.setHeight(Parameter.chainAnnotationHeight * alignment.getChainAnnotationCount() +
                Parameter.annotationRulerHeight * 2);
            zoomBox.setWidth(150.0f);
            zoomBox.setMinimumExtend(0.0f);

            // set zoomBoxPosition to the center of the 1. alignment
            position.x += alignment.getLength() / 2.0f;
            zoomBox.setCoordinate(position);
            if (alignmentLayout != null) {
                org.srs3d.viewer.annotation.creators.AlignmentGeometryCreator.adjustYValue(alignmentLayout,
                    zoomBox);
            } else {
                zoomBox.setHeight(Parameter.chainAnnotationHeight * alignment.getChainAnnotationCount() +
                    Parameter.annotationRulerHeight * 2);
                position.y -= 0.5f * alignment.getAnnotationUnits().size() * Parameter.annotationUnitHeight;
                position.y += 0.5f * alignment.getChainAnnotationCount() * Parameter.chainAnnotationHeight;
                zoomBox.setCoordinate(position);
            }

            // register/saves zoomBox, current alignment, zoomBoxThread
            sequenceContextData.setZoomBox(zoomBox);
            sequenceContextData.setActiveBoxAlignment(alignment);
            if (annotationContext != null) {

                // insert all zoomBox related things in annotationContext
                AnnotationModule.initializeZoomBox(annotationContext,
                    sequenceContextData);
            }
        }
    }
}
