/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.swing;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;

import javax.swing.JComboBox;
import javax.swing.JMenu;
import javax.swing.ToolTipManager;
import javax.swing.event.MenuEvent;
import javax.swing.event.MenuListener;

/**
 * Customized popup menu. This implementations gathers the modules that finally make up
 * the popup menu. Note that this is basically the datamodel for a JPopupMenu.
 *
 * @author Karsten Klein
 *
 * @created May 14, 2001
 */
public class Menu extends StrategyModule implements MenuListener {
    private JMenu menu = null;
    private String name = null;
    private boolean entriesUpdated = false;
    private ArrayList modules = new ArrayList();
    private transient JComboBox comboBox = null;

    /**
     * <code>Menu</code> constructor.
     *
     * @param name Description of parameter.
     */
    public Menu(String name) {
        this.name = name;
    }

    /**
     * Gets the <code>component</code> attribute of the <code>PopupMenu</code> object.
     *
     * @return The <code>component</code> value.
     */
    public Component getComponent() {
        if (menu == null) {
            menu = new JMenu(name, true);
            menu.addMenuListener(this);
        }
        return menu;
    }

    /**
     * Gets the <code>lastModule</code> attribute of the <code>Menu</code> object.
     *
     * @return The <code>lastModule</code> value.
     */
    public Module getLastModule() {
        return (Module) modules.get(modules.size() - 1);
    }

    /**
     * Gets the <code>name</code> attribute of the <code>Menu</code> object.
     *
     * @return The <code>name</code> value.
     */
    public String getName() {
        return name;
    }

    /**
     * Adds a <code>ActionListener</code> object to the <code>Menu</code> object.
     *
     * @param actionListener The <code>ActionListener</code> object to be added.
     */
    public void addActionListener(ActionListener actionListener) {
    }

    /**
     * Description of the method.
     *
     * @param actionListener Description of parameter.
     */
    public void removeActionListener(ActionListener actionListener) {
    }

    /**
     * Description of the method.
     */
    public void cleanup() {
        Iterator iterator = iterator();
        Module module;
        while (iterator.hasNext()) {
            ((Module) iterator.next()).cleanup();
        }
        clear();
        menu = null;
        name = null;
        if (comboBox != null) {
            comboBox.removeAll();
            comboBox = null;
        }
    }

    /**
     * Description of the method.
     */
    public void update() {
        if (menu != null) {
            Iterator iterator;
            Module module;

            // remove all entries
            menu.removeAll();

            // enter the entries
            iterator = iterator();
            Component component;
            while (iterator.hasNext()) {
                module = (Module) iterator.next();
                component = module.getComponent();
                menu.add(component);
            }
            if (getComponent() != null) {
                getComponent().setVisible(checkVisible());
            }
            if (menu.getItemCount() == 0) {
                menu.setEnabled(false);
            }
            entriesUpdated = false;
        }
    }

    /**
     * Description of the method.
     */
    public void updateEntries() {
        if (!entriesUpdated) {

            // update all entries
            Iterator iterator = iterator();
            Module module;
            while (iterator.hasNext()) {
                module = (Module) iterator.next();
                module.update();
            }
            entriesUpdated = true;
        }
    }

    /**
     * Description of the method.
     *
     * @return Description of the returned value.
     */
    public JComboBox createComboBox() {
        Vector items = new Vector();
        for (int i = 0; i < size(); i++) {
            items.add(get(i).getName());
        }

        // add action listener to combo box
        final JComboBox comboBox = new JComboBox(items);
        if (items.isEmpty()) {
            comboBox.setEnabled(false);
        }
        comboBox.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    if (comboBox.isEnabled()) {
                        boolean comboBoxEnable = comboBox.isEnabled();
                        comboBox.setEnabled(false);
                        Module module = get(comboBox.getSelectedIndex());
                        module.actionPerformed(e);
                        comboBox.setEnabled(comboBoxEnable);
                    }
                }
            });
        comboBox.setLightWeightPopupEnabled(false);
        comboBox.setMaximumSize(new Dimension(150, 30));
        ToolTipManager.sharedInstance().registerComponent(comboBox);
        if (this.comboBox != null) {
            this.comboBox.removeAll();
            this.comboBox = null;
        }
        this.comboBox = comboBox;
        return comboBox;
    }

    /**
     * Adds a <code>ComboBoxActionListeners</code> object to the <code>Menu</code>
     * object.
     *
     * @param comboBox The <code>ComboBoxActionListeners</code> object to be added.
     */
    public void addComboBoxActionListeners(final JComboBox comboBox) {
        Iterator iterator = iterator();
        while (iterator.hasNext()) {
            final Module module = (Module) iterator.next();
            module.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        boolean enable = comboBox.isEnabled();
                        comboBox.setEnabled(false);
                        comboBox.setSelectedItem(module.getName());
                        comboBox.setEnabled(enable);
                    }
                });
        }
    }

    /**
     * Description of the method.
     *
     * @param e Description of parameter.
     */
    public void menuCanceled(MenuEvent e) {
    }

    /**
     * Description of the method.
     *
     * @param e Description of parameter.
     */
    public void menuDeselected(MenuEvent e) {
    }

    /**
     * Description of the method.
     *
     * @param e Description of parameter.
     */
    public void menuSelected(MenuEvent e) {
        if (checkVisible()) {
            CursorManager.setWaitCursor((Component) e.getSource());
            updateEntries();
            CursorManager.restoreCursor((Component) e.getSource());
        }
    }

    /**
     * Description of the method.
     *
     * @param e Description of parameter.
     */
    public void actionPerformed(ActionEvent e) {
    }

    /**
     * Method description.
     *
     * @param module Parameter description.
     */
    public void add(Module module) {
        modules.add(module);
    }

    /**
     * Method description.
     *
     * @return Return description.
     */
    public Iterator iterator() {
        return modules.iterator();
    }

    /**
     * Method description.
     */
    public void clear() {
        modules.clear();
    }

    /**
     * Method description.
     *
     * @return Return description.
     */
    public int size() {
        return modules.size();
    }

    /**
     * Method description.
     *
     * @param index Parameter description.
     *
     * @return Return description.
     */
    public Module get(int index) {
        return (Module) modules.get(index);
    }

    /**
     * Method description.
     *
     * @param map Parameter description.
     */
    public void createReport(Map map) {
        update();
        updateEntries();
        Iterator iterator = iterator();
        String report;
        while (iterator.hasNext()) {
            ((Module) iterator.next()).createReport(map);
        }
    }
}
