/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.urn;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLStreamHandler;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import org.srs3d.viewer.util.Log;

/**
 * URN Stream handler. This Extension of the URLStreamHandler class is able to resolve
 * URLs given a URN. An application can register differnent URNResolvers for different
 * urn type ids. This class contains an extension of URLConnection called URNConnection,
 * which is able to also cache accesses to a URN/URL.
 *
 * @author Karsten Klein
 *
 * @created October 8, 2002
 */
public class URNStreamHandler extends URLStreamHandler {
    private static final Log log = new Log(URNStreamHandler.class);
    private Map urnMap = new HashMap();

    /**
     * URNStreamHandler constructor.
     */
    public URNStreamHandler() {
    }

    /**
     * Retrieves the resolver attribute of the URNStreamHandler object.
     *
     * @param urn Description of parameter.
     *
     * @return The resolver value.
     */
    public URNResolver getResolver(String urn) {
        URNResolver resolver = (URNResolver) urnMap.get(urn);
        if (resolver == null) {
            resolver = (URNResolver) urnMap.get(urn.toLowerCase());
        }
        return resolver;
    }

    /**
     * Retrieves the registeredURNs attribute of the URNStreamHandler object.
     *
     * @return The registeredURNs value.
     */
    public Collection getRegisteredURNs() {
        return Collections.unmodifiableSet(urnMap.keySet());
    }

    /**
     * Description of the method.
     *
     * @param urn Description of parameter.
     * @param resolver Description of parameter.
     */
    public void registerResolver(String urn, URNResolver resolver) {
        urnMap.put(urn, resolver);
    }

    /**
     * Description of the method.
     *
     * @param urn Description of parameter.
     */
    public void removeResolver(String urn) {
        urnMap.remove(urn);
    }

    /**
     * Description of the method.
     *
     * @param urn Description of parameter.
     *
     * @return Description of the returned value.
     */
    public URLConnection openConnection(URL urn) {
        return new URNConnection(urn);
    }

    /**
     * Description of the class.
     *
     * @author Karsten Klein
     *
     * @created October 8, 2002
     */
    public class URNConnection extends URLConnection {

        //    private byte[] cache = null;

        /**
         * URNConnection constructor.
         *
         * @param urn Description of parameter.
         */
        public URNConnection(URL urn) {
            super(urn);
        }

        /**
         * Retrieves the uRNResolver attribute of the URNConnection object.
         *
         * @param urnUrl Description of parameter.
         *
         * @return The uRNResolver value.
         */
        public URNResolver getURNResolver(URL urnUrl) {
            String urn = urnUrl.toString();
            URNResolver resolver = URNStreamHandler.this.getResolver(urn);
            while (resolver == null && urn.indexOf(":") != -1) {
                urn = urn.substring(0, urn.lastIndexOf(":"));
                resolver = URNStreamHandler.this.getResolver(urn);
            }
            if (resolver == null) {
                resolver = LocalURNResolver.getInstance();
            }
            return resolver;
        }

        /**
         * Retrieves the inputStream attribute of the URNConnection object.
         *
         * @return The inputStream value.
         */
        public InputStream getInputStream() {

            //      if (cache == null) {
            //
            //        cache();
            //
            //      }
            //
            //      if (cache != null) {
            //
            //        return new ByteArrayInputStream(cache);
            //
            //      }
            return getURNResolver(getURL()).openStream(getURL().toString());
        }

        /**
         * Description of the method.
         */
        public void connect() {
            getURNResolver(getURL()).connect(getURL().toString());
        }

        /**
         * Description of the method.
         *
         * @param in Description of parameter.
         * @param out Description of parameter.
         *
         * @exception IOException Description of exception.
         */
        private void copy(InputStream in, OutputStream out) throws IOException {
            byte[] buffer = new byte[200000];
            int bytesRead;
            while ((bytesRead = in.read(buffer)) > 0) {
                out.write(buffer, 0, bytesRead);
            }
            out.flush();
            out.close();
            in.close();
        }
    }
}
