/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.util;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;

/**
 * Chronometer for time measurements.
 *
 * @author Karsten Klein
 *
 * @created January 31, 2002
 */
public final class Chronometer {
    private static final Log log = new Log(Chronometer.class);
    private static final long INVALID_TIMESTAMP = -1;
    private static boolean debug = false;
    private static Chronometer defaultChronometer = new Chronometer();
    private static Map chronometerMap = debug ? new Hashtable() : null;
    private Vector timestamps = new Vector();
    private Vector durations = new Vector();
    private boolean isVerbose = false;
    private Map taskMap = debug ? new Hashtable() : null;

    /**
     * Sets the <code>verbose</code> attribute of the <code>Chronometer</code> object.
     *
     * @param isVerbose The new <code>verbose</code> value.
     */
    public void setVerbose(boolean isVerbose) {
        this.isVerbose = isVerbose;
    }

    /**
     * Description of the method.
     *
     * @return Description of the returned value.
     */
    public long computeTime() {
        if (debug) {
            if (!timestamps.isEmpty()) {
                long timestamp = ((Long) timestamps.lastElement()).longValue();
                if (timestamp != INVALID_TIMESTAMP) {
                    long duration =
                        ((Long) durations.lastElement()).longValue();
                    duration += System.currentTimeMillis() - timestamp;
                    timestamp = System.currentTimeMillis();
                    timestamps.set(timestamps.size() - 1, new Long(timestamp));
                    durations.set(durations.size() - 1, new Long(duration));
                    return duration;
                }
            } else {
                log.error("cannot compute time; no timestamp.");
            }
        }
        return 0;
    }

    /**
     * Description of the method.
     */
    public void hold() {
        if (debug) {
            computeTime();
        }
    }

    /**
     * Description of the method.
     */
    public void stop() {
        if (debug) {
            if (!timestamps.isEmpty()) {
                computeTime();
                timestamps.remove(timestamps.size() - 1);
                long duration = ((Long) durations.lastElement()).longValue();
                durations.remove(durations.size() - 1);
                TaskDuration taskDuration = addTask("", duration);
                if (isVerbose) {
                    taskDuration.display();
                }
            } else {
                log.error("stop called without start.");
            }
        }
    }

    /**
     * Description of the method.
     *
     * @param taskName Description of parameter.
     */
    public void stop(String taskName) {
        if (debug) {
            if (!timestamps.isEmpty()) {
                computeTime();
                timestamps.remove(timestamps.size() - 1);
                long duration = ((Long) durations.lastElement()).longValue();
                durations.remove(durations.size() - 1);
                TaskDuration taskDuration = addTask(taskName, duration);
                if (isVerbose) {
                    taskDuration.display();
                }
            } else {
                log.error("stop called without start.");
            }
        }
    }

    /**
     * Adds a <code>Task</code> object to the <code>Chronometer</code> object.
     *
     * @param taskName The <code>Task</code> object to be added.
     * @param duration The <code>Task</code> object to be added.
     *
     * @return Description of the returned value.
     */
    public TaskDuration addTask(String taskName, long duration) {
        if (debug) {
            TaskDuration taskDuration = (TaskDuration) taskMap.get(taskName);
            if (taskDuration == null) {
                taskDuration = new TaskDuration(taskName);
                taskMap.put(taskName, taskDuration);
            }
            taskDuration.add(duration);
            return taskDuration;
        }
        return null;
    }

    /**
     * Description of the method.
     */
    public void start() {
        if (debug) {
            long timestamp = System.currentTimeMillis();
            timestamps.add(new Long(timestamp));
            durations.add(new Long(0));
        }
    }

    /**
     * Description of the method.
     */
    public void display() {
        if (debug) {
            Iterator iterator = taskMap.values().iterator();
            TaskDuration taskDuration;
            while (iterator.hasNext()) {
                taskDuration = (TaskDuration) iterator.next();
                taskDuration.display();
            }
        }
    }

    /**
     * Description of the method.
     */
    public static void displayAll() {
        if (debug) {
            Iterator iterator = chronometerMap.keySet().iterator();
            Chronometer chronometer;
            Object object;
            while (iterator.hasNext()) {
                object = iterator.next();
                chronometer = (Chronometer) chronometerMap.get(object);
                log.info(object.toString() + " - Chronometer:");
                chronometer.display();
                log.info("");
            }
        }
    }

    /**
     * Method description.
     *
     * @param object Parameter description.
     *
     * @return Return description.
     */
    public static Chronometer getChronometer(Object object) {
        if (debug) {
            Chronometer c = (Chronometer) chronometerMap.get(object);
            if (c == null) {
                c = new Chronometer();
                chronometerMap.put(object, c);
            }
            return c;
        }
        return defaultChronometer;
    }

    /**
     * Description of the class.
     *
     * @author Karsten Klein
     *
     * @created January 31, 2002
     */
    private static final class TaskDuration {
        private String name;
        private long accumulatedDuration = 0;
        private long accumulatedDurationCount = 0;

        /**
         * <code>TaskDuration</code> constructor.
         *
         * @param name Description of parameter.
         */
        public TaskDuration(String name) {
            this.name = name;
        }

        /**
         * Description of the method.
         *
         * @param duration Description of parameter.
         */
        public void add(long duration) {
            if (accumulatedDurationCount == 1) {
                accumulatedDuration = 2 * duration;
            } else {
                accumulatedDuration += duration;
            }
            accumulatedDurationCount++;
        }

        /**
         * Description of the method.
         */
        public void display() {
            double duration = accumulatedDuration;
            if (duration > 0) {
                if (accumulatedDurationCount > 1) {
                    duration /= accumulatedDurationCount;
                    log.info(accumulatedDuration + " ms / " + duration +
                        " ms in " + name);
                } else if (accumulatedDurationCount == 1) {
                    log.info(accumulatedDuration + " ms in " + name);
                }
            }
        }
    }
}
