/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.util;

import java.io.BufferedReader;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;

/**
 * Holds, and parses the attributes and subtags of XML type tags. Minimal implementation
 * of an XML parser/writer.
 *
 * @author Karsten Klein
 *
 * @created January 23, 2002
 * @modified August 27, 2002
 */
public class XmlTag {
    private static final Log log = new Log(XmlTag.class);
    private HashMap attributes = new HashMap();
    private ArrayList attributeOrder = new ArrayList();
    private ArrayList subtags = new ArrayList();
    private String name = null;
    private String value = null;

    /**
     * <code>XmlTag</code> constructor.
     *
     * @param name Description of parameter.
     */
    public XmlTag(String name) {
        setName(name);
    }

    /**
     * Sets the <code>name</code> attribute of the <code>XmlTag</code> object.
     *
     * @param name The new <code>name</code> value.
     */
    public void setName(String name) {
        this.name = name;
    }

    /**
     * Sets the <code>attribute</code> attribute of the <code>XmlTag</code> object.
     *
     * @param attribute The new <code>attribute</code> value.
     * @param value The new <code>attribute</code> value.
     */
    public void setAttribute(String attribute, String value) {
        attributes.put(attribute, value);
        attributeOrder.remove(attribute);
        attributeOrder.add(attribute);
    }

    /**
     * Sets the <code>value</code> attribute of the <code>XmlTag</code> object.
     *
     * @param value The new <code>value</code> value.
     */
    public void setValue(String value) {
        this.value = value;
    }

    /**
     * Gets the <code>attribute</code> attribute of the <code>XmlTag</code> object.
     *
     * @param attribute Description of parameter.
     *
     * @return The <code>attribute</code> value.
     */
    public String getAttribute(String attribute) {
        return (String) attributes.get(attribute);
    }

    /**
     * Gets the <code>name</code> attribute of the <code>XmlTag</code> object.
     *
     * @return The <code>name</code> value.
     */
    public String getName() {
        return name;
    }

    /**
     * Gets the <code>comment</code> attribute of the <code>XmlTag</code> object.
     *
     * @return The <code>comment</code> value.
     */
    public boolean isComment() {
        String name = getName();
        if (name != null) {
            return name.startsWith("!--");
        }
        return false;
    }

    /**
     * Gets the <code>tag</code> attribute of the <code>XmlTag</code> object.
     *
     * @param tagName Description of parameter.
     *
     * @return The <code>tag</code> value.
     */
    public XmlTag getTag(String tagName) {
        Iterator iterator = subtags.iterator();
        XmlTag tag;
        while (iterator.hasNext()) {
            tag = (XmlTag) iterator.next();
            if (tag.getName().equalsIgnoreCase(tagName)) {
                return tag;
            }
        }
        return null;
    }

    /**
     * Gets the <code>tags</code> attribute of the <code>XmlTag</code> object.
     *
     * @param tagName Description of parameter.
     *
     * @return The <code>tags</code> value.
     */
    public Collection getTags(String tagName) {
        Iterator iterator = subtags.iterator();
        XmlTag tag;
        Collection tags = new ArrayList();
        while (iterator.hasNext()) {
            tag = (XmlTag) iterator.next();
            if (tag.getName().equalsIgnoreCase(tagName)) {
                tags.add(tag);
            }
        }
        return tags;
    }

    /**
     * Gets the <code>subtags</code> attribute of the <code>XmlTag</code> object.
     *
     * @return The <code>subtags</code> value.
     */
    public Collection getSubtags() {
        return subtags;
    }

    /**
     * Gets the <code>value</code> attribute of the <code>XmlTag</code> object.
     *
     * @return The <code>value</code> value.
     */
    public String getValue() {
        return value;
    }

    /**
     * Gets the <code>attributes</code> attribute of the <code>XmlTag</code> object.
     *
     * @return The <code>attributes</code> value.
     */
    public Collection getAttributes() {
        return attributeOrder;
    }

    /**
     * Description of the method.
     *
     * @param tagName Description of parameter.
     *
     * @return Description of the returned value.
     */
    public XmlTag findTag(String tagName) {
        if (getName().equals(tagName)) {
            return this;
        } else {

            // horizontal search
            XmlTag subtag = getTag(tagName);
            if (subtag == null) {
                Iterator iterator = getSubtags().iterator();
                while (iterator.hasNext()) {

                    // recurse
                    subtag = (XmlTag) iterator.next();
                    subtag = subtag.findTag(tagName);
                    if (subtag != null) {
                        return subtag;
                    }
                }
                return null;
            } else {
                return subtag;
            }
        }
    }

    /**
     * Description of the method.
     *
     * @return Description of the returned value.
     */
    public String toString() {
        return toString(0);
    }

    /**
     * Description of the method.
     *
     * @param spaces Description of parameter.
     *
     * @return Description of the returned value.
     */
    public String toString(int spaces) {
        byte[] spaceArray = new byte[spaces];
        Arrays.fill(spaceArray, (byte) ' ');
        String spaceString = new String(spaceArray);
        String string = spaceString + "<" + getName();
        if (attributes.size() > 0) {
            Iterator iterator = attributeOrder.iterator();
            String attribute;
            while (iterator.hasNext()) {
                attribute = (String) iterator.next();
                string += " " + attribute + "=\"" + getAttribute(attribute) +
                "\"";
            }
        }
        if (subtags.isEmpty() && value == null) {
            if (getName().equals("?xml")) {
                string += "?>";
            } else {
                string += "/>";
            }
        } else {
            string += ">";
            Iterator subTagIterator = subtags.iterator();
            while (subTagIterator.hasNext()) {
                string += "\r\n" +
                ((XmlTag) subTagIterator.next()).toString(spaces + 2);
            }
            if (value != null) {
                string += value;
                if (!subtags.isEmpty()) {
                    string += spaceString;
                }
                string += "</" + getName() + ">\r\n";
            } else {
                string += spaceString + "</" + getName() + ">\r\n";
            }
        }
        return string;
    }

    /**
     * Description of the method.
     *
     * @param string Description of parameter.
     */
    public void readSubtags(String string) {
        if (string.indexOf("/>") == string.length() - 2 ||
              string.indexOf("?>") == string.length() - 2) {

            // no subtags expected
        } else {

            // we cut away the parent tag first
            int startIndex = string.indexOf(">");
            int index = string.lastIndexOf("<");
            string = string.substring(startIndex + 1, index);
            setValue(string);
            boolean terminate;
            String substring;
            do {
                string = string.trim();
                index = getFullTagEndIndex(string);
                if (index != -1) {
                    substring = string.substring(0, index).trim();
                    XmlTag tag = new XmlTag(null);
                    tag.deserialize(substring);
                    subtags.add(tag);
                    string = string.substring(index);
                }
            } while (index != -1);
        }
    }

    /**
     * Description of the method.
     *
     * @param attribute Description of parameter.
     */
    public void removeAttribute(String attribute) {
        attributes.remove(attribute);
        attributeOrder.remove(attribute);
    }

    /**
     * Read the attributes from the passed tag string. Only the attributes of the tag
     * level are read, the attributes of the subtags are not parsed.
     *
     * @param string Tag string.
     */
    private void readAttributes(String string) {
        int index0;
        int index1;
        int keyIndex;
        String value;
        String key;

        // is delimeted by the next tag opening
        int maxIndex = string.indexOf("<", 1);
        if (maxIndex == -1) {
            maxIndex = string.indexOf("/>");
            if (maxIndex == -1) {
                maxIndex = string.indexOf("?>");
            }
        }
        do {
            index0 = string.indexOf("\"");
            index1 = -1;
            if (index0 != -1 && index0 < maxIndex) {
                value = string.substring(index0 + 1);
                index1 = value.indexOf("\"");
                if (index1 != -1) {
                    value = value.substring(0, index1);

                    // read key name
                    key = string.substring(0, index0);
                    keyIndex = key.lastIndexOf("=");
                    key = key.substring(0, keyIndex);
                    key.trim();
                    keyIndex = key.lastIndexOf(" ");
                    if (keyIndex != -1) {
                        key = key.substring(keyIndex).trim();
                    }
                    setAttribute(key, value);
                    string = string.substring(index0 + index1 + 2);
                    maxIndex -= index0 + index1 + 2;
                }
            }
        } while (index1 != -1);
    }

    /**
     * Gets the <code>tagName</code> attribute of the <code>XmlTag</code> class.
     *
     * @param line Description of parameter.
     *
     * @return The <code>tagName</code> value.
     */
    public static String getTagName(String line) {
        line = line.trim();
        int startIndex = line.indexOf("<");
        if (startIndex != -1) {
            int index = line.indexOf(">", startIndex);
            if (index != -1) {
                int spaceIndex = line.indexOf(" ", startIndex);
                if (spaceIndex != -1 && spaceIndex < index) {
                    index = spaceIndex;
                }
            }
            if (index != -1) {
                String name = line.substring(startIndex + 1, index).trim();
                return name;
            }
        }
        return null;
    }

    /**
     * Determines the tag end index in the specified string. If the tag end cannot be
     * determined the result will be -1.
     *
     * @param line Line to parse for the next full tag.
     *
     * @return The string relative index of the tag end.
     */
    public static int getFullTagEndIndex(String line) {
        int index = -1;
        String tagName = getTagName(line);
        if (tagName != null) {
            tagName = tagName.toLowerCase().trim();
            if (tagName.startsWith("!--")) {
                index = line.indexOf("-->");
                if (index == -1) {
                    return -1;
                }
                return index + 3;
            }
            if (tagName.equals("?xml")) {
                index = line.indexOf("?>");
                if (index == -1) {
                    return -1;
                }
                return index + 2;
            }

            // find the next lesser sign index
            int lesserIndex = line.indexOf("<");
            lesserIndex = line.indexOf("<", lesserIndex + 1);
            if (lesserIndex != -1) {
                int slashLesserIndex = line.indexOf("/>");
                if (slashLesserIndex != -1 && lesserIndex > slashLesserIndex) {
                    return slashLesserIndex + 2;
                }
            }
            index = line.toLowerCase().indexOf("</" + tagName + ">");
            if (index != -1) {
                return index + tagName.length() + 3;
            } else {
                int test = line.indexOf(">");
                index = line.indexOf("/>");
                if (index != -1) {
                    if (test < index) {
                        return -1;
                    }
                    return index + 2;
                }
            }
        }
        return index;
    }

    /**
     * Deserialize a XmlTag from the specified reader.
     *
     * @param in BufferedReader for reading the xml content.
     *
     * @return The parsed XML content.
     *
     * @exception java.io.IOException IOException.
     */
    public static XmlTag deserialize(BufferedReader in)
        throws java.io.IOException {
        XmlTag tag = null;
        String line;
        String tagString;
        int index;
        String residualTag = "";
        ArrayList tagStrings = new ArrayList();
        do {
            line = in.readLine();
            if (line != null) {
                line = residualTag + line;
                index = getFullTagEndIndex(line);
                if (index != -1) {
                    tagString = line.substring(0, index);
                    tagStrings.add(tagString.trim());
                    residualTag = line.substring(index);
                } else {
                    residualTag = line;
                }
            }
        } while (line != null);

        // converts string tags to real tags
        Iterator iterator = tagStrings.iterator();
        Collection tags = new ArrayList();
        while (iterator.hasNext()) {
            tagString = (String) iterator.next();
            tag = new XmlTag(null);
            tag.deserialize(tagString);
            if (!tag.isComment()) {
                tags.add(tag);
            }
        }
        if (tags.size() == 1) {
            return (XmlTag) tags.iterator().next();
        } else {

            // wrap in root tag
            XmlTag rootTag = new XmlTag("root");
            rootTag.getSubtags().addAll(tags);
            return rootTag;
        }
    }

    /**
     * Method description.
     *
     * @param string Parameter description.
     *
     * @return Return description.
     *
     * @throws java.io.IOException Exception description.
     */
    public static XmlTag deserializeString(String string)
        throws java.io.IOException {
        StringReader stringReader = new StringReader(string);
        BufferedReader reader = new BufferedReader(stringReader);
        return XmlTag.deserialize(reader);
    }

    /**
     * Create XmlTag from the provided string
     *
     * @param string String with xml formated content.
     */
    public void deserialize(String string) {
        try {

            // parse attributes
            // extract name
            setName(getTagName(string).trim());
            if (!isComment()) {

                // extract key/value pairs
                readAttributes(string);

                // parse subtags
                readSubtags(string);
                if (!subtags.isEmpty() || !attributes.isEmpty()) {
                    value = null;
                }
            }
        } catch (Exception e) {
            ExceptionHandler.handleException(e,
                ExceptionHandler.SILENT_IN_RELEASE, this);
            log.error(": unable to parse " + string, e);
        }
    }

    /**
     * Method description.deseri
     *
     * @param xmlTag Parameter description.
     *
     * @return Return description.
     */
    public static XmlTag copy(XmlTag xmlTag) {
        String serializedString = xmlTag.toString();
        XmlTag copyTag = null;
        try {
            copyTag = XmlTag.deserializeString(serializedString);
        } catch (Exception e) {
            ExceptionHandler.handleException(e,
                ExceptionHandler.SILENT_IN_RELEASE);
        }
        return copyTag;
    }
}
