/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.vecmath;

import javax.media.j3d.GeometryArray;
import javax.media.j3d.TriangleFanArray;
import javax.vecmath.Matrix3f;
import javax.vecmath.Point3f;
import javax.vecmath.TexCoord2f;
import javax.vecmath.Tuple3f;
import javax.vecmath.Vector3f;

/**
 * Description of the class.
 *
 * @author Karsten Klein
 *
 * @created March 20, 2001
 */
public final class Contour {
    private PointArray3f coordinates = null;
    private VectorArray3f normals = null;
    private TexCoordArray2f texture = null;
    private int size;
    private boolean closed = true;

    /**
     * <code>Contour</code> contructor.
     *
     * @param size Description of parameter.
     */
    public Contour(int size) {
        this.size = size;
        coordinates = new PointArray3f(size);
        normals = new VectorArray3f(2 * size);
    }

    /**
     * Sets the <code>Closed</code> attribute of the <code>Contour</code> object.
     *
     * @param closed The new <code>Closed</code> value.
     */
    public final void setClosed(boolean closed) {
        this.closed = closed;
    }

    /**
     * Gets the <code>Coordinates</code> attribute of the <code>Contour</code> object.
     *
     * @return The <code>Coordinates</code> value.
     */
    public final PointArray3f getCoordinates() {
        return coordinates;
    }

    /**
     * Method description.
     *
     * @return Return description.
     */
    public final TexCoordArray2f getTextureCoordinates() {
        return texture;
    }

    /**
     * Gets the <code>Normals</code> attribute of the <code>Contour</code> object.
     *
     * @return The <code>Normals</code> value.
     */
    public final VectorArray3f getNormals() {
        return normals;
    }

    /**
     * Gets the <code>Size</code> attribute of the <code>Contour</code> object.
     *
     * @return The <code>Size</code> value.
     */
    public final int getSize() {
        return size;
    }

    /**
     * Gets the <code>Closed</code> attribute of the <code>Contour</code> object.
     *
     * @return The <code>Closed</code> value.
     */
    public final boolean isClosed() {
        return closed;
    }

    /**
     * Gets the <code>Geometry</code> attribute of the <code>Contour</code> object.
     *
     * @param frame Description of parameter.
     *
     * @return The <code>Geometry</code> value.
     */
    public final GeometryArray getGeometry(Frame frame, boolean isClockwise,
        boolean isTextured) {
        int i;
        int j;
        int contourSize = getSize();
        int contourLoops = contourSize;
        int stripVertexCount = contourSize;
        int stripTriangleCount = contourSize;
        if (!isClosed()) {
            contourLoops--;
            stripTriangleCount--;
        }
        int[] stripIndexCount = new int[1];
        stripIndexCount[0] = stripVertexCount;
        TexCoord2f[] textureCoordinates = null;
        int flags = GeometryArray.COORDINATES | GeometryArray.NORMALS;
        isTextured &= texture != null;
        if (isTextured) {
            flags |= GeometryArray.TEXTURE_COORDINATE_2;
            textureCoordinates = texture.getBuffer();
        }
        TriangleFanArray triangleArray =
            new TriangleFanArray(stripVertexCount, flags, stripIndexCount);
        Point3f[] coordinates = getCoordinates().getBuffer();
        int vertexIndex = 0;
        int index = 0;
        if (isClockwise) {
            for (j = contourLoops - 1; j >= 0; j--) {
                index = j % contourSize;
                triangleArray.setCoordinate(vertexIndex, coordinates[index]);
                triangleArray.setNormal(vertexIndex,
                    frame.getNormal(isClockwise));
                if (isTextured) {
                    triangleArray.setTextureCoordinate(0, vertexIndex,
                        textureCoordinates[index]);
                }
                vertexIndex++;
            }
        } else {
            for (j = 0; j < contourLoops; j++) {
                index = j % contourSize;
                triangleArray.setCoordinate(vertexIndex, coordinates[index]);
                triangleArray.setNormal(vertexIndex,
                    frame.getNormal(isClockwise));
                if (isTextured) {
                    triangleArray.setTextureCoordinate(0, vertexIndex,
                        textureCoordinates[index]);
                }
                vertexIndex++;
            }
        }
        return triangleArray;
    }

    /**
     * Description of the method.
     *
     * @param rotation Description of parameter.
     * @param translation Description of parameter.
     * @param scale Description of parameter.
     *
     * @return Description of the returned value.
     */
    public final Contour transform(Matrix3f rotation, Tuple3f translation,
        Tuple3f scale) {
        Contour contour = new Contour(size);
        coordinates.transform(rotation, translation, scale,
            contour.getCoordinates(), 0);
        Vector3f normalScale = new Vector3f(1, 1, 1);
        if (scale.x != 1 && scale.x > 0) {
            normalScale.x = 1.0f / scale.x;
        }
        if (scale.y != 1 && scale.y > 0) {
            normalScale.y = 1.0f / scale.y;
        }
        if (scale.z != 1 && scale.z > 0) {
            normalScale.z = 1.0f / scale.z;
        }
        normalScale.normalize();
        normals.transform(rotation, normalScale, contour.getNormals(), 0);
        contour.getNormals().normalize();
        return contour;
    }

    /**
     * Creates a circular contour using <code>divisionY</code> points.
     *
     * @param startAngle angle (in radians) that identifies with the initial point.
     * @param division Description of parameter.
     *
     * @return The <code>CircleContour</code> value.
     */
    public static final Contour getCircleContour(int division, float startAngle) {
        float dAngle = (float) Math.PI / division * 2;
        float angle = startAngle;
        Contour contour = new Contour(division);
        Vector3f vector = new Vector3f();
        for (int i = 0; i < division; i++) {
            vector.y = (float) Math.cos(angle);
            vector.x = (float) -Math.sin(angle);
            contour.getCoordinates().setAt(i, vector);
            contour.getNormals().setAt(i, vector);
            contour.getNormals().setAt(i + division, vector);
            angle += dAngle;
        }
        return contour;
    }

    /**
     * Creates a circular contour using <code>divisionY</code> points, but only to
     * certain division.
     *
     * @param startAngle angle (in radians) that identifies with the initial point.
     * @param division Description of parameter.
     * @param usedDivisions Number of divisions to produce.
     *
     * @return The <code>CircleContour</code> value.
     */
    public static final Contour getCircleContour(int division,
        float startAngle, int usedDivisions) {
        float dAngle = (float) Math.PI / division * 2;
        float angle = startAngle;
        Contour contour = new Contour(usedDivisions);
        Vector3f vector = new Vector3f();
        for (int i = 0; i < usedDivisions; i++) {
            vector.y = (float) Math.cos(angle);
            vector.x = (float) -Math.sin(angle);
            contour.getCoordinates().setAt(i, vector);
            contour.getNormals().setAt(i, vector);
            contour.getNormals().setAt(i + usedDivisions, vector);
            angle += dAngle;
        }
        return contour;
    }

    /**
     * Gets the <code>RectangleContour</code> attribute of the <code>Contour
     * </code>class.
     *
     * @return The <code>RectangleContour</code> value.
     */
    public static final Contour getRectangleContour() {
        Contour contour;
        contour = new Contour(4);
        contour.setClosed(true);
        contour.getCoordinates().setAt(0, -1, -1, 0);
        contour.getCoordinates().setAt(1, 1, -1, 0);
        contour.getCoordinates().setAt(2, 1, 1, 0);
        contour.getCoordinates().setAt(3, -1, 1, 0);
        contour.getNormals().setAt(0, -1, 0, 0);
        contour.getNormals().setAt(1, 0, -1, 0);
        contour.getNormals().setAt(2, 1, 0, 0);
        contour.getNormals().setAt(3, 0, 1, 0);
        contour.getNormals().setAt(4, 0, -1, 0);
        contour.getNormals().setAt(5, 1, 0, 0);
        contour.getNormals().setAt(6, 0, 1, 0);
        contour.getNormals().setAt(7, -1, 0, 0);
        return contour;
    }

    /**
     * Method description.
     *
     * @param textureCoordinate Parameter description.
     */
    public void setTextureCoordinate(TexCoord2f textureCoordinate) {
        this.texture = new TexCoordArray2f(size);
        this.texture.setUniform(textureCoordinate);
    }
}
