/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.vecmath;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Map;

import javax.media.j3d.GeometryArray;
import javax.media.j3d.Shape3D;
import javax.media.j3d.TriangleStripArray;
import javax.vecmath.Matrix3f;
import javax.vecmath.Point3f;
import javax.vecmath.TexCoord2f;
import javax.vecmath.Tuple3f;
import javax.vecmath.Vector3f;

/**
 * Surface generating class.
 *
 * @author Karsten Klein
 *
 * @created March 20, 2001
 */
public final class SweepingSurface {
    private Contour contour = null;
    private ArrayList frameContours = new ArrayList();
    private Map contourRegistry = new Hashtable();
    private Map contourMap = new Hashtable();
    private Point3f firstCenter = null;
    private Vector3f firstNormal = null;
    private Point3f lastCenter = null;
    private Vector3f lastNormal = null;
    private boolean isTextured = false;
    private ColorArray4f colors = null;

    /**
     * <code>SweepingSurface</code> contructor.
     *
     * @param contour Description of parameter.
     */
    public SweepingSurface(Contour contour) {
        this.contour = contour;
    }

    /**
     * Gets the <code>FrameCount</code> attribute of the <code>SweepingSurface
     * </code>object.
     *
     * @return The <code>FrameCount</code> value.
     */
    public int getFrameCount() {
        return frameContours.size();
    }

    /**
     * Method description.
     *
     * @param startFrame Parameter description.
     * @param endFrame Parameter description.
     *
     * @return Return description.
     */
    public final GeometryArray getGeometry(int startFrame, int endFrame) {
        int stripVertexCount = (endFrame - startFrame + 1) * 2;
        int stripTriangleCount = (endFrame - startFrame) * 2;
        int contourSize = contour.getSize();
        int triangleIndex = 0;
        int i;
        int j;
        int stripCount = contourSize;
        if (!contour.isClosed()) {
            stripCount--;
        }
        int[] stripIndexCount = new int[stripCount];
        for (i = 0; i < stripCount; i++) {
            stripIndexCount[i] = stripVertexCount;
        }
        int vertexCount = stripCount * stripVertexCount;
        int flags = GeometryArray.COORDINATES | GeometryArray.NORMALS;
        if (isTextured) {
            flags |= GeometryArray.TEXTURE_COORDINATE_2;
        }
        TriangleStripArray triangleArray =
            new TriangleStripArray(vertexCount, flags, stripIndexCount);
        Point3f[] coordinates;
        Vector3f[] normals;
        TexCoord2f[] texture = null;
        int vertexIndex = 0;
        int index = 0;
        Contour contour;
        Frame frame;
        Vector3f normal = null;
        for (j = 0; j < stripCount; j++) {
            for (i = startFrame; i <= endFrame; i++) {
                contour = (Contour) frameContours.get(i);
                coordinates = contour.getCoordinates().getBuffer();
                normals = contour.getNormals().getBuffer();
                index = (j + 1) % contourSize;
                triangleArray.setCoordinate(vertexIndex, coordinates[index]);
                triangleArray.setNormal(vertexIndex, normals[index]);
                if (isTextured) {
                    texture = contour.getTextureCoordinates().getBuffer();
                    triangleArray.setTextureCoordinate(0, vertexIndex,
                        texture[index]);
                }
                vertexIndex++;
                triangleArray.setCoordinate(vertexIndex, coordinates[j]);
                triangleArray.setNormal(vertexIndex, normals[j + contourSize]);
                if (isTextured) {
                    triangleArray.setTextureCoordinate(0, vertexIndex,
                        texture[j]);
                }
                vertexIndex++;
            }
        }
        return triangleArray;
    }

    /**
     * Gets the <code>Geometry</code> attribute of the <code>SweepingSurface
     * </code>object.
     *
     * @param startFrame Description of parameter.
     *
     * @return The <code>Geometry</code> value.
     */
    public GeometryArray getGeometry(int startFrame) {
        int endFrame = startFrame + 1;
        return getGeometry(startFrame, endFrame);
    }

    /**
     * Method description.
     *
     * @param startFrame Parameter description.
     * @param endFrame Parameter description.
     *
     * @return Return description.
     */
    public final Shape3D getFullGeometry(int startFrame, int endFrame) {
        return getFullGeometry(startFrame, endFrame, true);
    }

    /**
     * Method description.
     *
     * @param startFrame Parameter description.
     * @param endFrame Parameter description.
     * @param isClosed Parameter description.
     *
     * @return Return description.
     */
    public Shape3D getFullGeometry(int startFrame, int endFrame,
        boolean isClosed) {
        Shape3D shape = new Shape3D();
        int start = startFrame;
        int end = start;
        Contour contour;
        while (end < endFrame) {
            do {
                end++;
                contour = (Contour) frameContours.get(end);
            } while (contourRegistry.get(contour) == null && end < endFrame);
            shape.addGeometry(getGeometry(start, end));

            // draw front and back if registered
            if (isClosed) {
                contour = (Contour) frameContours.get(start);
                Frame frame = (Frame) contourMap.get(contour);
                if (frame != null) {
                    shape.addGeometry(contour.getGeometry(frame, false,
                            isTextured));
                }
                contour = (Contour) frameContours.get(end);
                frame = (Frame) contourMap.get(contour);
                if (frame != null) {
                    shape.addGeometry(contour.getGeometry(frame, true,
                            isTextured));
                }
            }
            end++;
            start = end;
        }
        return shape;
    }

    /**
     * Adds a feature to the <code>Frame</code> attribute of the
     * <code>SweepingSurface</code> object.
     *
     * @param rotation The feature to be added to the <code>Frame</code> attribute.
     * @param translation The feature to be added to the <code>Frame</code> attribute.
     * @param scale The feature to be added to the <code>Frame</code> attribute.
     * @param register The feature to be added to the <code>Frame</code> attribute.
     */
    public void addFrame(Matrix3f rotation, Tuple3f translation, Tuple3f scale,
        boolean register) {
        addFrame(rotation, translation, scale);
        Frame frame = new Frame();
        Vector3f normal = new Vector3f();
        rotation.getColumn(2, normal);
        frame.setNormal(normal);
        if (register) {
            contourRegistry.put(frameContours.get(frameContours.size() - 1),
                frame);
        }
        contourMap.put(frameContours.get(frameContours.size() - 1), frame);
    }

    /**
     * Method description.
     *
     * @param rotation Parameter description.
     * @param translation Parameter description.
     * @param scale Parameter description.
     * @param register Parameter description.
     * @param textureCoordinate Parameter description.
     */
    public void addFrame(Matrix3f rotation, Tuple3f translation, Tuple3f scale,
        boolean register, TexCoord2f textureCoordinate) {
        addFrame(rotation, translation, scale, textureCoordinate);
        Frame frame = new Frame();
        Vector3f normal = new Vector3f();
        rotation.getColumn(2, normal);
        frame.setNormal(normal);
        if (register) {
            contourRegistry.put(frameContours.get(frameContours.size() - 1),
                frame);
        }
        contourMap.put(frameContours.get(frameContours.size() - 1), frame);
    }

    /**
     * Adds a feature to the <code>Frame</code> attribute of the
     * <code>SweepingSurface</code> object.
     *
     * @param rotation The feature to be added to the <code>Frame</code> attribute.
     * @param translation The feature to be added to the <code>Frame</code> attribute.
     * @param scale The feature to be added to the <code>Frame</code> attribute.
     */
    private void addFrame(Matrix3f rotation, Tuple3f translation, Tuple3f scale) {
        frameContours.add(contour.transform(rotation, translation, scale));
        if (getFrameCount() == 1) {
            firstCenter = new Point3f(translation);
            lastCenter = new Point3f(translation);
            firstNormal = new Vector3f();
            rotation.getColumn(2, firstNormal);
            lastNormal = new Vector3f(firstNormal);
            firstNormal.negate();
        } else {
            lastCenter.set(translation);
            rotation.getColumn(2, lastNormal);
        }
    }

    private void addFrame(Matrix3f rotation, Tuple3f translation,
        Tuple3f scale, TexCoord2f textureCoordinate) {
        Contour newContour = contour.transform(rotation, translation, scale);
        newContour.setTextureCoordinate(textureCoordinate);
        frameContours.add(newContour);
        if (getFrameCount() == 1) {
            firstCenter = new Point3f(translation);
            lastCenter = new Point3f(translation);
            firstNormal = new Vector3f();
            rotation.getColumn(2, firstNormal);
            lastNormal = new Vector3f(firstNormal);
            firstNormal.negate();
        } else {
            lastCenter.set(translation);
            rotation.getColumn(2, lastNormal);
        }
    }

    /**
     * Adds a feature to the <code>Geometry</code> attribute of the
     * <code>SweepingSurface</code> object.
     *
     * @param startFrame The feature to be added to the <code>Geometry</code> attribute.
     * @param endFrame The feature to be added to the <code>Geometry</code> attribute.
     * @param shape The feature to be added to the <code>Geometry</code> attribute.
     */
    public void addGeometry(int startFrame, int endFrame, Shape3D shape) {
        GeometryArray geometryArray = getGeometry(startFrame, endFrame);
        shape.addGeometry(geometryArray);
    }

    /**
     * Method description.
     *
     * @param isTextured Parameter description.
     */
    public void setTextured(boolean isTextured) {
        this.isTextured = isTextured;
    }
}
