/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.vecmath;

import javax.vecmath.Matrix3f;
import javax.vecmath.Tuple3f;

/**
 * This class is a slight extention to the javax.vecmath package.
 *
 * @author Karsten Klein
 *
 * @created March 20, 2001
 */
public abstract class TupleArray3f {

    /** Array with all the tuples of the instance. */
    protected Tuple3f[] buffer = null;

    /**
     * Sets the first tuple of this instance.
     *
     * @param tuple0 new tuple value.
     */
    public final void set(Tuple3f tuple0) {
        buffer[0].set(tuple0);
    }

    /**
     * Sets the <code>Tuple</code> at the given index.
     *
     * @param index index of the tuple to change.
     * @param tuple new values of the tuple
     */
    public final void setAt(int index, Tuple3f tuple) {
        buffer[index].set(tuple);
    }

    /**
     * Sets the <code>Tuple</code> at the given index.
     *
     * @param index index of the tuple to change.
     * @param x new x value.
     * @param y new y value.
     * @param z new z value.
     */
    public final void setAt(int index, float x, float y, float z) {
        buffer[index].x = x;
        buffer[index].y = y;
        buffer[index].z = z;
    }

    /**
     * Sets the first two tuples of the array.
     *
     * @param tuple0 new tuple at index 0.
     * @param tuple1 new tuple at index 1.
     */
    public final void set(Tuple3f tuple0, Tuple3f tuple1) {
        buffer[0].set(tuple0);
        buffer[1].set(tuple1);
    }

    /**
     * Sets the first three tuples of the array.
     *
     * @param tuple0 new tuple at index 0.
     * @param tuple1 new tuple at index 1.
     * @param tuple2 new tuple at index 2.
     */
    public final void set(Tuple3f tuple0, Tuple3f tuple1, Tuple3f tuple2) {
        buffer[0].set(tuple0);
        buffer[1].set(tuple1);
        buffer[2].set(tuple2);
    }

    /**
     * Sets all the buffer in the array to tuple.
     *
     * @param tuple The new <code>Uniform</code> value.
     */
    public final void setUniform(Tuple3f tuple) {
        for (int i = 0; i < buffer.length; i++) {
            buffer[i].set(tuple);
        }
    }

    /**
     * Sets the <code>Buffer</code> attribute of the <code>TupleArray3f</code> object.
     *
     * @param buffer The new <code>Buffer</code> value.
     */
    public final void setBuffer(Tuple3f[] buffer) {
        for (int i = 0; i < buffer.length; i++) {
            this.buffer[i].set(buffer[i]);
        }
    }

    /**
     * Gets the <code>Size</code> attribute of the <code>TupleArray3f</code> object.
     *
     * @return The <code>Size</code> value.
     */
    public final int getSize() {
        return buffer.length;
    }

    /**
     * Applies the translation to each tuple in the list.
     *
     * @param translation Description of parameter.
     */
    public final void add(Tuple3f translation) {
        for (int i = 0; i < buffer.length; i++) {
            buffer[i].add(translation);
        }
    }

    /**
     * Applies the translation to each tuple in the list.
     *
     * @param tranformation Description of parameter.
     */
    public final void transform(Matrix3f tranformation) {
        for (int i = 0; i < buffer.length; i++) {
            tranformation.transform(buffer[i]);
        }
    }

    /**
     * Description of the method.
     *
     * @param tuple Description of parameter.
     */
    public final void scale(Tuple3f tuple) {
        for (int i = 0; i < buffer.length; i++) {
            buffer[i].x *= tuple.x;
            buffer[i].y *= tuple.y;
            buffer[i].z *= tuple.z;
        }
    }

    /**
     * Computes the minumum tuple from the provided tuple and the tuple at the specified
     * index.
     *
     * @param tuple provided tuple to compare.
     * @param index index of the array tuple to compare.
     */
    public final void min(int index, Tuple3f tuple) {
        buffer[index].x = Math.min(tuple.x, buffer[index].x);
        buffer[index].y = Math.min(tuple.y, buffer[index].y);
        buffer[index].z = Math.min(tuple.z, buffer[index].z);
    }

    /**
     * Computes the maximum tuple from the provided tuple and the tuple at the specified
     * index.
     *
     * @param tuple provided tuple to compare.
     * @param index index of the array tuple to compare.
     */
    public final void max(int index, Tuple3f tuple) {
        buffer[index].x = Math.max(tuple.x, buffer[index].x);
        buffer[index].y = Math.max(tuple.y, buffer[index].y);
        buffer[index].z = Math.max(tuple.z, buffer[index].z);
    }
}
