/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.vecmath;

import javax.vecmath.Matrix4f;
import javax.vecmath.Tuple4f;

/**
 * This class is a slight extention to the javax.vecmath package.
 *
 * @author Karsten Klein
 *
 * @created March 20, 2001
 */
public abstract class TupleArray4f {

    /** Array with all the tuples of the instance. */
    protected Tuple4f[] buffer = null;

    /**
     * Sets the first tuple of the array.
     *
     * @param tuple0 new tuple setting.
     */
    public void set(Tuple4f tuple0) {
        buffer[0].set(tuple0);
    }

    /**
     * Sets the <code>Tuple</code> at the given index.
     *
     * @param index index of the tuple to change.
     * @param tuple new values of the tuple
     */
    public void setAt(int index, Tuple4f tuple) {
        buffer[index].set(tuple);
    }

    /**
     * Sets the <code>Tuple</code> at the given index.
     *
     * @param index index of the tuple to change.
     * @param x new x value.
     * @param y new y value.
     * @param z new z value.
     * @param w new w value.
     */
    public void setAt(int index, float x, float y, float z, float w) {
        buffer[index].x = x;
        buffer[index].y = y;
        buffer[index].z = z;
        buffer[index].w = w;
    }

    /**
     * Sets the first two tuples of the array.
     *
     * @param tuple0 new tuple at index 0.
     * @param tuple1 new tuple at index 1.
     */
    public void set(Tuple4f tuple0, Tuple4f tuple1) {
        buffer[0].set(tuple0);
        buffer[1].set(tuple1);
    }

    /**
     * Sets the first three tuples of the array.
     *
     * @param tuple0 new tuple at index 0.
     * @param tuple1 new tuple at index 1.
     * @param tuple2 new tuple at index 2.
     */
    public void set(Tuple4f tuple0, Tuple4f tuple1, Tuple4f tuple2) {
        buffer[0].set(tuple0);
        buffer[1].set(tuple1);
        buffer[2].set(tuple0);
    }

    /**
     * Sets all the buffer in the array to tuple.
     *
     * @param tuple The new <code>Uniform</code> value
     */
    public void setUniform(Tuple4f tuple) {
        for (int i = 0; i < buffer.length; i++) {
            buffer[i].set(tuple);
        }
    }

    /**
     * Sets the <code>Buffer</code> attribute of the <code>TupleArray4f</code> object.
     *
     * @param buffer The new <code>Buffer</code> value
     */
    public void setBuffer(Tuple4f[] buffer) {
        for (int i = 0; i < buffer.length; i++) {
            this.buffer[i].set(buffer[i]);
        }
    }

    /**
     * Applies the translation to each tuple in the list.
     *
     * @param translation Description of parameter
     */
    public void add(Tuple4f translation) {
        for (int i = 0; i < buffer.length; i++) {
            buffer[i].add(translation);
        }
    }

    /**
     * Applies the translation to each tuple in the list.
     *
     * @param tranformation Description of parameter
     */
    public void transform(Matrix4f tranformation) {
        for (int i = 0; i < buffer.length; i++) {
            tranformation.transform(buffer[i]);
        }
    }

    /**
     * Scales all tuples in the array using the specified scaling tuple.
     *
     * @param tuple tuple containing the scaling factors in each direction.
     */
    public void scale(Tuple4f tuple) {
        for (int i = 0; i < buffer.length; i++) {
            buffer[i].x *= tuple.x;
            buffer[i].y *= tuple.y;
            buffer[i].z *= tuple.z;
            buffer[i].w *= tuple.w;
        }
    }
}
