/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.visualization;

import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;

import javax.vecmath.Color3f;
import javax.vecmath.Color4f;
import javax.vecmath.Tuple3f;
import javax.vecmath.Vector3f;

import org.srs3d.viewer.bioatlas.objects.Atom;

/**
 * @author Karsten Klein
 *
 * @created May 31, 2001
 */
public class AtomDistanceProperty implements ScalarProperty, ColorProperty,
    GradientProperty {

    /** Description of the field. */
    public int cacheHits = 0;

    /** Description of the field. */
    public int cacheMisses = 0;

    /** Description of the field. */
    public int cacheFilled = 0;

    /** Description of the field. */
    protected Subspace subspace = null;

    /** Description of the field. */
    protected Sphere sphere = new Sphere();

    /** Description of the field. */
    protected Collection volumes = new HashSet();

    //  private Collection atoms = null;
    private Color4f color = null;
    private Subspace cachedSubspace = null;

    /**
     * Sets the <code>subspace</code> attribute of the <code>AtomDistanceProperty</code>
     * object.
     *
     * @param subspace The new <code>subspace</code> value.
     */
    public void setSubspace(Subspace subspace) {
        this.subspace = subspace;
    }

    /**
     * Samples the volume at a given positon. This implementation searches the closest
     * atom in the volume set and returns the distance from position to atom hull;
     *
     * @param vector Pposition.
     *
     * @return The distance from the postion to the atom hull.
     */
    public float getValue(Tuple3f vector) {
        AtomVolume atomVolume = (AtomVolume) computeClosestVolume(vector, 2);
        if (atomVolume != null) {
            Atom atom = atomVolume.getAtom();
            Vector3f v = new Vector3f(atom.getCoordinate());
            v.sub(vector);
            float length = v.length();
            length -= atom.getVanDerWaalsRadius();
            return length;
        }
        return Float.MAX_VALUE;
    }

    /**
     * Maps the CPK color onto the sampling position.
     *
     * @param vector Position.
     *
     * @return Mapped color.
     */
    public Color4f getColor(Tuple3f vector) {
        AtomVolume volume = (AtomVolume) computeClosestVolume(vector, 2);
        if (volume != null) {
            Color3f color = volume.getAtom().getColor();
            return new Color4f(color.x, color.y, color.z, 1);
        }
        return new Color4f();
    }

    /**
     * Maps the grafient onto the specified position.
     *
     * @param vector Position.
     *
     * @return Mapped gradient.
     */
    public Vector3f getGradient(Tuple3f vector) {
        AtomVolume volume = (AtomVolume) computeClosestVolume(vector, 2);
        if (volume != null) {
            Vector3f v = new Vector3f(vector);
            v.sub(volume.getAtom().getCoordinate());
            return v;
        }
        return new Vector3f();
    }

    /**
     * Sets the <code>atoms</code> attribute of the <code>AtomDistanceProperty </code>
     * object. The atoms are converted to volumetric elements using the
     * <code>AtomVolume</code> class.
     *
     * @param atoms The new <code>atoms</code> value.
     */
    public void addAtoms(Collection atoms) {
        Iterator iterator = atoms.iterator();
        Atom atom;
        AtomVolume atomVolume;
        while (iterator.hasNext()) {
            atom = (Atom) iterator.next();
            atomVolume = new AtomVolume(atom, 0f);
            subspace.addElement(atomVolume);
        }
    }

    /**
     * Description of the method.
     */
    public void computeSubdivision() {
        subspace.subdivide(7);
    }

    /**
     * Gets the <code>cached</code> attribute of the <code>AtomDistanceProperty</code>
     * object.
     *
     * @param volume Description of parameter.
     *
     * @return The <code>cached</code> value.
     */
    protected boolean isCached(Volume volume) {
        if (false && cachedSubspace != null) {
            if (cachedSubspace.getVolume().includes(volume)) {
                return true;
            }
        }
        return false;
    }

    /**
     * Description of the method.
     *
     * @param volume Description of parameter.
     */
    protected void collectVolumes(Volume volume) {
        boolean cached = false;
        if (cachedSubspace != null) {
            if (cachedSubspace.getVolume().includes(volume)) {
                cacheHits++;
                volumes.clear();
                Subspace cacheCandidate =
                    cachedSubspace.collectCandidates(volume, volumes);
                if (cacheCandidate != null) {
                    cachedSubspace = cacheCandidate;
                    cacheFilled++;
                }
                cached = true;
            }
        }
        if (!cached) {
            volumes.clear();
            cachedSubspace = subspace.collectCandidates(volume, volumes);
            if (cachedSubspace != null) {
                cacheFilled++;
            }
            cacheMisses++;
        }
    }

    /**
     * Description of the method.
     *
     * @param vector Description of parameter.
     * @param radius Description of parameter.
     *
     * @return Description of the returned value.
     */
    protected Volume computeClosestVolume(Tuple3f vector, float radius) {
        sphere.setCoordinate(vector);
        sphere.setRadius(radius);
        collectVolumes(sphere);
        return computeClosestVolume(vector);
    }

    protected final Volume computeClosestVolume(Tuple3f vector) {
        return computeClosestVolume(vector, volumes);
    }

    /**
     * Description of the method.
     *
     * @param vector Description of parameter.
     *
     * @return Description of the returned value.
     */
    protected final Volume computeClosestVolume(Tuple3f vector,
        Collection volumes) {
        Vector3f v = new Vector3f();
        float min = Float.MAX_VALUE;
        float l = 0;
        Iterator iterator = volumes.iterator();
        Volume closestVolume = null;
        Volume volume;
        while (iterator.hasNext()) {
            volume = (Volume) iterator.next();
            l = volume.computeDistance(vector);
            if (l < min) {
                min = l;
                closestVolume = volume;
            }
        }
        return closestVolume;
    }

    /**
     * Method description.
     *
     * @return Return description.
     */
    public Subspace getSubspace() {
        return subspace;
    }
}
