/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.visualization;

import java.util.Collection;

import javax.vecmath.Tuple3f;
import javax.vecmath.Vector3f;

/**
 * Volumetric description of a concave cylinder.
 *
 * @author Karsten Klein
 *
 * @created May 31, 2001
 */
public class ConcaveCylinder extends Cylinder {
    private float radius = 0;
    private Vector3f offset = new Vector3f();

    /**
     * Sets the <code>radius</code> attribute of the <code>ConcaveCylinder</code> object.
     *
     * @param radius The new <code>radius</code> value.
     */
    public void setRadius(float radius) {
        this.radius = radius;
    }

    /**
     * Sets the <code>offset</code> attribute of the <code>ConcaveCylinder</code> object.
     *
     * @param offset The new <code>offset</code> value.
     */
    public void setOffset(Tuple3f offset) {
        this.offset.set(offset);
    }

    /**
     * Gets the <code>radius</code> attribute of the <code>ConcaveCylinder</code> object.
     *
     * @return The <code>radius</code> value.
     */
    public float getRadius() {
        return radius;
    }

    /**
     * Gets the <code>offset</code> attribute of the <code>ConcaveCylinder</code> object.
     *
     * @return The <code>offset</code> value.
     */
    public Vector3f getOffset() {
        return offset;
    }

    /**
     * Description of the method.
     *
     * @param volumes Description of parameter.
     */
    public void computeSubvolumes(Collection volumes) {
        throw new RuntimeException(this.toString() +
            ": missing implementation!");
    }

    /**
     * Description of the method.
     *
     * @param tuple Description of parameter.
     * @param epsilon Description of parameter.
     *
     * @return Description of the returned value.
     */
    public boolean intersects(Tuple3f tuple, float epsilon) {
        if (super.intersects(tuple, epsilon)) {
            return computeConcaveBodyDistance(tuple) < epsilon;
        }
        return false;
    }

    /**
     * Description of the method.
     *
     * @param volume Description of parameter.
     *
     * @return Description of the returned value.
     */
    public boolean includes(Volume volume) {
        throw new RuntimeException(this.toString() +
            ": missing implementation!");
    }

    /**
     * Description of the method.
     *
     * @param tuple Description of parameter.
     *
     * @return Description of the returned value.
     */
    public float computeDistance(Tuple3f tuple) {
        float cylinderDistance = super.computeDistance(tuple);
        return Math.max(computeConcaveBodyDistance(tuple), cylinderDistance);
    }

    /**
     * Description of the method.
     *
     * @param tuple Description of parameter.
     *
     * @return Description of the returned value.
     */
    public float computeConcaveBodyDistance(Tuple3f tuple) {
        Vector3f vector = computeProbeCoordinate(tuple);
        vector.sub(tuple);
        return getRadius() - vector.length();
    }

    /**
     * Description of the method.
     *
     * @param vector Description of parameter.
     *
     * @return Description of the returned value.
     */
    public Vector3f computeProbeCoordinate(Tuple3f vector) {

        // compute connection vector between the atoms centers
        Vector3f w = new Vector3f(getAxis());
        w.scale(offset.x);
        Vector3f orthogonal = new Vector3f(vector);
        orthogonal.sub(getCoordinateA());
        Vector3f projection = new Vector3f(getAxis());
        float projectionLength = orthogonal.dot(projection);
        if (projectionLength != 0) {
            projection.scale(projectionLength);
            orthogonal.sub(projection);
        }
        orthogonal.normalize();
        orthogonal.scale(offset.y);

        // this is the probe coordinate
        w.add(getCoordinateA());
        w.add(orthogonal);
        return w;
    }
}
