/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.visualization;

import java.util.Collection;

import javax.vecmath.Point3f;
import javax.vecmath.Tuple3f;
import javax.vecmath.Vector3f;

/**
 * Description of the class.
 *
 * @author Karsten Klein
 *
 * @created April 26, 2001
 */
public final class Cube {

    /** Description of the field. */
    public int hit = 0;

    /** Description of the field. */
    public int miss = 0;

    /** Description of the field. */
    public int pointHit = 0;

    /** Description of the field. */
    public int pointMiss = 0;
    private int idI = -1;
    private int idJ = -1;
    private int idK = -1;
    private Point3f origin = new Point3f();
    private Vector3f size = new Vector3f(1, 1, 1);
    private float[] values = new float[8];
    private int caseIndex = 0;
    private float threshold = 0.5f;
    private SampleManager sampleManager = new SampleManager();
    private PointManager pointManager = new PointManager();

    /**
     * Sets the <code>Threshold</code> attribute of the <code>Cube</code> object.
     *
     * @param threshold The new <code>Threshold</code> value.
     */
    public void setThreshold(float threshold) {
        this.threshold = threshold;
    }

    /**
     * Sets the <code>Origin</code> attribute of the <code>Cube</code> object.
     *
     * @param origin The new <code>Origin</code> value.
     */
    public void setOrigin(Tuple3f origin) {
        this.origin.set(origin);
    }

    /**
     * Sets the <code>Size</code> attribute of the <code>Cube</code> object.
     *
     * @param size The new <code>Size</code> value.
     */
    public void setSize(Tuple3f size) {
        this.size.set(size);
    }

    /**
     * Sets the <code>Id</code> attribute of the <code>Cube</code> object.
     *
     * @param idI The new <code>Id</code> value.
     * @param idJ The new <code>Id</code> value.
     * @param idK The new <code>Id</code> value.
     */
    public final void setId(int idI, int idJ, int idK) {
        this.idI = idI;
        this.idJ = idJ;
        this.idK = idK;
    }

    /**
     * Returns the <code>SampleManager</code> of the cube
     *
     * @return The <code>SampleManager</code>.
     */
    public final SampleManager getSampleManager() {
        return sampleManager;
    }

    /**
     * Gets the <code>PointManager</code> attribute of the <code>Cube</code> object.
     *
     * @return The <code>PointManager</code>.
     */
    public final PointManager getPointManager() {
        return pointManager;
    }

    /**
     * Sample the cubes corners.
     *
     * @param property <code>ScalarProperty</code> used for sampling.
     */
    public final void sample(ScalarProperty property) {
        int i;
        int j;
        int k;
        Sample sample;
        for (int index = 0; index < 8; index++) {
            i = (index & 1) == 0 ? this.idI : this.idI + 1;
            j = (index & 2) == 0 ? this.idJ : this.idJ + 1;
            k = (index & 4) == 0 ? this.idK : this.idK + 1;
            if (index != 7) {
                sample = sampleManager.getSample(i, j, k);
            } else {

                // here we are sure that it's a really new sample
                sample = null;

                // we avoid a miss
                miss--;
            }
            if (sample == null) {
                sample = new Sample();
                sample.setScalarValue(property.getValue(computeSamplePoint(
                            index)));
                sampleManager.addSample(i, j, k, sample);
                miss++;
            } else {
                hit++;
            }
            values[index] = sample.getScalarValue();
        }
    }

    /**
     * Description of the method.
     *
     * @return Description of the returned value.
     */
    public final int computeCaseIndex() {
        caseIndex = 0;
        if (values[0] >= threshold) {
            caseIndex |= 1;
        }
        if (values[1] >= threshold) {
            caseIndex |= 2;
        }
        if (values[2] >= threshold) {
            caseIndex |= 4;
        }
        if (values[3] >= threshold) {
            caseIndex |= 8;
        }
        if (values[4] >= threshold) {
            caseIndex |= 16;
        }
        if (values[5] >= threshold) {
            caseIndex |= 32;
        }
        if (values[6] >= threshold) {
            caseIndex |= 64;
        }
        if (values[7] >= threshold) {
            caseIndex |= 128;
        }
        return caseIndex;
    }

    /**
     * Adds a feature to the <code>Triangles</code> attribute of the <code>Cube
     * </code>object.
     *
     * @param triangles The feature to be added to the <code>Triangles</code> attribute.
     * @param filterProperty The <code>Triangles</code> object to be added.
     */
    public final void addTriangles(Collection triangles,
        FilterProperty filterProperty) {
        int[][] caseCode = Cases.caseCodes[caseIndex];
        for (int i = 0; i < caseCode.length; i++) {
            addTriangle(caseCode[i], triangles, filterProperty);
        }
    }

    /**
     * Gets the <code>CaseIndex</code> attribute of the <code>Cube</code> object.
     *
     * @return The <code>CaseIndex</code> value.
     */
    private final int getCaseIndex() {
        return caseIndex;
    }

    /**
     * Gets the <code>EdgeIntersection</code> attribute of the <code>Cube</code> object.
     *
     * @param edgeCode Description of parameter.
     *
     * @return The <code>EdgeIntersection</code> value.
     */
    private final Point3f getEdgeIntersection(int edgeCode) {
        int index0 = (edgeCode >> 3) & 7;
        int index1 = index0 | (edgeCode & 7);

        //    // tetrahedral approach
        //    int index1 = edgeCode & 7;
        return computePoint(index0, index1);
    }

    /**
     * Adds a feature to the <code>Triangle</code> attribute of the <code>Cube
     * </code>object.
     *
     * @param edgeCodes The feature to be added to the <code>Triangle</code> attribute.
     * @param triangles The feature to be added to the <code>Triangle</code> attribute.
     * @param filterProperty The <code>Triangle</code> object to be added.
     */
    private final void addTriangle(int[] edgeCodes, Collection triangles,
        FilterProperty filterProperty) {
        int edgeCode;
        int edgeId = 0;
        Point point;
        int i;
        int j;
        int k;
        Triangle triangle = new Triangle();
        boolean filtered = false;
        boolean pointFiltered;
        for (int index = 0; index < 3 && !filtered; index++) {
            edgeCode = edgeCodes[index];

            // this doesn't work with the tetrahedral approch; no point mapping possible
            i = (edgeCode & 8) == 0 ? this.idI : this.idI + 1;
            j = (edgeCode & 16) == 0 ? this.idJ : this.idJ + 1;
            k = (edgeCode & 32) == 0 ? this.idK : this.idK + 1;
            edgeId = edgeCode & 3;

            //      // tetrahedral approach
            //      i = ( edgeCode & 64 ) == 0 ? this.idI : this.idI + 1;
            //      j = ( edgeCode & 128 ) == 0 ? this.idJ : this.idJ + 1;
            //      k = ( edgeCode & 256 ) == 0 ? this.idK : this.idK + 1;
            //      edgeId = ( edgeCode >> 9 ) & 15;
            point = pointManager.getPoint(i, j, k, edgeId);
            if (point == null) {
                point = new Point();
                point.setCoordinate(getEdgeIntersection(edgeCode));
                pointManager.addPoint(i, j, k, edgeId, point);
                pointMiss++;
            } else {
                pointHit++;
            }
            triangle.getPoints()[index] = point;
            pointFiltered = filterProperty.filter(point.getCoordinate());
            filtered |= pointFiltered;
        }
        if (!filtered) {
            triangle.getPoints()[0].getTriangles().add(triangle);
            triangle.getPoints()[1].getTriangles().add(triangle);
            triangle.getPoints()[2].getTriangles().add(triangle);
            triangles.add(triangle);
        }
    }

    /**
     * Computes the finally used point on the incorporate edge by linear interpolation of
     * the scalar values.
     *
     * @param index0 Corner index of 1st point.
     * @param index1 Corner index of 3nd point.
     *
     * @return Interpolated point in between the specified ones.
     */
    private final Point3f computePoint(int index0, int index1) {
        Point3f point0 = computeSamplePoint(index0);
        Point3f point1 = computeSamplePoint(index1);
        float slope = values[index1] - values[index0];
        float a;
        if (slope == 0) {
            a = 0.5f;
        } else {
            a = threshold - values[index0];
            a /= slope;
        }
        point1.sub(point0);
        point1.scale(a);
        point0.add(point1);
        return point0;
    }

    /**
     * Computes the coordinates of the sample point.
     *
     * @param index Corner index in the cube.
     *
     * @return Computed coordinate.
     */
    private final Point3f computeSamplePoint(int index) {
        Point3f coordinate = new Point3f(origin);
        if ((index & 1) != 0) {
            coordinate.x += size.x;
        }
        if ((index & 2) != 0) {
            coordinate.y += size.y;
        }
        if ((index & 4) != 0) {
            coordinate.z += size.z;
        }
        return coordinate;
    }
}
