/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.visualization;

import java.util.Collection;

import javax.vecmath.Point3f;
import javax.vecmath.Tuple3f;
import javax.vecmath.Vector3f;

/**
 * Volumetric description of a cubeoid.
 *
 * @author Karsten Klein
 *
 * @created May 31, 2001
 */
public final class Cuboid implements Volume {
    private Point3f coordinate = new Point3f();

    /** Extend is half of the cube length */
    private float extend = 0;

    /**
     * Sets the <code>coordinate</code> attribute of the <code>Sphere</code> object.
     *
     * @param coordinate The new <code>coordinate</code> value.
     */
    public final void setCoordinate(Tuple3f coordinate) {
        this.coordinate.set(coordinate);
    }

    /**
     * Sets the <code>extend</code> attribute of the <code>Cuboid</code> object. Note
     * that the extend is the half the length of one side of the cube .
     *
     * @param extend The new <code>extend</code> value.
     */
    public final void setExtend(float extend) {
        this.extend = extend;
    }

    /**
     * Gets the <code>coordinate</code> attribute of the <code>Sphere</code> object.
     *
     * @return The <code>coordinate</code> value.
     */
    public final Point3f getCoordinate() {
        return coordinate;
    }

    /**
     * Gets the <code>extend</code> attribute of the <code>Cuboid</code> object.
     *
     * @return The <code>extend</code> value.
     */
    public final float getExtend() {
        return extend;
    }

    /**
     * Description of the method.
     *
     * @return Description of the returned value.
     */
    public float computeDiameter() {
        return 3.465f * getExtend(); // 2 * sqrt( 3 ) * extend
    }

    /**
     * Description of the method.
     *
     * @param volumes Description of parameter.
     */
    public void computeSubvolumes(Collection volumes) {
        Cuboid volume;
        Point3f point = new Point3f();
        float extend = 0.5f * getExtend();
        float newExtend = 0.501f * getExtend();
        for (int i = 0; i < 8; i++) {
            volume = new Cuboid();
            point.set(getCoordinate());
            point.x += (i & 1) == 0 ? -extend : extend;
            point.y += (i & 2) == 0 ? -extend : extend;
            point.z += (i & 4) == 0 ? -extend : extend;
            volume.setCoordinate(point);
            volume.setExtend(newExtend);
            volumes.add(volume);
        }
    }

    /**
     * Description of the method.
     *
     * @param tuple Description of parameter.
     * @param epsilon Description of parameter.
     *
     * @return Description of the returned value.
     */
    public boolean intersects(Tuple3f tuple, float epsilon) {
        epsilon += extend;
        if (Math.abs(coordinate.x - tuple.x) < epsilon) {
            if (Math.abs(coordinate.y - tuple.y) < epsilon) {
                if (Math.abs(coordinate.z - tuple.z) < epsilon) {
                    return true;
                }
            }
        }
        return false;
    }

    //  // General implementation for oriented cuboid
    //  public boolean intersects_General( Tuple3f tuple, float epsilon ) {
    //
    //    Vector3f coordinate = new Vector3f( tuple );
    //    Vector3f normal = new Vector3f();
    //    Vector3f anchor = new Vector3f( getCoordinate() );
    //
    //    float distance;
    //
    //    epsilon += extend;
    //
    //    normal.x = 1;
    //    distance = Cylinder.computePlaneDistance( anchor, normal, coordinate );
    //    distance = ( float ) Math.abs( distance );
    //
    //    if ( distance < epsilon ) {
    //
    //      normal.x = 0;
    //      normal.y = 1;
    //      distance = Cylinder.computePlaneDistance( anchor, normal, coordinate );
    //      distance = ( float ) Math.abs( distance );
    //
    //      if ( distance < epsilon ) {
    //
    //        normal.y = 0;
    //        normal.z = 1;
    //        distance = Cylinder.computePlaneDistance( anchor, normal, coordinate );
    //        distance = ( float ) Math.abs( distance );
    //
    //        return distance < epsilon;
    //      }
    //
    //    }
    //
    //    return false;
    //  }

    /**
     * Description of the method.
     *
     * @param volume Description of parameter.
     *
     * @return Description of the returned value.
     */
    public boolean intersects(Volume volume) {

        // coarse implementation
        throw new RuntimeException(this + ": missing implementation!");

        //    return volume.intersects( getCoordinate(), 1.733f * getExtend() );
    }

    /**
     * Description of the method.
     *
     * @param volume Description of parameter.
     *
     * @return Description of the returned value.
     */
    public boolean includes(Volume volume) {
        Sphere sphere = (Sphere) volume;
        Vector3f vector = new Vector3f();

        // compute direction vector
        vector.set(sphere.getCoordinate());
        vector.sub(getCoordinate());
        float length = vector.length();
        return (length + sphere.getRadius() < getExtend());
    }

    /**
     * Description of the method.
     *
     * @param tuple Description of parameter.
     *
     * @return Description of the returned value.
     */
    public float computeDistance(Tuple3f tuple) {
        Vector3f coordinate = new Vector3f(tuple);
        Vector3f normal = new Vector3f();
        Vector3f anchor = new Vector3f(getCoordinate());
        normal.x = 1;
        float distanceX =
            Cylinder.computePlaneDistance(anchor, normal, coordinate);
        distanceX = (float) Math.abs(distanceX) - getExtend();
        normal.x = 0;
        normal.y = 1;
        float distanceY =
            Cylinder.computePlaneDistance(anchor, normal, coordinate);
        distanceY = (float) Math.abs(distanceY) - getExtend();
        normal.y = 0;
        normal.z = 1;
        float distanceZ =
            Cylinder.computePlaneDistance(anchor, normal, coordinate);
        distanceZ = (float) Math.abs(distanceZ) - getExtend();
        return Math.max(distanceX, Math.max(distanceY, distanceZ));
    }

    /**
     * Description of the method.
     *
     * @return Description of the returned value.
     */
    public String toString() {
        return "Cuboid volume definition: center " + getCoordinate() +
        "; extend " + getExtend();
    }
}
