/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.visualization;

import java.util.Collection;

import javax.vecmath.Point3f;
import javax.vecmath.Tuple3f;
import javax.vecmath.Vector3f;

/**
 * Volumetric description of a cylinder (generalized to a cone section).
 *
 * @author Karsten Klein
 *
 * @created May 31, 2001
 */
public class Cylinder implements Volume {
    private Point3f coordinateA = new Point3f();
    private Point3f coordinateB = new Point3f();
    private float radiusA = 0;
    private float radiusB = 0;
    private float length = -1;

    // some variable for enhancing the computations
    private transient Vector3f axis = null;
    private transient Vector3f anchor = new Vector3f();
    private transient Vector3f normal = new Vector3f();
    private transient Vector3f coordinate = new Vector3f();
    private transient float distance;
    private transient float diameter = -1;
    private transient Point3f center = null;

    /**
     * Sets the <code>coordinateA</code> attribute of the <code>Cylinder</code> object.
     *
     * @param coordinateA The new <code>coordinateA</code> value.
     */
    public void setCoordinateA(Tuple3f coordinateA) {
        axis = null;
        length = -1;
        diameter = -1;
        this.coordinateA.set(coordinateA);
    }

    /**
     * Sets the <code>coordinateB</code> attribute of the <code>Cylinder</code> object.
     *
     * @param coordinateB The new <code>coordinateB</code> value.
     */
    public void setCoordinateB(Tuple3f coordinateB) {
        axis = null;
        length = -1;
        diameter = -1;
        this.coordinateB.set(coordinateB);
    }

    /**
     * Sets the <code>radiusA</code> attribute of the <code>Cylinder</code> object.
     *
     * @param radiusA The new <code>radiusA</code> value.
     */
    public void setRadiusA(float radiusA) {
        this.radiusA = radiusA;
        diameter = -1;
    }

    /**
     * Sets the <code>radiusB</code> attribute of the <code>Cylinder</code> object.
     *
     * @param radiusB The new <code>radiusB</code> value.
     */
    public void setRadiusB(float radiusB) {
        this.radiusB = radiusB;
        diameter = -1;
    }

    /**
     * Gets the <code>coordinateA</code> attribute of the <code>Cylinder</code> object.
     *
     * @return The <code>coordinateA</code> value.
     */
    public Point3f getCoordinateA() {
        return coordinateA;
    }

    /**
     * Gets the <code>coordinateB</code> attribute of the <code>Cylinder</code> object.
     *
     * @return The <code>coordinateB</code> value.
     */
    public Point3f getCoordinateB() {
        return coordinateB;
    }

    /**
     * Gets the <code>radiusA</code> attribute of the <code>Cylinder</code> object.
     *
     * @return The <code>radiusA</code> value.
     */
    public float getRadiusA() {
        return radiusA;
    }

    /**
     * Gets the <code>radiusB</code> attribute of the <code>Cylinder</code> object.
     *
     * @return The <code>radiusB</code> value.
     */
    public float getRadiusB() {
        return radiusB;
    }

    /**
     * Gets the <code>axis</code> attribute of the <code>Cylinder</code> object.
     *
     * @return The <code>axis</code> value.
     */
    public Vector3f getAxis() {
        if (axis == null) {
            return computeAxis();
        }
        return axis;
    }

    /**
     * Gets the <code>length</code> attribute of the <code>Cylinder</code> object.
     *
     * @return The <code>length</code> value.
     */
    public float getLength() {
        if (length == -1) {
            computeAxis();
        }
        return length;
    }

    /**
     * Computes and returns the normalized axis vector of the cylinder.
     *
     * @return Normalized axis vector pointing from A to B.
     */
    public Vector3f computeAxis() {
        axis = new Vector3f(getCoordinateB());
        axis.sub(getCoordinateA());
        length = axis.length();
        axis.normalize();
        return axis;
    }

    /**
     * Description of the method.
     *
     * @param volumes Description of parameter.
     */
    public void computeSubvolumes(Collection volumes) {
        throw new RuntimeException(this.toString() +
            ": missing implementation!");
    }

    /**
     * Computes the distance of the specified tuple from the cylinder body. Checked.
     *
     * @param coordinate Description of parameter.
     *
     * @return Description of the returned value.
     */
    public float computeBodyDistance(Tuple3f coordinate) {
        Vector3f orthogonal = new Vector3f(coordinate);
        orthogonal.sub(getCoordinateA());
        float cos = axis.dot(orthogonal);
        Vector3f projection = new Vector3f(axis);
        projection.scale(cos);

        // take out component in axis direction, the remainder is orthogonal
        // to the axis
        orthogonal.sub(projection);

        // interpolate radius (at projection end position)
        float radius = getRadiusB() - getRadiusA();
        radius *= cos / getLength();
        radius += getRadiusA();
        return orthogonal.length() - radius;
    }

    /**
     * Description of the method.
     *
     * @param tuple Description of parameter.
     * @param epsilon Description of parameter.
     *
     * @return Description of the returned value.
     */
    public boolean intersects(Tuple3f tuple, float epsilon) {
        coordinate.set(tuple);
        normal.set(getAxis());
        anchor.set(getCoordinateB());
        float distance = computePlaneDistance(anchor, getAxis(), coordinate);
        if (distance < epsilon) {
            anchor.set(getCoordinateA());
            distance = -computePlaneDistance(anchor, getAxis(), coordinate);
            if (distance < epsilon) {
                return computeBodyDistance(coordinate) < epsilon;
            }
        }
        return false;
    }

    /**
     * Description of the method.
     *
     * @param volume Description of parameter.
     *
     * @return Description of the returned value.
     */
    public boolean intersects(Volume volume) {
        return volume.intersects(getCenter(), getDiameter() * 0.5f);
    }

    /**
     * Method description.
     *
     * @return Return description.
     */
    public float getDiameter_() {
        if (diameter == -1) {
            float maxRadius = Math.max(radiusA, radiusB);
            float d = getLength() * 0.5f;
            diameter = 2 * (float) Math.sqrt(maxRadius * maxRadius + d * d);
        }
        return diameter;
    }

    /**
     * Computes and returns the minimal diameter of the volume. In general a Sphere with
     * the this diameter around the center of mass of the volume should include the
     * volume.
     *
     * @return Computed diameter.
     */
    public float getDiameter() {
        if (diameter == -1) {
            computeDiameter();
        }
        return diameter;
    }

    /**
     * Method description.
     *
     * @return Return description.
     */
    public float computeDiameter() {
        float l = getLength();
        float radiusBSquared = getRadiusB();
        radiusBSquared *= radiusBSquared;
        float radiusASquared = getRadiusA();
        radiusASquared *= radiusASquared;
        diameter = radiusBSquared - radiusASquared - l * l;
        diameter /= 2 * l;
        diameter += radiusBSquared;
        diameter = 2.0f * (float) Math.sqrt(diameter);
        return diameter;
    }

    /**
     * Description of the method.
     *
     * @param volume Description of parameter.
     *
     * @return Description of the returned value.
     */
    public boolean includes(Volume volume) {
        throw new RuntimeException(this.toString() +
            ": missing implementation!");
    }

    /**
     * Description of the method.
     *
     * @param tuple Description of parameter.
     *
     * @return Description of the returned value.
     */
    public float computeDistance(Tuple3f tuple) {
        coordinate.set(tuple);
        normal.set(getAxis());
        anchor.set(getCoordinateB());
        float distancePlaneB =
            computePlaneDistance(anchor, getAxis(), coordinate);
        anchor.set(getCoordinateA());
        float distancePlaneA =
            -computePlaneDistance(anchor, getAxis(), coordinate);
        float distanceBody = computeBodyDistance(coordinate);
        return Math.max(distancePlaneA, Math.max(distancePlaneB, distanceBody));
    }

    /**
     * Method to compute the distance of a coordinate from a plane with given anchor and
     * normal vector.
     *
     * @param anchor Plane anchor.
     * @param normal Plane normal.
     * @param coordinate Coordinate.
     *
     * @return Distance of the coordinate from the plane (euclidean).
     */
    public static float computePlaneDistance(Vector3f anchor, Vector3f normal,
        Vector3f coordinate) {
        return normal.dot(coordinate) - normal.dot(anchor);
    }

    /**
     * Method description.
     *
     * @return Return description.
     */
    public Point3f getCenter() {
        if (center == null) {
            center = new Point3f(getAxis());
            center.scale(0.5f);
            center.add(getCoordinateA());
        }
        return center;
    }
}
