/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.visualization;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;

import javax.vecmath.Point3f;
import javax.vecmath.Tuple3f;
import javax.vecmath.Vector3f;

import org.srs3d.viewer.objects.Computation;

/**
 * Description of the class.
 *
 * @author Karsten Klein
 *
 * @created April 26, 2001
 */
public class MarchingCube implements Computation {
    private Collection triangles = null;
    private Collection points = null;

    // the value i = -1 is used for teminating the marching vube finished
    private transient int i = 0;
    private transient int j;
    private transient int k;
    private transient int subdivisions;
    private transient boolean isCanceled = false;
    private transient boolean isVerbose = false;

    /**
     * Gets the <code>triangles</code> attribute of the <code>MarchingCube</code> object.
     *
     * @return The <code>triangles</code> value.
     */
    public Collection getTriangles() {
        return triangles;
    }

    /**
     * Description of the method.
     *
     * @param origin Description of parameter.
     * @param size Description of parameter.
     * @param threshold Description of parameter.
     * @param subdivisions Description of parameter.
     * @param property Description of parameter.
     * @param filterProperty Description of parameter.
     */
    public void createSurface(Tuple3f origin, Tuple3f size, float threshold,
        int subdivisions, ScalarProperty property, FilterProperty filterProperty) {
        this.subdivisions = subdivisions;
        triangles = new HashSet();
        points = new HashSet();
        Vector3f cubeSize = new Vector3f(size);
        cubeSize.scale(1.0f / (subdivisions + 1));
        Point3f cubeOrigin = new Point3f();
        Cube cube = new Cube();
        cube.setSize(cubeSize);
        cube.setThreshold(threshold);
        cubeOrigin.x = origin.x;
        SampleManager sampleManager = cube.getSampleManager();
        PointManager pointManager = cube.getPointManager();
        sampleManager.setCapacity(subdivisions + 2);
        pointManager.setCapacity(subdivisions + 2);
        subdivisions++;
        for (i = 0; i < subdivisions && !isCanceled; i++) {
            cubeOrigin.y = origin.y;
            for (j = 0; j < subdivisions; j++) {
                cubeOrigin.z = origin.z;
                for (k = 0; k < subdivisions; k++) {

                    // update prepare cube
                    cube.setOrigin(cubeOrigin);
                    cube.setId(i, j, k);

                    // sample cube and add triangles to triangle list
                    cube.sample(property);
                    cube.computeCaseIndex();
                    cube.addTriangles(triangles, filterProperty);
                    cubeOrigin.z += cubeSize.z;

                    // drop the sample in the lower left front corner
                    sampleManager.clear(i, j, k);
                }
                cubeOrigin.y += cubeSize.y;
            }
            cubeOrigin.x += cubeSize.x;

            // drop all samples in i section of x subdivision
            sampleManager.clear(i);

            // drop all point samples assoicated with edges to and between i corners
            pointManager.clear(i - 1);
        }
        i = -1;
        cube.getSampleManager().clear();
        cube.getPointManager().clear();

        // collect all point covered by the triangles
        collectPoints();

        // grid collapse point (100 means two digits)
        pointCollapse(100);

        // identify points (needed for smotthing)
        identifyPoints(0.2f);

        // remove triangles that are degenerated by the sideeffects of the upper two
        // calls
        removeCollapsedTriangles();
    }

    /**
     * Method description.
     *
     * @param iterations Parameter description.
     */
    public void smoothNormals(int iterations) {
        Iterator iterator = points.iterator();
        Point point;
        Vector3f normal;
        for (int i = 0; i < iterations; i++) {
            iterator = points.iterator();
            while (iterator.hasNext()) {
                point = (Point) iterator.next();
                normal = point.computeAvarageNormal();
                if (!Float.isNaN(normal.lengthSquared())) {
                    point.setNormal(normal);
                }
            }
        }
    }

    /**
     * Method description.
     *
     * @param iterations Parameter description.
     */
    public void smoothColor(int iterations) {
        Iterator iterator = points.iterator();
        Point point;
        for (i = 0; i < iterations; i++) {
            iterator = points.iterator();
            while (iterator.hasNext()) {
                point = (Point) iterator.next();
                point.setColor(point.computeAvarageColor());
            }
        }
    }

    /**
     * Description of the method.
     */
    public void computeDefaultNormals() {
        Iterator iterator = triangles.iterator();
        Triangle triangle;
        Vector3f normal;

        // create normals at each vertex
        while (iterator.hasNext()) {
            triangle = (Triangle) iterator.next();
            normal = triangle.computeOrthogonal();
            triangle.modifyNormals(normal);
        }

        // normalize all normals
        iterator = triangles.iterator();
        while (iterator.hasNext()) {
            triangle = (Triangle) iterator.next();
            triangle.normalizeNormals();
        }
    }

    /**
     * Description of the method.
     *
     * @param colorProperty Description of parameter.
     */
    public void mapColors(ColorProperty colorProperty) {
        Iterator iterator = points.iterator();
        Point point;
        while (iterator.hasNext()) {
            point = (Point) iterator.next();
            point.setColor(colorProperty.getColor(point.getCoordinate()));
        }
    }

    /**
     * Description of the method.
     *
     * @param gradientProperty Description of parameter.
     */
    public void mapGradient(GradientProperty gradientProperty) {
        Iterator iterator = points.iterator();
        Point point;
        while (iterator.hasNext()) {
            point = (Point) iterator.next();
            point.setNormal(gradientProperty.getGradient(
                    new Vector3f(point.getCoordinate())));
            point.getNormal().normalize();
        }
    }

    /**
     * Description of the method.
     */
    private void collectPoints() {
        Iterator iterator = triangles.iterator();
        Triangle triangle;
        while (iterator.hasNext()) {
            triangle = (Triangle) iterator.next();
            points.add(triangle.getPoints()[0]);
            points.add(triangle.getPoints()[1]);
            points.add(triangle.getPoints()[2]);
        }
    }

    //    /**
    //     *  The main program for the <code>MarchingCube</code> class.
    //     *
    //     * @param  args  The command line arguments.
    //     */
    //    public static void main( String[] args ) {
    //
    //      Point3f origin = new Point3f( -20, -20, -20 );
    //      Vector3f size = new Vector3f( 40, 40, 40 );
    //
    //      MarchingCube mc = new MarchingCube();
    //
    //      ScalarProperty testProperty = new DistanceProperty();
    //  //    testProperty = new QuadricProperty();
    //  //    ( (QuadricProperty) testProperty ).setFactor( new Vector3f( 1, 2, -1 ) );
    //      mc.createSurface( origin, size, 6.5f, 40, testProperty, null );
    //
    //    }
    public int getProgress() {
        if (subdivisions > 0) {
            return 100 * i / subdivisions;
        }

        // happens when not initialized yet
        return 0;
    }

    /**
     * Method description.
     *
     * @return Return description.
     */
    public boolean canCancel() {
        return true;
    }

    /**
     * Method description.
     */
    public void cancel() {
        isCanceled = true;
    }

    /**
     * Method description.
     *
     * @return Return description.
     */
    public String getDescription() {
        return "Computing surface...";
    }

    /**
     * Method description.
     *
     * @return Return description.
     */
    public boolean hasFinished() {
        return i != -1;
    }

    /**
     * Method description.
     *
     * @param factor Parameter description.
     */
    public void pointCollapse(int factor) {
        Iterator iterator = points.iterator();
        Point point;
        while (iterator.hasNext()) {
            point = (Point) iterator.next();
            point.collapse(factor);
        }
    }

    /**
     * Description of the method.
     */
    public void removeCollapsedTriangles() {
        Iterator iterator = triangles.iterator();
        Triangle triangle;
        int counter = 0;
        int progress = 0;
        while (iterator.hasNext()) {
            triangle = (Triangle) iterator.next();
            if (triangle.isCollapsed()) {
                counter++;
                iterator.remove();
            }
        }
    }

    /**
     * Method description.
     *
     * @param epsilon Parameter description.
     */
    public void identifyPoints(float epsilon) {
        ArrayList points = new ArrayList(this.points);
        Point point0;
        Point point1;
        int counter = 0;
        for (int i = 0; i < points.size(); i++) {
            point0 = (Point) points.get(i);
            for (int j = i + 1; j < points.size(); j++) {
                point1 = (Point) points.get(j);
                if (point0.isIdentical(point1, epsilon)) {
                    point1.replaceBy(point0);
                    counter++;
                    points.remove(point1);
                    this.points.remove(point1);
                }
            }
        }
    }
}
