/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.visualization;

import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;

import javax.vecmath.Color4f;
import javax.vecmath.Point3f;
import javax.vecmath.Vector3f;

/**
 * Description of the class.
 *
 * @author Karsten Klein
 *
 * @created September 10, 2001
 */
public class Point {
    private Point3f coordinate = null;
    private Vector3f normal = null;
    private Color4f color = null;
    private Collection triangles = new HashSet();

    /**
     * Sets the <code>coordinate</code> attribute of the <code>Point</code> object.
     *
     * @param coordinate The new <code>coordinate</code> value.
     */
    public void setCoordinate(Point3f coordinate) {
        this.coordinate = coordinate;
    }

    /**
     * Sets the <code>normal</code> attribute of the <code>Point</code> object.
     *
     * @param normal The new <code>normal</code> value.
     */
    public void setNormal(Vector3f normal) {
        this.normal = normal;
    }

    /**
     * Sets the <code>color</code> attribute of the <code>Point</code> object.
     *
     * @param color The new <code>color</code> value.
     */
    public void setColor(Color4f color) {
        this.color = color;
    }

    /**
     * Gets the <code>coordinate</code> attribute of the <code>Point</code> object.
     *
     * @return The <code>coordinate</code> value.
     */
    public Point3f getCoordinate() {
        return coordinate;
    }

    /**
     * Gets the <code>normal</code> attribute of the <code>Point</code> object.
     *
     * @return The <code>normal</code> value.
     */
    public Vector3f getNormal() {
        return normal;
    }

    /**
     * Gets the <code>color</code> attribute of the <code>Point</code> object.
     *
     * @return The <code>color</code> value.
     */
    public Color4f getColor() {
        return color;
    }

    /**
     * Gets the <code>triangles</code> attribute of the <code>Point</code> object.
     *
     * @return The <code>triangles</code> value.
     */
    public Collection getTriangles() {
        return triangles;
    }

    /**
     * Method description.
     *
     * @return Return description.
     */
    public Vector3f computeAvarageNormal() {
        Iterator iterator = triangles.iterator();
        Triangle triangle;
        Vector3f avarageNormal = new Vector3f();
        while (iterator.hasNext()) {
            triangle = (Triangle) iterator.next();
            avarageNormal.add(triangle.computeAvarageNormal());
        }
        avarageNormal.normalize();
        float lengthSquared = avarageNormal.lengthSquared();
        return avarageNormal;
    }

    /**
     * Method description.
     *
     * @return Return description.
     */
    public Color4f computeAvarageColor() {
        Iterator iterator = triangles.iterator();
        Triangle triangle;
        Color4f avarageColor = new Color4f();
        int counter = 0;
        while (iterator.hasNext()) {
            triangle = (Triangle) iterator.next();
            avarageColor.add(triangle.computeAvarageColor());
            counter++;
        }
        if (counter > 0) {
            avarageColor.scale(1.0f / counter);
        }
        return avarageColor;
    }

    /**
     * Method description.
     *
     * @param point Parameter description.
     * @param epsilon Parameter description.
     *
     * @return Return description.
     */
    public boolean isIdentical(Point point, float epsilon) {
        if (point == this) {
            return true;
        }
        return getCoordinate().epsilonEquals(point.getCoordinate(), epsilon);
    }

    /**
     * Method description.
     *
     * @param p Parameter description.
     */
    public void replaceBy(Point p) {
        if (this != p) {
            Iterator iterator = triangles.iterator();
            Triangle triangle;
            while (iterator.hasNext()) {
                triangle = (Triangle) iterator.next();
                triangle.collapse(this, p);
            }
        }
    }

    /**
     * Method description.
     *
     * @param factor Parameter description.
     */
    public final void collapse(int factor) {
        Point3f point = getCoordinate();
        point.scale(factor);
        point.x = (float) (Math.round(point.x)) / factor;
        point.y = (float) (Math.round(point.y)) / factor;
        point.z = (float) (Math.round(point.z)) / factor;
    }
}
