/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.visualization;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;

import javax.vecmath.Tuple3f;
import javax.vecmath.Vector3f;

import org.srs3d.viewer.bioatlas.objects.Atom;

/**
 * @author Karsten Klein
 *
 * @created May 31, 2001
 */
public class ProbeDistanceProperty extends AtomDistanceProperty {
    public float probeRadius = 1.5f;
    private Subspace cylinderSubspace = null;
    private Collection cylinderVolumes = new HashSet();

    /**
     * Method description.
     *
     * @param atoms Parameter description.
     */
    public void addAtoms(Collection atoms) {
        super.addAtoms(atoms);
        Subspace atomSubspace = getSubspace();
        cylinderSubspace = new Subspace(subspace.getVolume());
        Atom atomA;
        Atom atomB;
        ConcaveCylinder cylinder;
        ArrayList atomVector = new ArrayList(atoms);
        int counter = 0;
        int i;
        int j;
        for (i = 0; i < atomVector.size(); i++) {
            atomA = (Atom) atomVector.get(i);
            for (j = i + 1; j < atomVector.size(); j++) {
                atomB = (Atom) atomVector.get(j);
                cylinder = computeConcaveCylinder(atomA, atomB);
                if (cylinder != null) {
                    cylinderSubspace.addElement(cylinder);
                    counter++;
                }
            }
        }
        cylinderSubspace.subdivide(5);
    }

    /**
     * Gets the <code>value</code> attribute of the <code>AtomDistanceProperty </code>
     * object.
     *
     * @param vector Description of parameter.
     *
     * @return The <code>value</code> value.
     */
    public float getValue(Tuple3f vector) {
        float distance = super.getValue(vector);
        if (distance < 0 || distance > 3) {
            return distance;
        } else {
            float cylinderDistance;
            sphere.setCoordinate(vector);
            sphere.setRadius(3.0f);
            cylinderSubspace.collectCandidates(sphere, cylinderVolumes);

            // check candidates for intersection with the probe
            // compute minimum distance from sample point
            Volume volume = computeClosestVolume(vector, cylinderVolumes);
            cylinderVolumes.clear();
            if (volume != null) {
                cylinderDistance = volume.computeDistance(vector);
                return cylinderDistance < distance ? cylinderDistance : distance;
            } else {
                return distance;
            }
        }
    }

    /**
     * Method description.
     *
     * @param atomA Parameter description.
     * @param atomB Parameter description.
     *
     * @return Return description.
     */
    public ConcaveCylinder computeConcaveCylinder(Atom atomA, Atom atomB) {
        ConcaveCylinder cylinder = new ConcaveCylinder();
        cylinder.setCoordinateA(atomA.getCoordinate());
        cylinder.setCoordinateB(atomB.getCoordinate());
        cylinder.setRadiusA(atomA.getVanDerWaalsRadius());
        cylinder.setRadiusB(atomB.getVanDerWaalsRadius());
        cylinder.setRadius(1.5f);
        if (cylinder.getLength() - cylinder.getRadiusA() -
              cylinder.getRadiusB() < cylinder.getRadius()) {

            // compute connection vector between the atoms centers
            Vector3f w = new Vector3f(cylinder.getAxis());
            float distance = cylinder.getLength();
            float radiusA = cylinder.getRadiusA() + cylinder.getRadius();
            float radiusB = cylinder.getRadiusB() + cylinder.getRadius();
            float lengthA = (radiusA * radiusA) - (radiusB * radiusB);
            lengthA /= 2 * distance;
            lengthA += 0.5f * distance;
            float lengthB = distance - lengthA;
            float lengthC =
                (float) Math.sqrt(radiusA * radiusA - lengthA * lengthA);
            cylinder.setOffset(new Vector3f(lengthA, lengthC, 0));
            float factor;
            w.set(cylinder.getAxis());
            factor = lengthA / radiusA * cylinder.getRadiusA();
            w.scale(factor);
            w.add(cylinder.getCoordinateA());
            cylinder.setCoordinateA(w);
            cylinder.setRadiusA(factor * lengthC / lengthA);
            cylinder.setOffset(new Vector3f(lengthA - factor, lengthC, 0));
            w.set(cylinder.getAxis());
            factor = lengthB / radiusB * cylinder.getRadiusB();
            w.scale(-factor);
            w.add(cylinder.getCoordinateB());
            cylinder.setCoordinateB(w);
            cylinder.setRadiusB(factor * lengthC / lengthB);
            return cylinder;
        }
        return null;
    }
}
