/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.visualization;

import java.util.Collection;

import javax.vecmath.Point3f;
import javax.vecmath.Tuple3f;
import javax.vecmath.Vector3f;

/**
 * Volumetric description of a sphere.
 *
 * @author Karsten Klein
 *
 * @created May 31, 2001
 */
public class Sphere implements Volume {
    private Point3f coordinate = new Point3f();
    private float radius = 0;

    /**
     * Sets the <code>coordinate</code> attribute of the <code>Sphere</code> object.
     *
     * @param coordinate The new <code>coordinate</code> value.
     */
    public void setCoordinate(Tuple3f coordinate) {
        this.coordinate.set(coordinate);
    }

    /**
     * Sets the <code>radius</code> attribute of the <code>Sphere</code> object.
     *
     * @param radius The new <code>radius</code> value.
     */
    public void setRadius(float radius) {
        this.radius = radius;
    }

    /**
     * Gets the <code>coordinate</code> attribute of the <code>Sphere</code> object.
     *
     * @return The <code>coordinate</code> value.
     */
    public Point3f getCoordinate() {
        return coordinate;
    }

    /**
     * Gets the <code>radius</code> attribute of the <code>Sphere</code> object.
     *
     * @return The <code>radius</code> value.
     */
    public float getRadius() {
        return radius;
    }

    /**
     * Method description.
     *
     * @return Return description.
     */
    public float computeDiameter() {
        return 2 * getRadius();
    }

    /**
     * Method description.
     *
     * @param volumes Parameter description.
     */
    public void computeSubvolumes(Collection volumes) {
        Sphere volume;
        Point3f point = new Point3f();
        float radius = 0.5f * this.radius;
        for (int i = 0; i < 8; i++) {
            volume = new Sphere();
            point.set(coordinate);
            point.x += (i & 1) == 0 ? -radius : radius;
            point.y += (i & 2) == 0 ? -radius : radius;
            point.z += (i & 4) == 0 ? -radius : radius;
            volume.setCoordinate(point);
            volume.setRadius(0.87f * this.radius);
            volumes.add(volume);
        }
    }

    /**
     * Method description.
     *
     * @param tuple Parameter description.
     * @param epsilon Parameter description.
     *
     * @return Return description.
     */
    public boolean intersects(Tuple3f tuple, float epsilon) {
        Vector3f vector = new Vector3f(getCoordinate());
        vector.sub(tuple);
        return (getRadius() + epsilon) > vector.length();
    }

    /**
     * Method description.
     *
     * @param volume Parameter description.
     *
     * @return Return description.
     */
    public boolean intersects(Volume volume) {
        return volume.intersects(getCoordinate(), getRadius());
    }

    /**
     * Method description.
     *
     * @param volume Parameter description.
     *
     * @return Return description.
     */
    public boolean includes(Volume volume) {
        Sphere sphere = (Sphere) volume;
        Vector3f vector = new Vector3f();

        // compute direction vector
        vector.set(sphere.getCoordinate());
        vector.sub(getCoordinate());
        float length = vector.length();
        return (length + sphere.getRadius() < getRadius());
    }

    /**
     * Method description.
     *
     * @param tuple Parameter description.
     *
     * @return Return description.
     */
    public float computeDistance(Tuple3f tuple) {
        Vector3f vector = new Vector3f(tuple);
        vector.sub(getCoordinate());
        return vector.length() - getRadius();
    }

    /**
     * Method description.
     *
     * @return Return description.
     */
    public String toString() {
        return "Sphere volume definition: center " + getCoordinate() +
        "; radius " + radius;
    }
}
