/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.visualization;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Vector;

/**
 * Provides the facility of spatial subdivision. This is achieved by using the
 * <code>Volume</code> interface. The user has to provide a volume that is the top of
 * the hierarchical subdivision. This volume should include the provided volumes
 *
 * @author Karsten Klein
 *
 * @created May 29, 2001
 */
public class Subspace {
    private Volume volume = null;
    private Vector subspaces = null;
    private Vector elements = new Vector(100, 100);
    private int level = 0;

    /**
     * <code>Subspace</code> constructor.
     *
     * @param volume Description of parameter.
     */
    public Subspace(Volume volume) {
        setVolume(volume);
    }

    /**
     * Sets the <code>volume</code> attribute of the <code>Subspace</code> object.
     *
     * @param volume The new <code>volume</code> value.
     */
    public void setVolume(Volume volume) {
        this.volume = volume;
    }

    /**
     * Sets the <code>level</code> attribute of the <code>Subspace</code> object.
     *
     * @param level The new <code>level</code> value.
     */
    public void setLevel(int level) {
        this.level = level;
    }

    /**
     * Gets the <code>volume</code> attribute of the <code>Subspace</code> object.
     *
     * @return The <code>volume</code> value.
     */
    public Volume getVolume() {
        return volume;
    }

    /**
     * Gets the <code>level</code> attribute of the <code>Subspace</code> object.
     *
     * @return The <code>level</code> value.
     */
    public int getLevel() {
        return level;
    }

    /**
     * Subdivision of the volume and distribute the volumetric objects to the enw
     * subvolumes.
     */
    public void subdivide() {

        // compute the subvolumes
        ArrayList volumes = new ArrayList(8);
        volume.computeSubvolumes(volumes);

        // iterate over proposed subspace volumes and create new subspaces
        Iterator volumeIterator = volumes.iterator();
        subspaces = new Vector();
        Subspace subspace;
        while (volumeIterator.hasNext()) {
            subspace = new Subspace((Volume) volumeIterator.next());
            subspace.setLevel(getLevel() + 1);
            subspaces.add(subspace);
        }

        // distribute volumetric elements over the new subspaces
        Iterator subspaceIterator = subspaces.iterator();
        Iterator iterator;
        Volume element;
        Volume subspaceVolume;
        HashSet removeable = new HashSet();
        while (subspaceIterator.hasNext()) {
            subspace = (Subspace) subspaceIterator.next();
            subspaceVolume = subspace.getVolume();
            iterator = elements.iterator();
            while (iterator.hasNext()) {
                element = (Volume) iterator.next();

                // the addElement() method performs the intersection check!
                subspace.addElement(element);
            }
        }

        // remove all the elements that are on this level of the subspace subdivision
        elements.clear();
    }

    /**
     * Description of the method.
     *
     * @return Description of the returned value.
     */
    public boolean needsSubdivision() {
        if (elements != null && elements.size() > 10) {
            return true;
        }
        return false;
    }

    /**
     * Adds a <code>Element</code> object to the <code>Subspace</code> object.
     *
     * @param volume The <code>Element</code> object to be added.
     */
    public final boolean addElement(Volume volume) {
        if (volume.intersects(this.volume)) {
            elements.add(volume);
            return true;
        }
        return false;
    }

    /**
     * Description of the method.
     *
     * @param volume Description of parameter.
     * @param candidates Description of parameter.
     *
     * @return Description of the returned value.
     */
    public Subspace collectCandidates(Volume volume, Collection candidates) {
        Subspace cachedSubspace = this;
        if (subspaces == null) {
            candidates.addAll(elements);
        } else {
            Iterator iterator = subspaces.iterator();
            Subspace subspace;

            // search for inclusions
            while (iterator.hasNext()) {
                subspace = (Subspace) iterator.next();
                if (subspace.getVolume().includes(volume)) {
                    cachedSubspace = subspace;
                    subspace = subspace.collectCandidates(volume, candidates);

                    // cache subspace if it includes the volume
                    if (subspace != null) {
                        cachedSubspace = subspace;
                    }
                }
            }

            // search for intersections
            iterator = subspaces.iterator();
            while (iterator.hasNext()) {
                subspace = (Subspace) iterator.next();
                if (volume.intersects(subspace.getVolume())) {
                    subspace.collectCandidates(volume, candidates);
                }
            }
        }
        return cachedSubspace;
    }

    /**
     * Description of the method.
     *
     * @param volumes Description of parameter.
     */
    public void checkIntersection(Collection volumes) {
        Iterator iterator = volumes.iterator();
        Volume volume;
        while (iterator.hasNext()) {
            volume = (Volume) iterator.next();
        }
    }

    /**
     * Description of the method.
     *
     * @param subspace Description of parameter.
     * @param level Description of parameter.
     */
    public final void subdivide(int level) {
        Subspace childSubspace;
        if (level > 0 && needsSubdivision()) {
            subdivide();
            Iterator iterator = subspaces.iterator();
            while (iterator.hasNext()) {
                childSubspace = (Subspace) iterator.next();
                if (childSubspace.elements.isEmpty()) {

                    // :NOTE: not removing the empty node enhances the caching
                    // iterator.remove();
                } else {

                    // recursion
                    childSubspace.subdivide(level - 1);
                }
            }
        }
    }
}
