/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.visualization;

import javax.vecmath.Color3f;
import javax.vecmath.Color4f;
import javax.vecmath.Vector3f;

/**
 * Description of the class.
 *
 * @author Karsten Klein
 *
 * @created May 28, 2001
 */
public class Triangle {
    Point[] points = new Point[3];
    Color3f color = new Color3f(0.7f, 0.7f, 0.2f);

    /**
     * Sets the <code>color</code> attribute of the <code>Triangle</code> object.
     *
     * @param color The new <code>color</code> value.
     */
    public void setColor(Color3f color) {
        this.color.set(color);
    }

    /**
     * Gets the <code>points</code> attribute of the <code>Triangle</code> object.
     *
     * @return The <code>points</code> value.
     */
    public Point[] getPoints() {
        return points;
    }

    /**
     * Description of the method.
     *
     * @param normal Description of parameter.
     */
    public void modifyNormals(Vector3f normal) {
        float lengthSquared = normal.lengthSquared();
        if (Float.isNaN(lengthSquared)) {
            for (int i = 0; i < 3; i++) {
                if (points[i].getNormal() == null) {
                    points[i].setNormal(new Vector3f(0, 0, 0));
                }
            }
            return;
        }
        Vector3f center = new Vector3f(points[0].getCoordinate());
        center.add(points[1].getCoordinate());
        center.add(points[2].getCoordinate());
        center.scale(1.0f / 3);
        Vector3f dummy = new Vector3f();
        Vector3f pointNormal;
        float[] distances = new float[4];
        distances[3] = 0;
        for (int index = 0; index < 3; index++) {
            dummy.set(points[index].getCoordinate());
            dummy.sub(center);
            distances[index] = dummy.length();
            distances[3] += distances[index];
        }
        for (int index = 0; index < 3; index++) {
            pointNormal = points[index].getNormal();
            dummy.set(normal);
            dummy.scale(distances[index] / distances[3]);
            if (pointNormal == null) {
                pointNormal = new Vector3f(dummy);
                points[index].setNormal(pointNormal);
            } else {
                pointNormal.add(dummy);
            }
        }
    }

    /**
     * Description of the method.
     */
    public void normalizeNormals() {
        Vector3f pointNormal;
        for (int index = 0; index < 3; index++) {
            pointNormal = points[index].getNormal();
            if (pointNormal != null) {
                pointNormal.normalize();
            }
        }
    }

    /**
     * Description of the method.
     *
     * @return Description of the returned value.
     */
    public Vector3f computeOrthogonal() {
        Vector3f connect1 = new Vector3f(points[1].getCoordinate());
        connect1.sub(points[0].getCoordinate());
        Vector3f connect2 = new Vector3f(points[2].getCoordinate());
        connect2.sub(points[0].getCoordinate());
        Vector3f orthogonal = new Vector3f();
        orthogonal.cross(connect1, connect2);
        orthogonal.normalize();
        return orthogonal;
    }

    /**
     * Description of the method.
     *
     * @return Description of the returned value.
     */
    public Vector3f computeAvarageNormal() {
        Vector3f normal = new Vector3f();
        Vector3f pointNormal;
        for (int index = 0; index < 3; index++) {
            pointNormal = points[index].getNormal();
            if (pointNormal != null) {
                normal.add(pointNormal);
            }
        }
        normal.normalize();
        return normal;
    }

    /**
     * Method description.
     *
     * @return Return description.
     */
    public Color4f computeAvarageColor() {
        Color4f avarageColor = new Color4f();
        Color4f color;
        int counter = 0;
        for (int index = 0; index < 3; index++) {
            color = points[index].getColor();
            if (color != null) {
                avarageColor.add(color);
                counter++;
            }
        }
        if (counter > 0) {
            avarageColor.scale(1.0f / counter);
        }
        return avarageColor;
    }

    /**
     * Method description.
     *
     * @param a Parameter description.
     * @param b Parameter description.
     */
    public void collapse(Point a, Point b) {

        // triangle is meant to replace point a by point b
        boolean isCollapsed = false;
        if (getPoints()[0] == a) {
            getPoints()[0] = b;
            isCollapsed = true;
        }
        if (getPoints()[1] == a) {
            getPoints()[1] = b;
            isCollapsed = true;
        }
        if (getPoints()[2] == a) {
            getPoints()[2] = b;
            isCollapsed = true;
        }
        if (isCollapsed) {
            b.getTriangles().addAll(a.getTriangles());
        }
    }

    /**
     * Method description.
     *
     * @param a Parameter description.
     * @param b Parameter description.
     * @param epsilon Parameter description.
     *
     * @return Return description.
     */
    public boolean collapsed(int a, int b, float epsilon) {
        Vector3f connect = new Vector3f(points[b].getCoordinate());
        connect.sub(points[a].getCoordinate());
        float length = connect.length();
        if (length < epsilon && length > 0.01f) {
            return true;
        }
        return false;
    }

    //  public boolean isCollapsedIdentical() {
    //
    //    return getPoints()[ 0 ] == getPoints()[1] ||
    //      getPoints()[ 0 ] == getPoints()[2] ||
    //      getPoints()[ 1 ] == getPoints()[2];
    //
    //  }
    public boolean isCollapsed() {
        return getPoints()[0].isIdentical(getPoints()[1], 0.01f) ||
        getPoints()[0].isIdentical(getPoints()[2], 0.01f) ||
        getPoints()[1].isIdentical(getPoints()[2], 0.01f);
    }

    /**
     * Method description.
     *
     * @param factor Parameter description.
     *
     * @return Return description.
     */
    public boolean pointCollapse(int factor) {
        getPoints()[0].collapse(factor);
        getPoints()[1].collapse(factor);
        getPoints()[2].collapse(factor);
        return isCollapsed();
    }
}
