/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.visualization;

import java.util.Collection;

import javax.vecmath.Tuple3f;

/**
 * Interface defining volumetric objects.
 *
 * @author Karsten Klein
 *
 * @created June 8, 2001
 */
public interface Volume {

    /**
     * Computes the subvolumes of an volumetric object. This method can be used for
     * computing hierachical subdivisions. The subdivisions should compose the full
     * original <code>Volume</code> using elements that are definitly smaller. V &lt; S1
     * ^ S2 ^ S3 ^ ... ^ Sn, Si &lt; V
     *
     * @param volumes Collection to be filled with new <code>Volume</code> s.
     */
    public void computeSubvolumes(Collection volumes);

    /**
     * Checks for an intersection with the specified <code>Volume</code> .
     *
     * @param volume <code>Volume</code> to check intersection with.
     *
     * @return Boolean value indicating intersection.
     */
    public boolean intersects(Volume volume);

    /**
     * Checks for an intersection of the <code>Volume</code> with a epsilon sphere at a
     * specified position.
     *
     * @param tuple Position of interest.
     * @param epsilon Scalar value interpreted as epsilon area around the position.
     *
     * @return Intersection result.
     */
    public boolean intersects(Tuple3f tuple, float epsilon);

    /**
     * Checks for an inclusion of the <code>Volume</code> . If the specified <code>Volume
     * </code> is included by this volumetric object, this method returns
     * <code>true</code> .
     *
     * @param volume <code>Volume</code> of interest.
     *
     * @return Inclusion result.
     */
    public boolean includes(Volume volume);

    /**
     * Computes the distance from the specified coordinate to the surface of the volume.
     * For a coordinate, which is inside the volume the method returns a negative
     * result. The distance is always the minimal distance in euclidean space.
     *
     * @param tuple Specifies the coordinate for the distance measurement.
     *
     * @return Minimal euclidean distance of the coordinate to the volume surface. If the
     *         coordinate is inside the <code>Volume</code> this gives a negative
     *         result.
     */
    public float computeDistance(Tuple3f tuple);

    /**
     * Computes the diameter of the <code>Volume</code> object. The max diameter can be
     * an estimation of the diameter of the <code>Volume</code> object. It dosn't have
     * to be exact, but it has to be at least bigger than the real diameter of the
     * <code>Volume</code> instance. One can consider it as diameter of a bounding
     * sphere.
     *
     * @return The <code>maxDiameter</code> value.
     */
    public float computeDiameter();
}
