/** 
 *    Part of SRS 3D, a system for viewing macromolecules (http://srs3d.org)
 *    
 *    Copyright (C) 2007-2008
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 */

package org.srs3d.viewer.j3d;

import javax.media.j3d.BranchGroup;
import javax.media.j3d.Node;

import org.srs3d.viewer.j3d.creators.GeometryCreator;
import org.srs3d.viewer.objects.AbstractObject;
import org.srs3d.viewer.util.ExceptionHandler;

/**
 * The <code>Spawner</code> is a special <code>SceneGraphObject</code> that is able to
 * change representation dynamically no matter if it's alive and/or compiled. This
 * functionality is achieved by keeping an internal <code>BranchGroup</code> node that
 * can be replaced and recreated any time the <code>Spawner</code> wants/is meant to.
 * This behavior is meant to be fully transparent to the user adding additional nodes to
 * the <code>Spawner</code> . The SceneGraph traversing algorithm has full insight into
 * the structure. Moreover the <code>Spawner</code> works on an
 * <code>AbstractObject</code> instance.
 *
 * @author Karsten Klein, 02/2001
 *
 * @created February 02, 2001
 *
 * @see AbstractObject, GeometryCreator
 */
public class Spawner extends BranchGroup {

    /** Boolean flag indicating whether the spawner is spawned yet */
    private boolean isSpawned = false;

    /** The associated <code>AbstractObject</code> */
    private AbstractObject object = null;

    /** Internal <code>BranchGroup</code> to achieve full dynamic capabilities */
    private BranchGroup internalBranch = null;

    /**
     * Constructs a new <code>Spawner</code> instance.
     *
     * @param object object the associated object.
     */
    public Spawner(AbstractObject object) {
        this.object = object;
        BranchGroupHelper.setDefaultCapabilities(this);
    }

    /**
     * Gets the <code>object</code> attribute of the <code>Spawner</code> object.
     *
     * @return The <code>object</code> value.
     */
    public AbstractObject getObject() {
        return object;
    }

    /**
     * This method should be called before releasing the object forever. After this call
     *
     * @param contextData Description of parameter.
     */
    public void cleanup(ContextData contextData) {
        if (object != null) {

            // decouple from user data
            setUserData(null);
            try {

                // detach and delete internal branch (deep deletion)
                deleteInternalBranch(contextData);

                // detach from parent node
                detach();
            } catch (Exception e) {

                // silently
                ExceptionHandler.handleException(e,
                    ExceptionHandler.SILENT_IN_DEBUG);
            }
            AbstractObject localObject = object;
            object = null;
            try {

                // remove associated shapes
                contextData.getSpawnerManager().remove(localObject);
                contextData.getShapeManager().remove(localObject);
            } catch (java.lang.NullPointerException e) {
                ExceptionHandler.handleException(e,
                    ExceptionHandler.SILENT_IN_RELEASE);
            }
        }
    }

    /**
     * Description of the method. Description of the method. Description of the method.
     *
     * @param contextData Description of parameter.
     * @param complete Description of parameter.
     */
    public void spawn(ContextData contextData, boolean complete) {
        if (isSpawned) {
            deleteInternalBranch(contextData);
        }
        if (object != null) {
            GeometryCreator geometryCreator =
                contextData.getGeometryCreatorFactory().getGeometryCreator(object);
            if (geometryCreator != null) {
                geometryCreator.create(object, this);
            }
        }

        // just in case it was not created yet
        createInternalBranch();
        if (complete) {
            completeSpawn();
        }
    }

    /**
     * Description of the method.
     */
    public void completeSpawn() {

        // the last step is attaching the internal branch group
        super.addChild(internalBranch);
        isSpawned = true;
    }

    /**
     * Spawn the objects geometry.
     *
     * @param contextData Description of parameter.
     */
    public void spawn(ContextData contextData) {
        spawn(contextData, true);
    }

    /**
     * Overwriting the Group method addChild
     *
     * @param child The <code>Child</code> object to be added.
     */
    public void addChild(Node child) {
        createInternalBranch();
        try {

            // delegate to internal branch
            internalBranch.addChild(child);
        } catch (Exception e) {
            ExceptionHandler.handleException(e,
                ExceptionHandler.SILENT_IN_RELEASE);
        }
    }

    /**
     * Description of the method.
     *
     * @param contextData Description of parameter.
     */
    public void deleteInternalBranch(ContextData contextData) {
        if (internalBranch != null) {
            try {
                BranchGroupHelper.detach(contextData, internalBranch);
                internalBranch.detach();
            } catch (Exception e) {
                ExceptionHandler.handleException(e,
                    ExceptionHandler.SILENT_IN_RELEASE);
            }
            internalBranch = null;
        }
        isSpawned = false;
    }

    /**
     * Description of the method.
     */
    public void createInternalBranch() {
        if (internalBranch == null) {
            internalBranch = new BranchGroup();
            BranchGroupHelper.setDefaultCapabilities(internalBranch);
            isSpawned = false;
        }
    }

    /**
     * Description of the method.
     *
     * @return Description of the returned value.
     */
    public String toString() {
        return getClass().toString() + ": representing object " + object;
    }

    /**
     * Method description.
     */
    public void compile() {
        if (internalBranch != null && !isSpawned) {
            internalBranch.compile();
        }
    }
}
